package agent;

import geometry.Mathematic;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import rescuecore2.messages.Command;
import rescuecore2.standard.components.StandardAgent;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.AKMove;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import utilities.Logger;
import utilities.Timer;
import utilities.config.Config;
import worldModel.Area;
import worldModel.Blockade;
import worldModel.Entrance;
import worldModel.GeometryTools;
import worldModel.Refuge;
import worldModel.Road;
import worldModel.WorldModel;
import agent.ThinkAgent.NoCommiStrategy;

import com.vividsolutions.jts.algorithm.Angle;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.triangulate.Segment;

public abstract class Agent<E extends StandardEntity> extends StandardAgent<E> {

	public boolean wasIInSearch = false;
	public boolean amIInSearch = false;

	public NoCommiStrategy strategy = null;

	public Config psdConfig = null;
	public Logger logger = null;
	public boolean bigMap = true;
	public FireBrigadeAgentNoHear fbNoHear = new FireBrigadeAgentNoHear();

	public static int maxDistanceExtinguish = -1;
	public double maxVolumeOfAllBuildings = -Integer.MAX_VALUE;
	public double averageOfSurfaces = 0;
	public int time = -1;
	protected Timer cycleTimer = null;
	protected WorldModel worldModel = null;
	public static final String MAX_DISTANCE_EXTINGUISH_KEY = "fire.extinguish.max-distance";
	public static final String MAX_DISTANCE_VIEW_KEY = "perception.los.max-distance";
	public static final String WATER_REFILL_RATE_REFUGE_KEY = "fire.tank.refill_rate";
	public static final String WATER_REFILL_RATE_HYDRANT_KEY = "fire.tank.refill_hydrant_rate";
	public boolean preCompute = false;
	protected Collection<Command> heard = null;
	protected Map<Integer, ActionCommandException> commandHistory = new HashMap<Integer, ActionCommandException>();
	protected ArrayList<worldModel.Road> sentBlocks = new ArrayList<worldModel.Road>();
	public int manLastDeadTime = 0;
	public int cvRecognizer = 32;
	protected final int STUCKCYCLES = 4;

	protected boolean isHuman = false;
	public boolean movingToRefuge = false;
	public boolean lowCommi = false;
	public boolean noCommi = false;
	public boolean amIStucking = false;
	public boolean isInMoveToFire = false;
	boolean IamOnfire = false;

	protected Map<EntityID, Set<EntityID>> neighbours = new HashMap<EntityID, Set<EntityID>>();
	public int bfsLayerSize = -1;
	int centerForPolice = 0;
	int centerForAmb = 0;
	int centerForFire = 0;
	protected CenterForAmbulance pg = null;
	protected CenterForFire kimpar = null;
	protected CenterForPolice safie = null;
	// shapeRe
	protected int timeForNewWayMemory = -1;

	public HashMap<MapType, int[][]> mapDetails = new HashMap<>();
	public double[][][] mapBorders = new double[5][5][4];

	// No Commi variables
	protected EntityID extinguishedTarg = null;
	protected int extinguishTime = -1;
	protected HashSet<worldModel.Building> shouldSayFireds = new HashSet<>();
	protected HashSet<worldModel.Building> shouldSayOutFireds = new HashSet<>();
	protected HashSet<worldModel.Building> shouldSayExtinguished = new HashSet<>();
	protected HashMap<worldModel.Building, Integer> lastSeenBy = new HashMap<worldModel.Building, Integer>();

	// public static void main(String[] args) {
	// fillMapDetails();
	// int min = 100;
	// for (Entry<MapType, int[][]> e1 : mapDetails.entrySet())
	// for (Entry<MapType, int[][]> e2 : mapDetails.entrySet()) {
	// int d = diff(e1.getValue(), e2.getValue());
	// if (d < min && d != 0)
	// min = d;
	// System.out.println("diff: " + d);
	// }
	// System.out.println("min: " + min);
	// }
	public worldModel.Building getBestBuildingForSearchArondZone(FireZone fz) {
		double minDist = Integer.MAX_VALUE;
		worldModel.Building best = null;
		for (worldModel.Building b : fz.buildings) {
			if (minDist > Math.hypot(b.getX() - fz.centerX, b.getY()
					- fz.centerY)) {
				minDist = Math.hypot(b.getX() - fz.centerX, b.getY()
						- fz.centerY);
				best = b;
			}

		}
		return best;
	}

	public void fillMapDetails() {
		// Maps has diffrence up to minimum 10 cell nums! Interesting!
		int[][] aVC = { { 2, 0, 0, 0, 1 }, { 0, 0, 0, 0, 0 },
				{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 1, 0, 0, 0, 1 } };
		int[][] aMexico = { { 0, 2, 1, 0, 3 }, { 1, 0, 0, 0, 1 },
				{ 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 2, 0, 0, 1 } };
		int[][] aBerlin = { { 1, 2, 1, 2, 4 }, { 2, 0, 0, 0, 1 },
				{ 0, 0, 0, 0, 0 }, { 2, 0, 0, 0, 0 }, { 4, 1, 0, 0, 4 } };
		int[][] aEindhoven = { { 0, 1, 2, 2, 1 }, { 1, 0, 0, 0, 2 },
				{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 1 }, { 3, 3, 0, 0, 2 } };
		int[][] aIstanbul = { { 6, 0, 0, 0, 3 }, { 1, 0, 0, 0, 2 },
				{ 0, 0, 0, 0, 2 }, { 0, 0, 0, 0, 1 }, { 4, 0, 0, 3, 1 } };
		int[][] aParis = { { 3, 1, 2, 0, 2 }, { 1, 0, 0, 0, 0 },
				{ 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 2 }, { 4, 1, 1, 0, 2 } };
		int[][] aKobe = { { 0, 3, 1, 2, 0 }, { 1, 0, 0, 0, 2 },
				{ 1, 0, 0, 0, 1 }, { 0, 0, 0, 0, 0 }, { 4, 1, 0, 0, 2 } };
		int[][] aJoao = { { 2, 0, 0, 2, 3 }, { 0, 0, 0, 0, 0 },
				{ 2, 0, 0, 0, 0 }, { 0, 0, 0, 0, 1 }, { 1, 4, 0, 0, 1 } };
		int[][] aTest = { { 0, 0, 2, 1, 2 }, { 0, 1, 0, 0, 0 },
				{ 0, 1, 0, 0, 1 }, { 0, 1, 0, 1, 0 }, { 0, 0, 2, 0, 0 } };
		int[][] aNY = { { 4, 0, 1, 0, 5 }, { 3, 0, 0, 0, 0 },
				{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 1 }, { 3, 1, 0, 1, 1 } };
		mapDetails.put(MapType.Kobe, aKobe);
		mapDetails.put(MapType.VC, aVC);
		mapDetails.put(MapType.Berlin, aBerlin);
		mapDetails.put(MapType.Paris, aParis);
		mapDetails.put(MapType.Istanbul, aIstanbul);
		mapDetails.put(MapType.Eindhoven, aEindhoven);
		mapDetails.put(MapType.Mecxico, aMexico);
		mapDetails.put(MapType.Joao, aJoao);
		mapDetails.put(MapType.Test, aTest);
		mapDetails.put(MapType.NY, aNY);
	}

	public int diff(int[][] a1, int[][] a2) {
		int diffs = 0;
		for (int i = 0; i < a1.length; i++)
			for (int j = 0; j < a1[0].length; j++)
				if (a1[i][j] != a2[i][j])
					diffs += Math.abs(a1[i][j] - a2[i][j]);
		return diffs;
	}

	public void hearRF(worldModel.Building build, int rf) {
		if (rf % 2 == 0) {
			build.timeOfStartOfFire = -1;
			build.stFire = 0;
		} else {
			build.stFire = 1;
			if (build.timeOfStartOfFire == -1)
				build.timeOfStartOfFire = time;
		}
		if (rf < 2)
			build.setFireness(0);
		if (rf == 2 || rf == 3)
			build.setFireness(1);
		if (rf == 4 || rf == 5)
			build.setFireness(5);
		if (rf == 6 || rf == 7)
			build.setFireness(8);
		build.lastFireness = rf;
	}

	public void setSTFire() {
		for (EntityID id : worldModel.changeSet.getChangedEntities()) {
			if (worldModel.getEntity(id) instanceof Building) {

				worldModel.Building build = (worldModel.Building) worldModel
						.getArea(id);
				if (build.isFirenessDefined() && build.getFieryness() == 0) {
					if (build.isTemperatureDefined()
							&& build.getTemperature() > 20) {
						build.stFire = 1;
					} else {
						build.stFire = 0;
					}
				}
				if (build.isFirenessDefined()
						&& (build.getFieryness() > 0 && build.getFieryness() < 4)) {
					build.stFire = 1;
				}
				if (build.isFirenessDefined() && build.getFieryness() > 4) {
					build.stFire = 0;
				}
				if (build.isFirenessDefined() && build.getFieryness() == 4) {
					if (build.isTemperatureDefined()
							&& build.getTemperature() > 20) {
						build.stFire = 1;
					} else {
						build.stFire = 0;
					}
				}
				if (build instanceof Refuge
						|| build.serverArea instanceof PoliceOffice
						|| build.serverArea instanceof FireStation
						|| build.serverArea instanceof AmbulanceCentre) {
					build.stFire = 0;
				}
			}
		}
	}

	// shapeRe
	protected void initBaseVars() {
		// Human.HP_PRECISION =
		// config.getIntValue("perception.los.precision.hp");
		// Human.DAMAGE_PRECISION = config
		// .getIntValue("perception.los.precision.damage");
		cycleTimer = new Timer();
		cycleTimer.setTime(0);
		if (psdConfig.get("Agent", "AgentLog", "LogToFile").equals("True"))
			logger = new Logger(
					psdConfig.get("Agent", "AgentLog", "LogAddress") + "/"
							+ getName().replaceAll("agent.", "") + "-"
							+ me().getID().getValue(), cycleTimer);
		else
			logger = new Logger(null, cycleTimer);
		if (psdConfig.get("Agent", "Public", "PreCompute").equals("True")) {
			preCompute = true;
		}

	}

	// @SuppressWarnings("unused")
	// private void setMapType() {
	// int buildingsSize = model.getEntitiesOfType(StandardEntityURN.BUILDING)
	// .size(), roadsSize = model.getEntitiesOfType(
	// StandardEntityURN.ROAD).size();
	// if (1253 <= buildingsSize && buildingsSize <= 1273 && 1944 <= roadsSize
	// && roadsSize <= 1964)
	// map = MapType.VC;
	// if (1608 <= buildingsSize && buildingsSize <= 1628 && 3015 <= roadsSize
	// && roadsSize <= 3035)
	// map = MapType.Paris;
	// if (1416 <= buildingsSize && buildingsSize <= 1436 && 3375 <= roadsSize
	// && roadsSize <= 3395)
	// map = MapType.Berlin;
	// if ((726 <= buildingsSize && buildingsSize <= 746 && 1505 <= roadsSize &&
	// roadsSize <= 1525)
	// || (747 <= buildingsSize && buildingsSize <= 767
	// && 1592 <= roadsSize && roadsSize <= 1612))
	// map = MapType.Kobe;
	// if (1234 <= buildingsSize && buildingsSize <= 1254 && 3327 <= roadsSize
	// && roadsSize <= 3347)
	// map = MapType.Istanbul;
	//
	// if (1298 <= buildingsSize && buildingsSize <= 1318 && 5162 <= roadsSize
	// && roadsSize <= 5182)
	// map = MapType.Eindhoven;
	// if (1537 <= buildingsSize && buildingsSize <= 1557 && 5084 <= roadsSize
	// && roadsSize <= 5104)
	// map = MapType.Mecxico;
	//
	// if (map == MapType.Kobe || map == MapType.VC)
	// bigMap = false;
	// System.out.println("///// roads " + roadsSize + " buildings "
	// + buildingsSize);
	// System.out.println("mapType " + map);
	// }

	long lll = -1;

	public void logTime() {
		if (lll < 0) {
			lll = System.nanoTime();
		} else {
			long m = System.nanoTime();
			System.out.println("Time: " + ((m - lll) / 1000000));
			lll = m;
		}
	}

	protected void postConnect() {
		// for (int i = 0; i < harchi.length; i++)
		// for (int j = 0; j < harchi[i].length; j++)
		// harchi[i][j] = 0;
		//
		// for (int i = 0; i < lastPacksSize.length; i++)
		// lastPacksSize[i] = 0;
		//
		// for (int i = 0; i < notInNearsSize.length; i++)
		// notInNearsSize[i] = 0;
		//
		// for (int i = 0; i < lastBuildsToSaySize.length; i++)
		// lastBuildsToSaySize[i] = 0;

		super.postConnect();
		// long l = System.nanoTime();
		System.out.println("Start of Precomputation");
		initBaseVars();
		maxDistanceExtinguish = config.getIntValue(MAX_DISTANCE_EXTINGUISH_KEY);
		// refillRateRefuge = config
		// .getIntValue(Agent.WATER_REFILL_RATE_REFUGE_KEY);
		// refillRateHydrant = config
		// .getIntValue(Agent.WATER_REFILL_RATE_HYDRANT_KEY);
		isHuman = (me()) instanceof Human;
		initWorldModel();
		worldModel.myEntrances.addAll(getMyPos().entrances);
		worldModel.updateReachableAreas();
		for (worldModel.Building b : worldModel.buildings) {
			b.maxDistanceExtingiush = maxDistanceExtinguish;
		}
		// long m = System.nanoTime();
		// long d = (m - l) / 1000000;
		// System.out.println("Time of precompuation of base: " + d);
		saveToFile();
		// mapCalculations();
		buildWorks();
		// visionRoadsAndBuildings();
		for (Area a : worldModel.areas) {
			Set<EntityID> s = new HashSet<EntityID>();
			for (Entrance e : a.entrances)
				s.add(e.neighbour.area.getID());
			neighbours.put(a.getID(), s);
		}
		for (Human h : worldModel.agents) {
			Area a = worldModel.getArea(h.getPosition());
			h.firstPosition = a;
			h.firstX = a.getX();
			h.firstY = a.getY();
		}
		fillLastSeenBy();
	}

	private void fillLastSeenBy() {
		for (worldModel.Building b : worldModel.buildings)
			lastSeenBy.put(b, 0);
	}

	private void buildWorks() {
		for (worldModel.Building b : worldModel.buildings) {
			double hajm = b.getGroundArea() * b.getFloors();
			averageOfSurfaces += b.getGroundArea();
			if (hajm > maxVolumeOfAllBuildings)
				maxVolumeOfAllBuildings = hajm;
		}

		averageOfSurfaces /= worldModel.buildings.size();

	}

	private void updatingWayMemory() {
		sentBlocks = new ArrayList<>();
		worldModel.getTrafficSimulator().clearBlockades();
		for (Area area : worldModel.areas) {
			area.blockades = new ArrayList<Blockade>();
			ArrayList<Entrance> enterances = area.entrances;
			for (int i = 0; i < enterances.size(); i++) {
				for (int j = 0; j < i; j++) {
					if (!enterances.get(i).internalEntrances
							.contains(enterances.get(j)))
						enterances.get(i).internalEntrances.add(enterances
								.get(j));
					if (!enterances.get(j).internalEntrances
							.contains(enterances.get(i)))
						enterances.get(j).internalEntrances.add(enterances
								.get(i));
					enterances.get(j).isItConnectedToNeighbour = true;
					enterances.get(j).updatePoints(area.blockades);
					enterances.get(j).resetShapeInArea();
				}
				enterances.get(i).isItConnectedToNeighbour = true;
				enterances.get(i).updatePoints(area.blockades);
				enterances.get(i).resetShapeInArea();
			}
		}
	}

	public void visionRoadsAndBuildings() {
		// long l = System.nanoTime();
		visionRoads();
		// long m = System.nanoTime();
		// long d = (m - l) / 1000000;
		visionBuildings();
		// long mm = System.nanoTime();
		// long dd = (mm - m) / 1000000;
	}

	protected void saveToFile() {

		// long l = System.nanoTime();
		// System.out.println("(hasRead) " + (hasRead) + " precompute "
		// + preCompute);
		if (!preCompute || hasRead)
			return;
		try {
			File file = getWorldConfigFile();
			if (file.exists()) {
				return;
			}
			ObjectOutputStream oos = new ObjectOutputStream(
					new FileOutputStream(file));
			oos.writeObject(worldModel);
			hasRead = true;
			oos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		// long m = System.nanoTime();
		// long d = (m - l) / 1000000;
	}

	private File getWorldConfigFile() {
		return new File("./precompute/model-" + mapHashCode() + ".bin");
	}

	private String mapHash = null;

	protected String mapHashCode() {
		if (mapHash == null) {
			int buildingsCount = model.getEntitiesOfType(
					StandardEntityURN.BUILDING).size();
			int roadsCount = model.getEntitiesOfType(StandardEntityURN.ROAD)
					.size();
			mapHash = String.valueOf(buildingsCount * 2
					+ (long) Math.pow(roadsCount, 3));
		}
		return mapHash;
	}

	public boolean hasRead = false;

	public rescuecore2.config.Config getConfig() {
		return config;
	}

	protected void initWorldModel() {
		// long l = System.nanoTime();
		boolean shouldInstantiate = true;
		try {
			if (!preCompute) {
				File file = getWorldConfigFile();
				if (file.exists()) {
					ObjectInputStream ois = new ObjectInputStream(
							new FileInputStream(file));
					worldModel = (WorldModel) ois.readObject();
					ois.close();
					worldModel.setTransients(this, model);
					shouldInstantiate = false;
					hasRead = true;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		if (shouldInstantiate) {
			worldModel = new WorldModel(this, model);
			visionRoadsAndBuildings();
		}
		// long m = System.nanoTime();
		// long d = (m - l) / 1000000;
	}

	protected void selectCenterForPolice() {
		centerForPolice = 0;
		if (worldModel.policeOffices.size() > 0)
			for (Entity e : worldModel.policeOffices) {
				centerForPolice = e.getID().getValue();
				break;
			}
		else if (worldModel.ambulanceCenters.size() > 0)
			for (Entity e : worldModel.ambulanceCenters) {
				centerForPolice = e.getID().getValue();
				break;
			}
		else if (worldModel.fireStations.size() > 0)
			for (Entity e : worldModel.fireStations) {
				centerForPolice = e.getID().getValue();
				break;
			}
		else if (worldModel.policeForces.size() > 0) {
			ArrayList<Integer> policeForces = new ArrayList<Integer>();
			ArrayList<Integer> goodPoliceForces = new ArrayList<Integer>();
			for (PoliceForce e : worldModel.policeForces) {
				policeForces.add(e.getID().getValue());
				if (e.firstPosition instanceof Road)
					goodPoliceForces.add(e.getID().getValue());
			}
			if (goodPoliceForces.size() == 0)
				goodPoliceForces.addAll(policeForces);
			Collections.sort(goodPoliceForces);
			centerForPolice = goodPoliceForces.get(0);
		}
	}

	protected void selectCenterForFire() {
		centerForFire = 0;
		if (worldModel.fireStations.size() > 0)
			for (Entity e : worldModel.fireStations) {
				centerForFire = e.getID().getValue();
				break;
			}
		else if (worldModel.policeOffices.size() > 0)
			for (Entity e : worldModel.policeOffices) {
				centerForFire = e.getID().getValue();
				break;
			}
		else if (worldModel.ambulanceCenters.size() > 0)
			for (Entity e : worldModel.ambulanceCenters) {
				centerForFire = e.getID().getValue();
				break;
			}
		else if (worldModel.fireBrigades.size() > 0) {
			ArrayList<Integer> fireBg = new ArrayList<Integer>();
			ArrayList<Integer> goodFireBg = new ArrayList<Integer>();
			for (FireBrigade e : worldModel.fireBrigades) {
				fireBg.add(e.getID().getValue());
				if (e.firstPosition instanceof Road)
					goodFireBg.add(e.getID().getValue());
			}
			if (goodFireBg.size() == 0)
				goodFireBg.addAll(fireBg);
			Collections.sort(goodFireBg);
			centerForFire = goodFireBg.get(0);
		}
	}

	protected void selectCenterForAmb() {
		centerForAmb = 0;
		if (worldModel.ambulanceCenters.size() > 0)
			for (Entity e : worldModel.ambulanceCenters) {
				centerForAmb = e.getID().getValue();
				break;
			}
		else if (worldModel.policeOffices.size() > 0)
			for (Entity e : worldModel.policeOffices) {
				centerForAmb = e.getID().getValue();
				break;
			}
		else if (worldModel.fireStations.size() > 0)
			for (Entity e : worldModel.fireStations) {
				centerForAmb = e.getID().getValue();
				break;
			}
		else if (worldModel.ambulanceTeams.size() > 0) {
			ArrayList<Integer> ambTeam = new ArrayList<Integer>();
			ArrayList<Integer> goodAmbTeam = new ArrayList<Integer>();
			for (AmbulanceTeam e : worldModel.ambulanceTeams) {
				ambTeam.add(e.getID().getValue());
				if (e.firstPosition instanceof Road)
					goodAmbTeam.add(e.getID().getValue());
			}
			if (goodAmbTeam.size() == 0)
				goodAmbTeam.addAll(ambTeam);
			Collections.sort(goodAmbTeam);
			centerForAmb = goodAmbTeam.get(0);
		}
	}

	public int calculatedRF(worldModel.Building build) {
		int rf = 0;
		if (build.isFirenessDefined()) {
			if (build.getFieryness() == 0)
				rf = build.stFire;
			if (build.isOnFire())
				rf = 2 + build.stFire;
			if (build.getFieryness() > 3 && build.getFieryness() < 8)
				rf = 4 + build.stFire;

			if (build.getFieryness() == 8)
				rf = 6 + build.stFire;
		}
		return rf;
	}

	protected void updateWays(Area lastPos, Area newPos, int agentID) {
		// if (me().getID().getValue() != centerForAmb
		// && me().getID().getValue() != centerForFire)
		// return;
		if (lastPos == null || lastPos == newPos)
			return;
		ArrayList<Area> way = worldModel.getMinPathByHashSet(lastPos, newPos,
				false);
		Entrance lastE = null;
		if (way == null || way.size() == 0)
			return;
		for (int i = 0; i < way.size() - 1; i++) {
			way.get(i).HLTV = time - 1;
			way.get(i).setHLTS(time - 1, agentID, lastSeenBy);
			// log("daram HLTV e " + way.get(i) + " " + (time - i) +
			// " mikonim");
			for (Entrance enter : way.get(i).entrances)
				if (enter.neighbour.area.getID().getValue() == way.get(i + 1)
						.getID().getValue()) {
					if (lastE != null
							&& !lastE.internalEntrances.contains(enter)) {
						lastE.internalEntrances.add(enter);
						enter.internalEntrances.add(lastE);
					}
					enter.isItConnectedToNeighbour = true;
					enter.neighbour.isItConnectedToNeighbour = true;
					lastE = enter.neighbour;
					break;
				}
		}
	}

	// static int[][] harchi = new int[500][34];
	// int[][] harchiNo = new int[500][34];

	// public static int[] lastPacksSize = new int[500];
	// public static int[] notInNearsSize = new int[500];
	// public static int[] lastBuildsToSaySize = new int[500];
	//
	// public static ArrayList<Integer> CVsSended = new ArrayList<Integer>();

	public void printRadarDetails() {
		// // if (me() instanceof AmbulanceTeam
		// // && worldModel.ambulanceTeams.indexOf(me()) == 0) {
		// // // String s = "";
		// // // int rrrr = 0;
		// // // for (int cv : ciiiiivilllsss) {
		// // // s += cv + " ";
		// // // if (worldModel.getAreaByInt(cv).isReachable)
		// // // rrrr++;
		// // // }
		// // // System.out.println("CCCCCVVVVVVVVVV: " + s);
		// // // System.out.println("all size: " + ciiiiivilllsss.size());
		// // // System.out.println("reachable size: " + rrrr);
		// // }
		// // System.out.println("CCCCCVVVVVVVVVV: " + s);
		// // System.out.println("all size: " + ciiiiivilllsss.size());
		// // System.out.println("reachable size: " + rrrr);
		// }
		// if (noCommi) {
		// if (time < 2)
		// return;
		// System.out.println("-------------TIME " + time + "-------------");
		// System.out.println("-------------ID: " + me().getID()
		// + "-------------");
		// System.out
		// .println("-------------Channel: " + harchiNo[time - 1][0]);
		// System.out.println("Way: " + harchiNo[time - 1][1] + " NoWay: "
		// + harchiNo[time - 1][2] + " CV: " + harchiNo[time - 1][3]
		// + " Fired: " + harchiNo[time - 1][4] + " OutFired: "
		// + harchiNo[time - 1][5] + " Clearer: "
		// + harchiNo[time - 1][12] + " AC: " + harchiNo[time - 1][13]
		// + " PC: " + harchiNo[time - 1][14] + " FC: "
		// + harchiNo[time - 1][15] + " Info: "
		// + harchiNo[time - 1][19] + " Pos: "
		// + harchiNo[time - 1][21] + " Zone: "
		// + harchiNo[time - 1][23] + " Stuck: "
		// + harchiNo[time - 1][26] + " UStuck: "
		// + harchiNo[time - 1][27] + " Help: "
		// + harchiNo[time - 1][28] + " UHelp: "
		// + harchiNo[time - 1][29]);
		// System.out.println("-------------Other Channel: "
		// + harchiNo[time - 1][6]);
		// System.out.println("Way: " + harchiNo[time - 1][7] + " NoWay: "
		// + harchiNo[time - 1][8] + " CV: " + harchiNo[time - 1][9]
		// + " Fired: " + harchiNo[time - 1][10] + " OutFired: "
		// + harchiNo[time - 1][11] + " Clearer: "
		// + harchiNo[time - 1][25] + " AC: " + harchiNo[time - 1][16]
		// + " PC: " + harchiNo[time - 1][17] + " FC: "
		// + harchiNo[time - 1][18] + " Info: "
		// + harchiNo[time - 1][20] + " Pos: "
		// + harchiNo[time - 1][22] + " Zone: "
		// + harchiNo[time - 1][24] + " Stuck: "
		// + harchiNo[time - 1][30] + " UStuck: "
		// + harchiNo[time - 1][31] + " Help: "
		// + harchiNo[time - 1][32] + " UHelp: "
		// + harchiNo[time - 1][33]);
		// System.out.println("last CV packs: " + lastPacksSize[time - 1]);
		// System.out.println("notInNearsSize: " + notInNearsSize[time - 1]);
		// System.out.println("lastBuildsToSaySize: "
		// + lastBuildsToSaySize[time - 1]);
		// System.out.println("-------------------------------------------");
		//
		// } else {

		// if (time > 1
		// && ((worldModel.ambulanceTeams.size() > 0 && me().getID()
		// .getValue() == worldModel.ambulanceTeams.get(0)
		// .getID().getValue())
		// || (worldModel.policeForces.size() > 0 && me()
		// .getID().getValue() == worldModel.policeForces
		// .get(0).getID().getValue()) || (worldModel.fireBrigades
		// .size() > 0 && me().getID().getValue() == worldModel.fireBrigades
		// .get(0).getID().getValue()))) {
		// System.out.println("-------------TIME " + time
		// + "-------------");
		// System.out.println("-------------Channel: "
		// + harchi[time - 1][0]);
		// System.out.println("Way: " + harchi[time - 1][1] + " NoWay: "
		// + harchi[time - 1][2] + " CV: " + harchi[time - 1][3]
		// + " Fired: " + harchi[time - 1][4] + " OutFired: "
		// + harchi[time - 1][5] + " Clearer: "
		// + harchi[time - 1][12] + " AC: " + harchi[time - 1][13]
		// + " PC: " + harchi[time - 1][14] + " FC: "
		// + harchi[time - 1][15] + " Info: "
		// + harchi[time - 1][19] + " Pos: "
		// + harchi[time - 1][21] + " Zone: "
		// + harchi[time - 1][23] + " Stuck: "
		// + harchi[time - 1][26] + " UStuck: "
		// + harchi[time - 1][27] + " Help: "
		// + harchi[time - 1][28] + " UHelp: "
		// + harchi[time - 1][29]);
		// System.out.println("-------------Other Channel: "
		// + harchi[time - 1][6]);
		// System.out.println("Way: " + harchi[time - 1][7] + " NoWay: "
		// + harchi[time - 1][8] + " CV: " + harchi[time - 1][9]
		// + " Fired: " + harchi[time - 1][10] + " OutFired: "
		// + harchi[time - 1][11] + " Clearer: "
		// + harchi[time - 1][25] + " AC: " + harchi[time - 1][16]
		// + " PC: " + harchi[time - 1][17] + " FC: "
		// + harchi[time - 1][18] + " Info: "
		// + harchi[time - 1][20] + " Pos: "
		// + harchi[time - 1][22] + " Zone: "
		// + harchi[time - 1][24] + " Stuck: "
		// + harchi[time - 1][30] + " UStuck: "
		// + harchi[time - 1][31] + " Help: "
		// + harchi[time - 1][32] + " UHelp: "
		// + harchi[time - 1][33]);
		// System.out.println("last CV packs: " + lastPacksSize[time - 1]);
		// System.out.println("notInNearsSize: "
		// + notInNearsSize[time - 1]);
		// System.out.println("lastBuildsToSaySize: "
		// + lastBuildsToSaySize[time - 1]);
		// System.out.println("CVs Said: " + CVsSended);
		// System.out.println("-------------------------------------------");
		// // System.out.println("outFiredCheck: " + outFiredCheck[time - 1]);
		// CVsSended = new ArrayList<Integer>();
		// }
		// }
	}

	@SuppressWarnings("unused")
	private void checkOutForRadChange() {
		if (me() instanceof PoliceForce) {
			worldModel.getTrafficSimulator().agentRad = 500;
			return;
		}
		worldModel.getTrafficSimulator().agentRad = 100;
		if (((RadarAgent<?>) this).isUnStuckTrueOnce) {
			if (((ThinkAgent<?>) this).stuckToolKeshid > 7) {
				worldModel.getTrafficSimulator().agentRad = 375;
				// System.out.println("Changed my rad to 375 in time " + time
				// + " id " + me().getID().getValue());
			}
			if (((ThinkAgent<?>) this).stuckToolKeshid > 14) {
				worldModel.getTrafficSimulator().agentRad = 500;
				// System.out.println("Changed my rad to 500 in time " + time
				// + " id " + me().getID().getValue());
			}
		}
	}

	protected void think(int time, ChangeSet changes, Collection<Command> heard) {

		// printRadarDetails();
		try {
			cycleTimer.setTime(0);
			this.time = time;
			this.heard = heard;
			logger.logTime(time);
			// checkOutForRadChange();
			if (time % timeForNewWayMemory == 0)
				if (worldModel.map != MapType.NY)
					updatingWayMemory();
				else
					System.out
							.println("didn't update way memories --> NY! time "
									+ time);
			try {
				worldModel.update(changes);

			} catch (ActionCommandException ac) {
				throw ac;
			}
			// update buriedness
			if (myHum() != null)
				myHum().hasBuriedness = (myHum().isBuriednessDefined() && myHum()
						.getBuriedness() > 0);
			setSTFire();
			for (worldModel.Building b : worldModel.buildings)
				b.shouldSend = false;
			// update LTS
			for (EntityID changed : worldModel.changeSet.getChangedEntities()) {
				Entity e = worldModel.getEntity(changed);
				if (e instanceof rescuecore2.standard.entities.Area) {
					((rescuecore2.standard.entities.Area) e).area.LTS = time;
					if (e instanceof rescuecore2.standard.entities.Building) {
						worldModel.Building build = (worldModel.Building) worldModel
								.getArea(changed);
						build.isSimulatedForFire = false;
						hearRF(build, calculatedRF(build));
					}
				} else if (e instanceof Human) {
					Human h = (Human) e;
					if (!(e instanceof Civilian)) {
						if (time > 4 && time - h.lastPosUpdatedTime < 10) {
							updateWays(h.lastCyclePos, worldModel.getArea(h
									.getPosition()), h.getID().getValue());
						}
						h.lastPosUpdatedTime = time;
						worldModel.getArea(h.getPosition()).HLTV = time;
					} else if (worldModel.getEntity(h.getPosition()) instanceof Area)// TODO
																						// Is
																						// that
																						// right?
						worldModel.getArea(h.getPosition()).HLTV = time;
					// update burriedness

					h.hasBuriedness = (h.isBuriednessDefined() && h
							.getBuriedness() > 0);
					h.setBury(h.getBuriedness());
					h.setHp(h.getHP(), time);
					h.setDmg(h.getDamage(), time);
					ArrayList<worldModel.Building> ignitedBuildings = new ArrayList<worldModel.Building>();
					for (worldModel.Building b : worldModel.buildings)
						if (b.isOnFire())
							ignitedBuildings.add(b);
					boolean emergency = false;
					for (worldModel.Building b : ignitedBuildings)
						if (worldModel.getEntity(h.getPosition()) instanceof Area
								&& worldModel.getArea(h.getPosition())
										.getNeighbours().contains(b.getID()))
							emergency = true;
					h.setDT(time, ignitedBuildings, worldModel, emergency);
					if (h.deadtime > 502)
						h.deadtime = 502;
				}
			}
			// update LTV
			if (lastMovePath.contains(getMyPos().id)) {
				for (EntityID eID : lastMovePath) {
					if (eID.getValue() == getMyPos().id)
						break;

					worldModel.getArea(eID).LTV = time - 1;
				}
			} else {
				boolean find = false;
				boolean arive = false;
				Area MyPlaceinPath = null;
				ArrayList<Area> MyPlacesinPath = new ArrayList<Area>();
				for (Entrance entr : getMyPos().entrances) {
					if (lastMovePath.contains(entr.neighbour.area))
						MyPlacesinPath.add(entr.neighbour.area);
				}
				for (EntityID eID : lastMovePath) {
					if (MyPlacesinPath.contains(worldModel.getArea(eID))) {
						if (find == false) {
							find = true;
							MyPlaceinPath = worldModel.getArea(eID);
						}
					}
				}
				if (MyPlaceinPath != null) {
					for (EntityID eID : lastMovePath) {
						if (eID.getValue() == MyPlaceinPath.id)
							arive = true;
						worldModel.getArea(eID).LTV = time - 1;
						if (arive == true)
							break;
					}
				}
			}
			getMyPos().LTV = time;
			try {
				ActionCommandException ace = null;
				try {
					process();
				} catch (ActionCommandException ac) {
					ace = ac;
					if (me() instanceof AmbulanceTeam && !noCommi)
						((AmbulanceTeamAgent) this).sendInfo();
				}
				if (ace == null)
					rest();
				throw ace;
			} catch (ActionCommandException ace) {
				setLastPoses();
				commandHistory.put(time, ace);
			}
		} catch (Exception e) {
			System.err.println("Exception: " + me().getID().getValue()
					+ " Time: " + time);
			e.printStackTrace();
			if (logger.getWriter() != null)
				e.printStackTrace(logger.getWriter());
			sendRest(time);
			commandHistory.put(time, new ActionCommandException(
					StandardMessageURN.AK_REST));
		}
		logger.logEndOfCycle();
	}

	private void setLastPoses() {
		for (Human h : worldModel.agents)
			h.lastCyclePos = worldModel.getArea(h.getPosition());
	}

	public enum MapType {
		VC, Paris, Kobe, Berlin, Istanbul, Eindhoven, Mecxico, Unknown, Joao, Test, NY
	}

	Refuge nearestRefuge = null;
	boolean shouldIChangeNearestRefuge = true;

	public void findNearestRefuge(Human man) {
		int minDis = Integer.MAX_VALUE;
		if (!(man instanceof PoliceForce)
				|| worldModel.reachableRefuges.size() > 0) {
			for (Refuge ref : worldModel.reachableRefuges)
				if (minDis > worldModel.getArea(ref.getID()).distanceFromSelf) {
					minDis = worldModel.getArea(ref.getID()).distanceFromSelf;
					nearestRefuge = ref;
				}
		} else if (me() instanceof PoliceForce) {
			for (Refuge ref : worldModel.refuges)
				if (minDis > worldModel.getArea(ref.getID()).distFromeSelfWOB) {
					minDis = worldModel.getArea(ref.getID()).distFromeSelfWOB;
					nearestRefuge = ref;
				}
		}
	}

	int zaman = 0;

	public void mikhadBemire() throws ActionCommandException {
		Human man = (Human) me();
		if (me() instanceof PoliceForce
				&& worldModel.reachableRefuges.size() == 0) {
			zaman = 35;
		} else {
			zaman = 25;
		}

		if (man.hasBuriedness)
			return;
		if (shouldIChangeNearestRefuge)
			findNearestRefuge(man);
		if (movingToRefuge
				&& man.isDamageDefined()
				&& man.getDamage() != 0
				&& (worldModel.reachableRefuges.size() > 0 || (me() instanceof PoliceForce && worldModel.refuges
						.size() > 0))
				&& man.getPosition().getValue() == nearestRefuge.getID()
						.getValue()) {
			rest();
		}

		if (movingToRefuge && worldModel.reachableRefuges.size() > 0) {
			EntityID entID = nearestRefuge.getID();
			if (((Human) me()).getPosition().getValue() != entID.getValue()) {
				if (me() instanceof PoliceForce
						&& !worldModel.reachableRefuges.contains(nearestRefuge)) {
					((PoliceForceAgent) this).fastClearPath(nearestRefuge,
							nearestRefuge.getX(), nearestRefuge.getY());
				} else {
					move(entID);
				}
			} else if (man.isDamageDefined() && man.getDamage() == 0) {
				shouldIChangeNearestRefuge = true;
				movingToRefuge = false;
			}
		}
		ArrayList<EntityID> dist = new ArrayList<EntityID>();
		if (man.isDamageDefined() && man.getDamage() > 0) {
			if (worldModel.reachableRefuges.size() > 0) {
				dist = worldModel.getMinPathDijk(nearestRefuge.getID(), true,
						false, worldModel, myHum());
				if (man.deadtime < dist.size() + zaman + time) {
					movingToRefuge = true;
					shouldIChangeNearestRefuge = false;
					manLastDeadTime = man.deadtime;
					move(dist, worldModel.getArea(dist.get(dist.size() - 1))
							.getX(),
							worldModel.getArea(dist.get(dist.size() - 1))
									.getY(), true);
				}
			} else if (me() instanceof PoliceForce
					&& worldModel.refuges.size() > 0) {
				ArrayList<EntityID> myPathToRefuge = worldModel.getMinPathDijk(
						nearestRefuge.getID(), false, false, worldModel,
						myHum());
				if (man.deadtime < myPathToRefuge.size() + zaman + time) {
					movingToRefuge = true;
					shouldIChangeNearestRefuge = false;
					((PoliceForceAgent) this).fastClearPath(
							worldModel.getArea(nearestRefuge.getID()),
							worldModel.getArea(nearestRefuge.getID()).getX(),
							worldModel.getArea(nearestRefuge.getID()).getY());
				}
			}
		}
	}

	public ArrayList<worldModel.Blockade> getBlocksInRange(int range) {
		ArrayList<worldModel.Blockade> blockList = new ArrayList<worldModel.Blockade>();
		Coordinate rightUp = new Coordinate(this.getX() + range, this.getY()
				+ range);
		Coordinate rightDown = new Coordinate(this.getX() + range, this.getY()
				- range);
		Coordinate leftUp = new Coordinate(this.getX() - range, this.getY()
				+ range);
		Coordinate leftDown = new Coordinate(this.getX() - range, this.getY()
				- range);
		Coordinate[] coordinates = { leftDown, leftUp, rightUp, rightDown };
		ArrayList<Coordinate> coords = new ArrayList<Coordinate>();
		coords.add(rightUp);
		coords.add(rightDown);
		coords.add(leftDown);
		coords.add(leftUp);
		coords.add(rightUp);
		blockList.addAll(worldModel.getTrafficSimulator().getBlockadesInRange(
				GeometryTools.makeRectangle(coordinates)));
		return blockList;
	}

	protected Human myHum() {
		if (me() instanceof Human)
			return (Human) me();
		return null;
	}

	EntityID lastMovePathTarget;
	EntityID lastRealMovePathTarget;

	public static ArrayList<Coordinate> getVisionPoints(WorldModel model,
			Area area, worldModel.Building building, int maxDistance,
			int minMax, GeometryFactory geom) {

		Coordinate builCenter = new Coordinate((int) building.getX(),
				(int) building.getY());
		if (!building.circumstance.contains(GeometryTools.coordToGeometry(
				builCenter, geom))) {
			builCenter = new Coordinate((int) building.circumstance
					.getInteriorPoint().getX(), (int) building.circumstance
					.getInteriorPoint().getY());
		}
		maxDistance *= 4;
		double minDegree = Integer.MAX_VALUE, maxDegree = -Integer.MAX_VALUE;
		ArrayList<Coordinate> array = new ArrayList<Coordinate>();
		for (Coordinate c : area.circumstance.getCoordinates())
			array.add(c);
		ArrayList<Coordinate> roadPoints = array;
		ArrayList<Coordinate> visionPoints = new ArrayList<Coordinate>();
		Coordinate buildingPoint = new Coordinate(builCenter.x, builCenter.y);
		Coordinate minPoint = null, maxPoint = null;
		for (int i = 0; i < roadPoints.size() - 1; i++) {
			Segment line = new Segment(buildingPoint, roadPoints.get(i));
			double degree = Angle.normalizePositive(line.getLineSegment()
					.angle());

			if (i == 0) {
				minDegree = maxDegree = degree;
				minPoint = maxPoint = roadPoints.get(i);
			} else {
				if (GeometryTools.isClockWise(roadPoints.get(i - 1),
						buildingPoint, roadPoints.get(i))) {
					if (GeometryTools.isClockWise(minPoint, buildingPoint,
							roadPoints.get(i))) {
						minDegree = degree;
						minPoint = roadPoints.get(i);
					}
				} else {
					if (!GeometryTools.isClockWise(maxPoint, buildingPoint,
							roadPoints.get(i))) {
						maxDegree = degree;
						maxPoint = roadPoints.get(i);
					}
				}
			}
		}
		int m = 3;
		if (area.circumstance.getCoordinates().length < 7)
			m = 1;
		ArrayList<worldModel.Building> nearBuildings = new ArrayList<worldModel.Building>();

		for (Area a : building.nearAreas50000)
			if (a instanceof worldModel.Building)
				nearBuildings.add((worldModel.Building) a);
		double deg1 = Angle.normalizePositive((minDegree + maxDegree) / 2);
		for (int j = 0; j < m; j++) {
			if (j == 1)
				deg1 = Angle
						.normalizePositive((float) ((minDegree + maxDegree) / 2 - (GeometryTools
								.getDeltaAngle((float) maxDegree,
										(float) minDegree) / 4)));
			if (j == 2)
				deg1 = Angle
						.normalizePositive((float) ((minDegree + maxDegree) / 2 + (GeometryTools
								.getDeltaAngle((float) maxDegree,
										(float) minDegree)) / 4));
			double deg2 = Angle.normalizePositive((float) Math.PI + deg1);
			double degree = deg1;
			Coordinate tmpPoint = new Coordinate(builCenter.x + 10
					* (int) Math.cos(deg2), builCenter.y + 10
					* (int) Math.sin(deg2));
			if (GeometryTools.isClockWise(tmpPoint, buildingPoint, minPoint)
					&& !GeometryTools.isClockWise(tmpPoint, buildingPoint,
							maxPoint))
				degree = deg2;
			int x = (int) (builCenter.x + maxDistance * Math.cos(degree));
			int y = (int) (builCenter.y + maxDistance * Math.sin(degree));
			Segment buildingToRoad = new Segment(buildingPoint, new Coordinate(
					x, y));
			float minDist = Integer.MAX_VALUE, maxDist = -Integer.MAX_VALUE;
			for (int i = 0; i < roadPoints.size() - 1; i++) {
				Segment line = new Segment(roadPoints.get(i),
						roadPoints.get(i + 1));
				Coordinate intersectPoint = buildingToRoad.intersection(line);
				if (intersectPoint != null) {
					float dist = (float) intersectPoint.distance(buildingPoint);
					if (dist < minDist)
						minDist = dist;
					if (dist > maxDist)
						maxDist = dist;
				}
			}
			if (maxDist < 0)
				continue;
			// if (minDist > minMax)// remove
			// continue;
			// maxDist = Math.min(maxDist, minMax);// remove
			for (int i = 0; i < 3; i++) {
				float dist = (minDist + maxDist) / 2;
				if (i == 1)
					dist = maxDist - 500; // TAVAJJOH
				if (i == 2)
					dist = minDist + 500; // TAVAJJOH
				int newX = (int) (builCenter.x + dist * Math.cos(degree));
				int newY = (int) (builCenter.y + dist * Math.sin(degree));
				Coordinate newPoint = new Coordinate(newX, newY);
				buildingToRoad.getEnd().setCoordinate(newPoint);
				float kamine = (float) buildingPoint.distance(newPoint);
				Coordinate behtarinPoint = buildingPoint;
				Coordinate[] coords = building.circumstance.getCoordinates();
				for (int jj = 0; jj < coords.length - 1; jj++) {
					Segment lineEdge = new Segment(coords[jj], coords[jj + 1]);
					if (GeometryTools.isIntersectBetweenLines(buildingToRoad,
							lineEdge)) {
						Coordinate intersectPoint = buildingToRoad
								.intersection(lineEdge);
						if (intersectPoint != null) {
							if (intersectPoint.distance(newPoint) < kamine) {
								kamine = (float) intersectPoint
										.distance(newPoint);
								behtarinPoint = intersectPoint;
							}
						}
					}
				}
				behtarinPoint.x = (int) behtarinPoint.x;
				behtarinPoint.y = (int) behtarinPoint.y;
				while (!building.circumstance.contains(GeometryTools
						.coordToGeometry(behtarinPoint, geom))) {
					Coordinate c = new Coordinate();
					c.x = (int) behtarinPoint.x - buildingPoint.x;
					c.y = (int) behtarinPoint.y - buildingPoint.y;
					c.x = (int) (c.x * 0.99);
					c.y = (int) (c.y * 0.99);
					behtarinPoint.x = (int) (c.x + buildingPoint.x);
					behtarinPoint.y = (int) (c.y + buildingPoint.y);
				}
				buildingToRoad = new Segment(behtarinPoint,
						buildingToRoad.getEnd());
				boolean hasIntersect = false;
				done: for (worldModel.Building bld : nearBuildings) {
					for (Edge edge : bld.serverArea.getEdges()) {
						if (edge.getNeighbour() == null) {
							Segment lineEdge = new Segment(new Coordinate(
									edge.getStartX(), edge.getStartY()),
									new Coordinate(edge.getEndX(), edge
											.getEndY()));
							if (GeometryTools.isIntersectBetweenLines(
									buildingToRoad, lineEdge)) {
								hasIntersect = true;
								break done;
							}
						}
					}
				}
				if (behtarinPoint.distance(newPoint) < model.maxDistanceView
						&& newPoint.distance(buildingPoint) < maxDistance
						&& !hasIntersect
						&& Mathematic.getDistance(
								(int) buildingToRoad.getEnd().x,
								(int) buildingToRoad.getEnd().y,
								(int) builCenter.x, (int) builCenter.y) < model.maxDistanceExtinguish) {
					visionPoints.add((Coordinate) buildingToRoad.getEnd()
							.clone());
				}
			}
		}
		return visionPoints;
	}

	public void visionRoads() {
		for (worldModel.Building b : worldModel.buildings)
			b.getVisionRoads(worldModel);
	}

	public void visionBuildings() {
		for (worldModel.Building building : worldModel.buildings) {
			for (Edge edge : building.serverArea.getEdges()) {
				if (edge.getNeighbour() == null)
					continue;
				Entity entity = worldModel.getEntity(edge.getNeighbour());
				if (entity instanceof Building) {
					worldModel.Building area = (worldModel.Building) worldModel
							.getArea(entity.getID());
					Coordinate c = new Coordinate(
							(edge.getStartX() + edge.getEndX()) / 2,
							(edge.getStartY() + edge.getEndY()) / 2);
					done: for (int i = 5000; i >= 5000; i -= 5000) {
						for (float theta = 0; theta < 2 * Math.PI; theta += 2 * Math.PI / 20.0) {
							Coordinate p = new Coordinate(
									(int) (i * Math.cos(theta)),
									(int) (i * Math.sin(theta)));
							p = new Coordinate(c.x + p.x, c.y + p.y);
							if (area.shape
									.contains(GeometryTools.coordToGeometry(
											p,
											worldModel.getTrafficSimulator().geom))) {
								building.visionBuildings.put(area, p);
								area.sumlooked++;
								break done;
							}
						}
					}
				}
			}
		}
	}

	protected void rest() throws ActionCommandException {
		sendRest(time);
		throw new ActionCommandException(StandardMessageURN.AK_REST);
	}

	public int Timeofstartext = -1, Timeoflastex = -1;
	Area lastaim = null;

	protected void extinguish(EntityID target, int power)
			throws ActionCommandException {
		extinguishedTarg = target;
		extinguishTime = time;
		power = Math.min(power, ((FireBrigade) myHum()).getWater());
		sendExtinguish(time, target, power);
		throw new ActionCommandException(StandardMessageURN.AK_EXTINGUISH);
	}

	int xLastMove = -1, yLastMove = -1;
	int xLastCycle = -1, yLastCycle = -1;
	int xRealMove = -1, yRealMove = -1;
	int lastMoveCycle;
	public int stuckToolKeshid = 0;
	ArrayList<EntityID> lastMovePath = new ArrayList<EntityID>();
	ArrayList<EntityID> lastRealMovePath = new ArrayList<EntityID>();

	protected void move(List<EntityID> path, int x, int y, boolean djkstraPath)
			throws ActionCommandException {
		ArrayList<EntityID> pd = dijkstraCalculations(djkstraPath, path, true);
		if (pd != null)
			path = pd;
		xLastMove = x;
		yLastMove = y;
		xRealMove = x;
		yRealMove = y;
		lastMoveCycle = time;

		lastMovePathTarget = path.get(path.size() - 1);
		// lastMovePath = (ArrayList<EntityID>) path;
		lastMovePath.clear();
		lastMovePath.addAll(path);

		lastRealMovePathTarget = path.get(path.size() - 1);
		// lastRealMovePath = (ArrayList<EntityID>) path;
		lastRealMovePath.clear();
		lastRealMovePath.addAll(path);

		sendMove(time, path, x, y);
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	protected void clear(EntityID target) throws ActionCommandException {
		sendClear(time, target);
		ActionCommandException ace = new ActionCommandException(
				StandardMessageURN.AK_CLEAR);
		ace.lastClearedTarget = ((rescuecore2.standard.entities.Blockade) worldModel
				.getEntity(target)).getPosition();
		throw ace;
	}

	protected void rescue(EntityID target) throws ActionCommandException {
		sendRescue(time, target);
		throw new ActionCommandException(StandardMessageURN.AK_RESCUE);
	}

	protected void load(EntityID target) throws ActionCommandException {
		sendLoad(time, target);
		throw new ActionCommandException(StandardMessageURN.AK_LOAD);
	}

	protected void unLoad() throws ActionCommandException {
		sendUnload(time);
		throw new ActionCommandException(StandardMessageURN.AK_UNLOAD);
	}

	public void log(String msg) {
		// logger.log(msg);
	}

	public int getTime() {
		return time;
	}

	public StandardWorldModel getModel() {
		return model;
	}

	public WorldModel getWorldModel() {
		return worldModel;
	}

	public int getX() {
		if (isHuman)
			return ((Human) me()).getX();
		return ((rescuecore2.standard.entities.Building) me()).getX();
	}

	public int getY() {
		if (isHuman)
			return ((Human) me()).getY();
		return ((rescuecore2.standard.entities.Building) me()).getY();
	}

	public Area getMyPos() {
		if (isHuman)
			return worldModel.getArea(((Human) me()).getPosition());

		return ((rescuecore2.standard.entities.Building) me()).area;
	}

	public boolean isLastMoveBFS = true;
	public int lastTargetForDijk = -1;

	public ArrayList<EntityID> dijkstraCalculations(boolean djkstraPath,
			List<EntityID> path) {
		return dijkstraCalculations(djkstraPath, path, false);
	}

	public ArrayList<EntityID> dijkstraCalculations(boolean djkstraPath,
			List<EntityID> path, boolean isCalledByMove) {
		if (!djkstraPath) {
			try {
				if (path.size() > 1) {
					ArrayList<EntityID> epath = null;
					Area a = worldModel.getArea(path.get(path.size() - 1));
					if (isCalledByMove
							&& a.getID().getValue() != lastTargetForDijk)
						isLastMoveBFS = true;
					lastTargetForDijk = a.getID().getValue();
					int thres = 40;
					if (!isLastMoveBFS
							&& a.getID().getValue() == lastTargetForDijk)
						thres = 120;
					if (path.size() <= thres) {
						if (isCalledByMove)
							isLastMoveBFS = false;
						if (me() instanceof PoliceForce)
							epath = worldModel.getMinPathDijk(a.getID(), false,
									true, worldModel, myHum());
						else if (a.isReachable) {
							// System.out.println("Time:" + time);
							epath = worldModel.getMinPathDijk(a.getID(), true,
									true, worldModel, myHum());
						}
						if (epath != null && epath.size() > 0)
							path = epath;
					} else {
						if (isCalledByMove)
							isLastMoveBFS = true;
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
				if (logger.getWriter() != null) {
					e.printStackTrace(logger.getWriter());
				}
			}
		}
		return (ArrayList<EntityID>) path;
	}

	protected boolean isReachable(EntityID target) {
		return worldModel.getArea(target).isReachable;
	}

	public void move(Area target) throws ActionCommandException {
		move(target, target.getX(), target.getY(), false);
		// move(target, -1, -1, true);
	}

	public void moveForPolice(Area target, int x, int y, boolean djkstraPath)
			throws ActionCommandException {
		ArrayList<EntityID> pathId = worldModel.getMinPathDijk(target.getID(),
				false, true, worldModel, myHum());
		// log("path: " + pathId);
		if (pathId == null || pathId.size() == 0) {
			pathId = new ArrayList<EntityID>();
			pathId.add(target.getID());
		}
		move(pathId, x, y, djkstraPath);
	}

	public void moveForPolice(Area target, int x, int y, boolean djkstraPath,
			ArrayList<EntityID> pathID) throws ActionCommandException {

		// log("path: " + pathID);
		if (pathID == null || pathID.size() == 0) {
			pathID = new ArrayList<EntityID>();
			pathID.add(target.getID());
		}

		move(pathID, x, y, djkstraPath);
	}

	public void move(Area target, int x, int y, boolean djkstraPath)
			throws ActionCommandException {
		ArrayList<Area> path = worldModel.getMinPath(worldModel.myEntrances,
				target);
		ArrayList<EntityID> pathId = new ArrayList<EntityID>();
		if (path == null || path.size() == 0) {
			path = new ArrayList<Area>();
			path.add(target);
		}

		for (Area area : path)
			pathId.add(area.getID());
		move(pathId, x, y, djkstraPath);
	}

	public void move(EntityID e) throws ActionCommandException {
		move(worldModel.getArea(e));
	}

	private boolean hasGoodShapeAndSituation(Geometry point) {
		HashSet<Geometry> availableShapes = new HashSet<>();
		for (Entrance e : worldModel.myEntrances)
			if (e.shapeInArea != null)
				availableShapes.add(e.shapeInArea);

		for (Geometry g : availableShapes)
			if (g.contains(point))
				return true;

		return false;
	}

	static final double PECENTAGE_OF_MAX_DISTANCE = 0.9;

	public ArrayList<Area> unEmploymentSituationPath(worldModel.Building target) {
		worldModel.clearAreas();
		HashSet<Entrance> targets = new HashSet<Entrance>();

		for (worldModel.Area area : target.nearAreas50000)
			if ((area instanceof worldModel.Road || (area instanceof worldModel.Building && ((worldModel.Building) area).stFire == 0))
					&& Math.hypot(area.getX() - target.getX(), area.getY()
							- target.getY()) < maxDistanceExtinguish
							* PECENTAGE_OF_MAX_DISTANCE && area.isReachable) {
				targets.addAll(area.entrances);
			}
		return worldModel.getMinPathByHashSet(worldModel.myEntrances, targets,
				true);
	}

	public void checkandExtinguish(worldModel.Building building)
			throws ActionCommandException {
		if (building != null)
			if (((FireBrigadeAgent) this).canExtinguish(building) && !IamOnfire) {
				extinguish(building.getID(), ((FireBrigadeAgent) this).maxPower);
			} else {
				return;
			}
	}

	// private void unEmploymentMoveDecision(ArrayList<EntityID> moveToFirePath,
	// worldModel.Building mainTarget, int x, int y, boolean dijkstra,
	// boolean isInSearch) throws ActionCommandException {
	// ArrayList<Area> unEmploymentSituationPath = new ArrayList<Area>();
	// unEmploymentSituationPath.addAll(unEmploymentSituationPath(mainTarget));
	// log("issimulatedForFire: " + mainTarget.isSimulatedForFire
	// + " MainTarg: " + mainTarget + " isinSearch: " + isInSearch
	// + "path.size: " + unEmploymentSituationPath.size()
	// + "movetofirepath: " + moveToFirePath.size() + "canExtinguish:"
	// + ((FireBrigadeAgent) this).canExtinguish(mainTarget)
	// + "isOnFire: " + mainTarget.isOnFire());
	// if (unEmploymentSituationPath != null
	// && (unEmploymentSituationPath.size() != 0 || ((FireBrigadeAgent)
	// this)
	// .canExtinguish(mainTarget))
	// && unEmploymentSituationPath.size() / moveToFirePath.size() < 0.8
	// && !isInSearch && !mainTarget.isSimulatedForFire
	// && mainTarget.isOnFire() && mainTarget.getLTS() + 10 >= time) {
	// checkUnemploymentExtinguish(false);
	// log("unemploymentSituationPath: " + unEmploymentSituationPath);
	// move(areaToEntityID(unEmploymentSituationPath), mainTarget.getX(),
	// mainTarget.getY(), false);
	// } else
	// move(moveToFirePath, x, y, dijkstra);
	// }
	//
	public ArrayList<EntityID> areaToEntityID(ArrayList<Area> path) {
		if (path == null)
			return null;
		ArrayList<EntityID> pathID = new ArrayList<EntityID>();
		for (Area a : path)
			pathID.add(a.getID());
		return pathID;
	}

	@SuppressWarnings("unchecked")
	public boolean hasBeenRecentlySeen(worldModel.Building target) {
		return target.getLTS() + 10 >= time
				|| ((RadarAgent<StandardEntity>) this).trapped.keySet()
						.contains(lastSeenBy.get(target))
		/*
		 * || worldModel.getEntityByInt(lastSeenBy.get(target)) instanceof
		 * Building
		 */;
	}

	public void moveToFire(worldModel.Building target)
			throws ActionCommandException {
		// if (unEmploymentSituationPath != null
		// && unEmploymentSituationPath.size() == 0
		// && ((FireBrigadeAgent) this).canExtinguish(target)
		// && target.isOnFire() && hasBeenRecentlySeen(target))
		// checkUnemploymentExtinguish(false); extinguish(target.getID(),
		// ((FireBrigadeAgent) this).maxPower);

		ArrayList<EntityID> moveToFirePath = new ArrayList<EntityID>();
		AKMove moveToFireMove = moveToFireComputations(target);
		if (moveToFireMove != null) {
			moveToFirePath.addAll(moveToFireMove.getPath());
			if (moveToFirePath.size() <= 10)
				moveToFireAndSee(target);
		}
		ArrayList<EntityID> unEmploymentSituationPath = new ArrayList<EntityID>();
		unEmploymentSituationPath = areaToEntityID(unEmploymentSituationPath(target));
		if (unEmploymentSituationPath != null
				&& (unEmploymentSituationPath.size() != 0 || ((FireBrigadeAgent) this)
						.canExtinguish(target))
				&& (moveToFirePath.size() == 0 || (unEmploymentSituationPath
						.size() / moveToFirePath.size() < 0.8))
				&& hasBeenRecentlySeen(target) && target.isOnFire()) {
			// log("unemployment : " + unEmploymentSituationPath);
			if (((FireBrigadeAgent) this).canExtinguish(target)) {
				extinguish(target.getID(), ((FireBrigadeAgent) this).maxPower);
			}
			Area unEmploymentTarget = worldModel
					.getArea(unEmploymentSituationPath
							.get(unEmploymentSituationPath.size() - 1));
			move(unEmploymentSituationPath, unEmploymentTarget.getX(),
					unEmploymentTarget.getY(), false);
		} else if (moveToFireMove != null && moveToFirePath.size() != 0) {
			// log("moveToFirePath : " + moveToFirePath);
			if (worldModel.isInChangeSet(target.getID())
					&& ((FireBrigadeAgent) this).canExtinguish(target)
					&& !IamOnfire) {
				extinguish(target.getID(), ((FireBrigadeAgent) this).maxPower);
			}
			moveToFireAndSee(target);
		}
	}

	public void moveToFireAndSee(worldModel.Building target)
			throws ActionCommandException {
		AKMove moveToFireMoveCommand = moveToFireComputations(target);
		if (moveToFireMoveCommand == null) {
			// Exception melException = new Exception(
			// "to move to fire and see darim random walk mizanim !!!!!!!");
			// melException.printStackTrace();
			if (me() instanceof FireBrigade)
				((FireBrigadeAgent) this).randomWalk(false);
			if (me() instanceof AmbulanceTeam)
				((AmbulanceTeamAgent) this).randomWalk(false);
		}
		ArrayList<EntityID> moveToFirePath = new ArrayList<EntityID>();
		moveToFirePath.addAll(moveToFireMoveCommand.getPath());
		int x = moveToFireMoveCommand.getDestinationX();
		int y = moveToFireMoveCommand.getDestinationY();
		move(moveToFirePath, x, y, false);
	}

	public void initMoveToFireComputation() {
		for (Entrance e : worldModel.allEntrancesOfAreas) {
			e.isTarget = false;
			e.distForDijk = 1000000000;
			e.prev = null;
			e.isTarget = false;
		}
	}

	public AKMove moveToFireComputations(worldModel.Building target)
			throws ActionCommandException {
		initMoveToFireComputation();
		for (Road road : target.getVisionRoads(worldModel).keySet()) {
			if (road.isReachable && road.canBeVisited)
				for (int j = 0; j < target.getVisionRoads(worldModel).get(road)
						.size(); j++) {
					boolean i = true;
					Geometry bb = GeometryTools.coordToGeometry(target
							.getVisionRoads(worldModel).get(road).get(j),
							worldModel.getTrafficSimulator().geom);
					for (Blockade blockade : road.blockades) {
						if (blockade.circumstance.contains(bb)) {
							i = false;
							break;
						}
					}
					double dist = Math.hypot(
							getX()
									- target.getVisionRoads(worldModel)
											.get(road).get(j).x,
							getY()
									- target.getVisionRoads(worldModel)
											.get(road).get(j).y);

					if (dist < worldModel.getTrafficSimulator().agentRad
							&& !worldModel.isInChangeSet(target.getID())) {
						target.getVisionRoads(worldModel).get(road).remove(j);
						if (target.getVisionRoads(worldModel).get(road).size() == 0)
							target.getVisionRoads(worldModel).remove(road);
						target.buildingReachabilityCalculated = false;
						if (me() instanceof FireBrigade)
							((FireBrigadeAgent) this).act();
						if (me() instanceof AmbulanceTeam)
							((AmbulanceTeamAgent) this).act();
					}
					if (i == true) {
						if (getMyPos().getID().getValue() == road.getID()
								.getValue() && hasGoodShapeAndSituation(bb)) {
							ArrayList<EntityID> path = new ArrayList<EntityID>();
							path.add(road.getID());
							// unEmploymentMoveDecision(path, target, (int)
							// target
							// .getVisionRoads(worldModel).get(road)
							// .get(j).x,
							// (int) target.getVisionRoads(worldModel)
							// .get(road).get(j).y, false,
							// isInSearch);

							int destX = (int) target.getVisionRoads(worldModel)
									.get(road).get(j).x;
							int destY = (int) target.getVisionRoads(worldModel)
									.get(road).get(j).y;
							return (new AKMove(getID(), time, path, destX,
									destY));
						}

						for (Entrance e : road.entrances) {
							if (e.shapeInArea != null && e.isReachable
									&& !e.isTarget
									&& e.shapeInArea.contains(bb)) {
								e.isTarget = true;
								for (Entrance inter : e.internalEntrances)
									inter.isTarget = true;
							}
						}
					}
				}
		}

		for (worldModel.Building building : target.visionBuildings.keySet())
			if (building.isReachable && building.stFire == 0) {
				double dist = Math.hypot(
						getX() - target.visionBuildings.get(building).x, getY()
								- target.visionBuildings.get(building).y);
				if (dist < worldModel.getTrafficSimulator().agentRad
						&& !worldModel.isInChangeSet(target.getID())) {
					target.visionBuildings.remove(building);
					target.buildingReachabilityCalculated = false;
					if (me() instanceof FireBrigade)
						((FireBrigadeAgent) this).act();
					if (me() instanceof AmbulanceTeam)
						((AmbulanceTeamAgent) this).act();
				}
				if (getMyPos().getID().getValue() == building.getID()
						.getValue()) {
					ArrayList<EntityID> path = new ArrayList<EntityID>();
					path.add(building.getID());
					isInMoveToFire = true;
					// unEmploymentMoveDecision(path, target,
					// (int) (int) target.visionBuildings.get(building).x,
					// (int) target.visionBuildings.get(building).y,
					// false, isInSearch);
					int destX = (int) target.visionBuildings.get(building).x;
					int destY = (int) target.visionBuildings.get(building).y;
					return (new AKMove(getID(), time, path, destX, destY));
				}
				building.setAreaAsTarget(true);
			}

		int x = -1, y = -1;
		ArrayList<Area> getMinPathesh = new ArrayList<Area>();
		ArrayList<EntityID> minPath = new ArrayList<EntityID>();
		ArrayList<EntityID> path = new ArrayList<EntityID>();
		minPath = worldModel.getMinPathDijk(true, true, worldModel, myHum());
		if (minPath == null || minPath.size() == 0) {
			return null;
		}
		for (EntityID eID : minPath)
			getMinPathesh.add(worldModel.getArea(eID));

		for (Area a : getMinPathesh)
			path.add(a.getID());

		for (Road r : target.getVisionRoads(worldModel).keySet())
			if (path.get(path.size() - 1).getValue() == r.getID().getValue()) {
				for (int j = 0; j < target.getVisionRoads(worldModel).get(r)
						.size(); j++) {
					boolean i = true;
					Geometry l = null;
					l = GeometryTools.coordToGeometry(
							target.getVisionRoads(worldModel).get(r).get(j),
							worldModel.getTrafficSimulator().geom);

					for (Blockade blockade : r.blockades)
						if (blockade.circumstance.contains(l)) {
							i = false;
							break;
						}

					if (i == true) {
						if (r.getID().getValue() == getMyPos().id) {
							if (hasGoodShapeAndSituation(GeometryTools
									.coordToGeometry(
											new Coordinate(target
													.getVisionRoads(worldModel)
													.get(r).get(j).x, target
													.getVisionRoads(worldModel)
													.get(r).get(j).y),
											worldModel.getTrafficSimulator().geom))) {
								isInMoveToFire = true;
								// unEmploymentMoveDecision(path, target,
								// (int) target.getVisionRoads(worldModel)
								// .get(r).get(j).x, (int) target
								// .getVisionRoads(worldModel)
								// .get(r).get(j).y, r.getID()
								// .getValue() == getMyPos().id,
								// isInSearch);
								int destX = (int) target
										.getVisionRoads(worldModel).get(r)
										.get(j).x;
								int destY = (int) target
										.getVisionRoads(worldModel).get(r)
										.get(j).y;
								return new AKMove(getID(), time, path, destX,
										destY);
							}
						}

						else if (r.getID().getValue() != getMyPos().id)
							if (hasGoodShape(r,
									target.getVisionRoads(worldModel).get(r)
											.get(j))) {
								// unEmploymentMoveDecision(path, target,
								// (int) target.getVisionRoads(worldModel)
								// .get(r).get(j).x, (int) target
								// .getVisionRoads(worldModel)
								// .get(r).get(j).y, r.getID()
								// .getValue() == getMyPos().id,
								// isInSearch);
								int destX = (int) target
										.getVisionRoads(worldModel).get(r)
										.get(j).x;
								int destY = (int) target
										.getVisionRoads(worldModel).get(r)
										.get(j).y;
								return new AKMove(getID(), time, path, destX,
										destY);
							}

					}
				}
			}

		for (worldModel.Building b : target.visionBuildings.keySet())
			if (b.stFire == 0
					&& path.get(path.size() - 1).getValue() == b.getID()
							.getValue()) {
				x = (int) target.visionBuildings.get(b).x;
				y = (int) target.visionBuildings.get(b).y;
				if (getMyPos().getID() == path.get(path.size() - 1))
					isInMoveToFire = true;
				return new AKMove(getID(), time, path, x, y);
				// unEmploymentMoveDecision(path, target, x, y, false,
				// isInSearch);
			}
		return null;

	}

	protected boolean hasGoodShape(Road road, Coordinate point) {
		HashSet<Geometry> availableShapes = new HashSet<Geometry>();
		for (Entrance e : road.entrances)
			if (e.isReachable && e.shapeInArea != null) {
				availableShapes.add(e.shapeInArea);
			}
		for (Geometry shape : availableShapes) {
			if (shape.contains(GeometryTools.coordToGeometry(point,
					worldModel.getTrafficSimulator().geom))) {
				return true;
			}
		}
		return false;
	}

	protected void initMoveToFire() {
		for (Area area : worldModel.areas)
			for (Entrance ents : area.entrances) {
				ents.distForDijk = 1000000000;
				ents.prev = null;
				ents.isTarget = false;
			}
	}

	public boolean buildingReachability(worldModel.Building building) {
		// TODO buildingReachabilityCalculated hichvaght true nemishe!!!!
		if (building.buildingReachabilityCalculated) {
			return building.buildingReachability;
		}
		// log("ffffffffffffffffff "+building+" building.hasGoodVisionPoint(worldModel) "+building.hasGoodVisionPoint(worldModel));
		// log("stfire "+((worldModel.Building)worldModel.getAreaByInt(53964)).stFire);
		if (worldModel.isInChangeSet(building.getID())
				&& (me() instanceof AmbulanceTeam || ((FireBrigadeAgent) this)
						.canExtinguish(building)) && !IamOnfire
				&& building.hasGoodVisionPoint(worldModel)) {
			building.buildingReachability = true;
			return true;
		}
		if (building.getVisionRoads(worldModel).size() == 0
				&& building.visionBuildings.size() == 0) {
			return false;
		}
		if (building.isReachable && !IamOnfire
				&& building.hasGoodVisionPoint(worldModel)) {
			building.buildingReachability = true;
			return true;
		}
		for (Road road : building.getVisionRoads(worldModel).keySet())
			if (road.isReachable) {
				for (Coordinate point : building.getVisionRoads(worldModel)
						.get(road)) {
					if (hasGoodShape(road, point)) {
						boolean aa = true;
						for (worldModel.Blockade blockade : road.blockades)
							if (blockade.circumstance
									.contains(GeometryTools.coordToGeometry(
											point,
											worldModel.getTrafficSimulator().geom))) {
								aa = false;
								break;
							}
						if (aa == true) {
							building.buildingReachability = true;
							return true;
						}
					}
				}
			}
		for (worldModel.Building b : building.visionBuildings.keySet())
			if (b.isReachable && b.stFire == 0) {
				building.buildingReachability = true;
				return true;
			}
		building.buildingReachability = false;
		return false;
	}

	protected void setIamOnFire() {
		IamOnfire = false;
		if (getMyPos() instanceof worldModel.Building
				&& ((worldModel.Building) getMyPos()).isOnFire())
			IamOnfire = true;
		if (worldModel.reachableRoads.size() == 0) {
			boolean noReachableNotIgnited = true;
			for (worldModel.Building building : worldModel.reachableBuildings) {
				if (building.stFire == 0) {
					noReachableNotIgnited = false;
					break;
				}
			}
			if (noReachableNotIgnited)
				IamOnfire = false;
		}
	}

	public int getDistToFire(worldModel.Building building) {
		if (!building.distToFireCalculated) {
			building.distToFire = getDistToFireCalc(building);
			building.distToFireCalculated = true;
		}
		return building.distToFire;
	}

	private int getDistToFireCalc(worldModel.Building building) {
		int minDist = Integer.MAX_VALUE;
		TreeSet<Road> roads = new TreeSet<Road>(new Comparator<Road>() {
			public int compare(Road r1, Road r2) {
				if (r1.distanceFromSelf < r2.distanceFromSelf)
					return -1;
				if (r1.distanceFromSelf > r2.distanceFromSelf)
					return 1;
				if (r1.getID().getValue() > r2.getID().getValue())
					return 1;
				if (r1.getID().getValue() < r2.getID().getValue())
					return -1;
				return 0;
			}
		});
		roads.addAll(building.getVisionRoads(worldModel).keySet());
		done: for (Road road : roads) {
			if (road.isReachable)
				for (Coordinate point : building.getVisionRoads(worldModel)
						.get(road)) {
					Entrance e = null;
					for (Entrance ent : road.entrances)
						if (ent.isReachable
								&& ent.shapeInArea != null
								&& ent.shapeInArea.contains(GeometryTools
										.coordToGeometry(point, worldModel
												.getTrafficSimulator().geom))) {
							if (minDist > ent.neighbour.distanceFromSelf + 1
									&& ent.neighbour.distanceFromSelf >= 0) {
								minDist = ent.neighbour.distanceFromSelf + 1;
								e = ent;
								break;
							}
						}

					if (e != null) {
						if (getMyPos().getID().getValue() == road.getID()
								.getValue())
							return 1;
						minDist = e.neighbour.distanceFromSelf + 1;
						break done;
					}
				}
		}
		for (worldModel.Building buil : building.visionBuildings.keySet()) {
			if (buil.isReachable) {
				if (getMyPos().getID().getValue() == buil.getID().getValue())
					return 1;
				if (buil.distanceFromSelf < minDist)
					minDist = buil.distanceFromSelf;
			}
		}
		if (minDist == Integer.MAX_VALUE)
			return 1000;
		return minDist;
	}

	protected abstract void process() throws ActionCommandException;
}