package agent;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;

import radar.Package;
import radar.Package.HeaderType;
import rescuecore2.messages.Command;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.Entity;
import worldModel.WorldModel;

public class CenterForAmbulance {
	public ThinkAgent<? extends StandardEntity> agent = null;
	protected WorldModel worldModel = null;
	protected ArrayList<Human> humans = new ArrayList<Human>();
	protected ArrayList<AmbulanceTeam> allAmbs = new ArrayList<AmbulanceTeam>();
	protected ArrayList<AmbulanceTeam> PgAmb = new ArrayList<AmbulanceTeam>();
	protected BitSet message = new BitSet();
	protected ArrayList<AmbulanceTeam> freeAmbs = new ArrayList<AmbulanceTeam>();
	private HashMap<AmbulanceTeam, Integer> lastSetTime = new HashMap<AmbulanceTeam, Integer>();
	private HashMap<AmbulanceTeam, worldModel.Area> lastSetPos = new HashMap<AmbulanceTeam, worldModel.Area>();
	private HashMap<AmbulanceTeam, HashMap<worldModel.Area, Integer>> unreachableAreas = new HashMap<AmbulanceTeam, HashMap<worldModel.Area, Integer>>();
	private static final int SAFE_TIME = 1;
	private int numOfAmb = 1;
	private int generateNum = 10000;

	public CenterForAmbulance(ThinkAgent<? extends StandardEntity> age) {
		agent = age;
	}

	public Comparator<Entity> newComparator = new Comparator<Entity>() {
		public int compare(Entity a1, Entity a2) {
			int m = 2;
			if (agent.bigMap)
				m = 1;
			Human hu1 = (Human) a1;
			Human hu2 = (Human) a2;
			double d1 = worldModel.getArea(hu1.getPosition()).distanceFromSelf;
			double d2 = worldModel.getArea(hu2.getPosition()).distanceFromSelf;
			double f1 = ((hu1.deadtime) * m + d1);
			double f2 = ((hu2.deadtime) * m + d2);
			if (hu1 instanceof Civilian && hu2 instanceof Civilian) {
				if (hu1.deadtime > time && hu2.deadtime <= time)
					return 1;
				if (hu2.deadtime > time && hu1.deadtime <= time)
					return -1;
			}
			if (f1 > f2)
				return 1;
			if (f2 > f1)
				return -1;
			return 0;
		}
	};

	public boolean isInRefuge(Human h) {
		if (worldModel.getEntityByInt(h.getPosition().getValue()) instanceof Refuge)
			return true;
		return false;
	}

	public boolean isOnFire(Entity building) {
		if (building instanceof Building
				&& ((((Building) building).isFierynessDefined()
						&& ((Building) building).getFieryness() > 0 && ((Building) building)
						.getFieryness() != 4) || ((worldModel.Building) ((Building) building).area).stFire == 1))
			return true;
		return false;
	}

	// change2
	public boolean goodEnough(Human h) {
		if ((h.hasBuriedness || h instanceof Civilian)
				&& h.isPositionDefined()
				&& !isOnFire(worldModel.getEntity(h.getPosition()))
				&& h.deadtime > agent.time
				&& (worldModel.getEntity(h.getPosition()) instanceof Building || (worldModel
						.getEntity(h.getPosition()) instanceof Road
						&& h.isDamageDefined() && h.getDamage() != 0))
				&& !isInRefuge(h)) {
			return true;
		}
		return false;
	}

	protected void sort() {
		humans = new ArrayList<Human>();
		for (Human hu : worldModel.agents) {
			// System.out.println("CenterAmb==>>"+"CV"+hu+"Time:"+time);

			if (goodEnough(hu))
				humans.add(hu);
		}
		for (Civilian cv : worldModel.getCivilians()) {
			if (goodEnough(cv))
				humans.add(cv);
		}
		Collections.sort(humans, new Comparator<Human>() {
			@Override
			public int compare(Human arg0, Human arg1) {
				if (arg0 instanceof Civilian && !(arg1 instanceof Civilian))
					return 1;
				if (!(arg0 instanceof Civilian) && arg1 instanceof Civilian)
					return -1;
				int dead0 = arg0.deadtime;
				int dead1 = arg1.deadtime;
				if (dead0 > dead1)
					return 1;
				if (dead1 > dead0)
					return -1;
				return 0;
			}
		});
		// System.out.println("CenterAmb==>>" + "firstCV:" + worldModel.agents
		// + worldModel.getCivilians());
		// System.out.println("secCV:" + humans + "Time:" + time);
		// for (Human hh : worldModel.agents)
		// System.out.println("=====<<" + "agent.deadtime" + hh.deadtime);
		// for (Human hh : worldModel.getCivilians())
		// System.out.println("=====<<" + "CV.model.deadtime" + hh.deadtime);
		// for (Human hh : humans)
		// System.out.println("=====<<" + "secCV" + hh.deadtime);
	}

	private ArrayList<TaskAssign> allTasks = new ArrayList<TaskAssign>();

	private void setTimeAndPos(AmbulanceTeam amb) {
		lastSetTime.put(amb, agent.time);
		lastSetPos.put(amb, ((Area) amb.getPosition(worldModel)).area);
		amb.isFree = true;
		amb.PgAim = null;
	}

	private void needToBeFree() {
		for (AmbulanceTeam a : allAmbs) {
			if (a.lastSetAim != null
					&& (a.lastSetAim.employedDeadTime - a.lastSetAim.deadtime) >= 20
			/* && (a.lastSetAim.deadtime - a.lastSetAim.employedDeadTime) < 502 */) {
				setTimeAndPos(a);
				for (int i = 0; i < allTasks.size(); i++) {
					if (allTasks.get(i).human.equals(a.lastSetAim)) {
						allTasks.remove(i);
						i--;
					}
				}
			}
		}
	}

	boolean firstTime = true;

	@SuppressWarnings("unchecked")
	private void setOwner() {
		checkAreaReachablity();
		checkHumansOnAreas();
		needToBeFree();
		for (AmbulanceTeam amb : allAmbs) {
			// reachable nist
			if (amb.feedback && !amb.isReachable && amb.lastSetAim != null
					&& amb.setAimTime + SAFE_TIME < agent.time) {
				if (worldModel.getEntity(amb.lastSetAim.getPosition()) instanceof Area)
					unreachableAreas
							.get(amb)
							.put(((Area) amb.lastSetAim.getPosition(worldModel)).area,
									agent.time);
				setTimeAndPos(amb);
				for (int i = 0; i < allTasks.size(); i++) {
					if (allTasks.get(i).ambs.contains(amb)) {
						for (AmbulanceTeam a : allTasks.get(i).ambs)
							if (a.lastSetAim != null
									&& a.lastSetAim.equals(amb.lastSetAim)) {
								setTimeAndPos(a);
							}
						TaskAssign removed = allTasks.remove(i);
						if (removed.human == amb.lastSetAim)
							break;
						i--;
					}
				}

			}
			// age mozakhraf bashe faghat khodesho bar midare
			if (amb.feedback && amb.isMozakhraf && amb.lastSetAim != null
					&& amb.setAimTime + SAFE_TIME < agent.time) {
				for (TaskAssign t : allTasks) {
					if (amb.lastSetAim != null
							&& t.human.equals(amb.lastSetAim)
							&& t.ambs.contains(amb)) {
						t.ambs.remove(amb);
						setTimeAndPos(amb);
					}
				}
			}
		}

		// TODO: tofe exception-- baad chera 2 tas?
		// ---------------------------------------------------------
		// for (TaskAssign t : allTasks) {
		// if (t.human.getPosition(model) instanceof Refuge) {
		// for (AmbulanceTeam amb : t.ambs) {
		// log("tof" + amb);
		// setTimeAndPos(amb);
		// }
		// t.ambs.clear();
		// }
		// }
		//
		// for (AmbulanceTeam amb : allAmbs)
		// if (amb.PgAim != null
		// && amb.PgAim.getPosition(model) instanceof Refuge) {
		// log("##tof");
		// setTimeAndPos(amb);
		// }
		// ---------------------------------------------------------

		for (AmbulanceTeam amb : allAmbs) {
			if (!lastSetTime.containsKey(amb)) {
				setTimeAndPos(amb);
			} else if (lastSetTime.get(amb) <= agent.time) {
				if (amb.isFree) {
					setTimeAndPos(amb);
				} else {
					// TODO: check konim be jaye inke 3 ta ezafe besh
					lastSetTime.put(amb, lastSetTime.get(amb) + 3);
					for (int i = 0; i < allTasks.size(); i++) {
						if (allTasks.get(i).ambs.contains(amb)) {
							int index = allTasks.get(i).ambs.indexOf(amb);
							if (allTasks.get(i).time.get(index) == agent.time)
								allTasks.get(i).time.set(index,
										allTasks.get(i).time.get(index) + 3);
							if (allTasks.get(i).time.get(index) < lastSetTime
									.get(amb)) {
								for (AmbulanceTeam a : allTasks.get(i).ambs)
									if (a.lastSetAim != null
											&& a.lastSetAim.equals(allTasks
													.get(i).human)) {
										setTimeAndPos(a);
									}
								allTasks.remove(i);
								i--;
							}
						}
					}
				}
			}
		}

		// taske ba'adish
		for (AmbulanceTeam amb : allAmbs) {
			if (lastSetTime.get(amb) == agent.time) {
				for (TaskAssign t : allTasks) {
					if (t.ambs.contains(amb)) {
						int a = t.ambs.indexOf(amb);
						if (t.time.get(a) > agent.time) {
							setTarget(amb, t.human, t.time.get(a));
							break;
						}
					}
				}
			}
		}

		ArrayList<Human> taskHumans = new ArrayList<Human>();
		taskHumans.addAll(humans);
		/*
		 * for (Human h : taskHumans) if (h.isPositionDefined() &&
		 * h.getPosition().getValue() == 72021) System.out.println("&&&&&&&&&&&"
		 * + time); for (TaskAssign h : allTasks) if
		 * (h.human.isPositionDefined() && h.human.getPosition().getValue() ==
		 * 72021) System.out.println("###########" + time);
		 */
		for (TaskAssign t : allTasks) {
			for (Human h : taskHumans)
				if (t.human.equals(h)) {
					taskHumans.remove(h);
					break;
				}
		}
		for (Human h : taskHumans)
			h.updateDistance(worldModel);

		HashMap<AmbulanceTeam, Integer> lastSetTimeCopy = (HashMap<AmbulanceTeam, Integer>) lastSetTime
				.clone();
		HashMap<AmbulanceTeam, worldModel.Area> lastSetPosCopy = (HashMap<AmbulanceTeam, worldModel.Area>) lastSetPos
				.clone();

		done: for (AmbulanceTeam amb : lastSetPos.keySet()) {
			for (int i = allTasks.size() - 1; i >= 0; i--) {
				if (allTasks.get(i).ambs.contains(amb)) {
					int a = allTasks.get(i).ambs.indexOf(amb);
					if (allTasks.get(i).time.get(a) > agent.time) {
						lastSetTimeCopy.put(amb, allTasks.get(i).time.get(a));
						if (allTasks.get(i).human.getPosition(worldModel) instanceof Area)
							lastSetPosCopy.put(amb,
									((Area) allTasks.get(i).human
											.getPosition(worldModel)).area);
						// TODO: zoodtar refuge beshe age gharare beshe
						continue done;
					}
				}
			}
		}

		State bestState = null;
		numOfAmb = allAmbs.size();
		/*
		 * if (time < 72 && time > 20) for (Human h : taskHumans)
		 * System.out.println("////////////////////////" +
		 * h.getPosition().getValue());
		 */

		/*
		 * for (Human h : taskHumans) if (h.isPositionDefined() &&
		 * h.getPosition().getValue() == 72021)
		 * System.out.println("@@@@@@@@@@@@" + time);
		 */

		for (int i = 0; i < generateNum; i++) {
			State state = new State(worldModel);
			state.time = agent.time;
			state.lastSetPos = (HashMap<AmbulanceTeam, worldModel.Area>) lastSetPosCopy
					.clone();
			state.lastSetTime = (HashMap<AmbulanceTeam, Integer>) lastSetTimeCopy
					.clone();
			state.generateRandom(numOfAmb, taskHumans, unreachableAreas);
			if (bestState == null
					|| state.getFitness() > bestState.getFitness())
				bestState = state;

		}
		// if (bestState.tasks.size() == 0)
		// System.err.println("bugggggg " + agent.time);
		for (TaskAssign t : bestState.tasks) {
			for (int i = 0; i < t.ambs.size(); i++) {
				if (lastSetTime.get(t.ambs.get(i)) == agent.time) {
					setTarget(t.ambs.get(i), t.human, t.time.get(i));
					if (!allTasks.contains(t)) {
						allTasks.add(t);
					}
				}
			}
		}
	}

	private void setTarget(AmbulanceTeam amb, Human h, int time) {
		amb.PgAim = h;
		amb.lastSetAim = h;
		amb.setAimTime = agent.time;
		lastSetTime.put(amb, time);
		lastSetPos.put(amb, ((Area) h.getPosition(worldModel)).area);
		amb.isFree = false;
		h.employedDeadTime = h.deadtime;
	}

	public void checkHumansOnAreas() {
		for (worldModel.Area area : worldModel.areas)
			area.checkForAmbForBFS = false;
		for (AmbulanceTeam amb : worldModel.ambulanceTeams)
			(worldModel.getArea(amb.getPosition())).checkForAmbForBFS = true;
		for (worldModel.Refuge refuge : worldModel.refuges)
			refuge.checkForAmbForBFS = true;
		for (FireBrigade fb : worldModel.fireBrigades)
			if (fb.hasBuriedness)
				(worldModel.getArea(fb.getPosition())).checkForAmbForBFS = true;
		for (PoliceForce pf : worldModel.policeForces)
			if (pf.hasBuriedness)
				((Area) pf.getPosition(worldModel)).area.checkForAmbForBFS = true;

		for (Civilian civil : worldModel.getCivilians())
			if ((civil.getPosition(worldModel) instanceof Building || civil
					.getPosition(worldModel) instanceof Road))
				((Area) civil.getPosition(worldModel)).area.checkForAmbForBFS = true;
	}

	private HashMap<AmbulanceTeam, worldModel.Area> lastAssigned = new HashMap<AmbulanceTeam, worldModel.Area>();

	protected void hearInfo(Package p, Entity agent) {
		if (!(agent instanceof AmbulanceTeam))
			return;
		AmbulanceTeam amb = ((AmbulanceTeam) agent);
		amb.feedback = true;
		amb.isReachable = (p.getInt(2) == 1);
		amb.isMozakhraf = (p.getInt(2) == 1);
	}

	protected void setmessage(radar.Package p, ArrayList<AmbulanceTeam> workers) {
		int messageSize = worldModel.buildings.size();
		boolean targetInRoad = false;
		for (AmbulanceTeam a : workers)
			if (allAmbs.contains(a)
					&& a.PgAim != null
					&& (a.PgAim.getPosition(worldModel) instanceof Road || (a.PgAim
							.getPosition(worldModel) instanceof AmbulanceTeam && lastAssigned
							.get((AmbulanceTeam) a.PgAim
									.getPosition(worldModel)) instanceof worldModel.Road))) {
				targetInRoad = true;
				messageSize = worldModel.areas.size();
				break;
			}
		if (targetInRoad)
			p.addInt(1, 2);
		else
			p.addInt(0, 2);

		for (AmbulanceTeam amb : workers) {
			if (!lastAssigned.containsKey(amb))
				lastAssigned.put(amb, null);
			if (amb.PgAim == null) {
				amb.isFree = true;
				p.addInt(0, messageSize + 1);
				lastAssigned.put(amb, null);
			} else {
				amb.isFree = false;
				if (!(amb.PgAim.getPosition(worldModel) instanceof AmbulanceTeam)) {
					lastAssigned.put(amb,
							worldModel.getArea(amb.PgAim.getPosition()));
				}
				if (targetInRoad)
					p.addInt(
							worldModel.areaToIndex.get(lastAssigned.get(amb)) + 1,
							messageSize + 1);
				else
					p.addInt(worldModel.buildingToIndex.get(lastAssigned
							.get(amb)) + 1, messageSize + 1);
				p.addInt(amb.PgAim.getID().getValue() % agent.cvRecognizer,
						agent.cvRecognizer);
			}
		}
	}

	protected void sendAmbAL() {
		if (agent.time > 2) {
			radar.Package p = new radar.Package(worldModel);
			p.addHeader(HeaderType.ACenter);
			ArrayList<AmbulanceTeam> workers = new ArrayList<AmbulanceTeam>();
			if (agent.lowCommi)
				if (time % 7 == 4)
					for (int i = 0; i < worldModel.ambulanceTeams.size() / 2; i++)
						workers.add(worldModel.ambulanceTeams.get(i));
				else
					for (int i = worldModel.ambulanceTeams.size() / 2; i < worldModel.ambulanceTeams
							.size(); i++)
						workers.add(worldModel.ambulanceTeams.get(i));
			else
				workers.addAll(worldModel.ambulanceTeams);
			setmessage(p, workers);
			// String w = "";
			// for (int i = 0; i < p.getBits(0, p.start).size(); i++) {
			// if (p.getBits(0, p.start).get(i))
			// w += "1";
			// else
			// w += "0";
			// }
			// log(w);
			agent.speak(time, HeaderType.ACenter, p.send());
			if (agent.lowCommi)
				for (int i = 0; i < 3; i++)
					agent.speak(time, HeaderType.ACenter, p.send());
		}
	}

	protected int time = -1;

	// protected void log(String msg) {
	// agent.logger.log(msg);
	// }

	private void checkAreaReachablity() {
		for (AmbulanceTeam amb : allAmbs) {
			HashMap<worldModel.Area, Integer> map = unreachableAreas.get(amb);
			ArrayList<worldModel.Area> removeMap = new ArrayList<worldModel.Area>();
			for (worldModel.Area arr : map.keySet()) {
				if (agent.time - map.get(arr) > 10) {
					removeMap.add(arr);
				}
			}
			for (worldModel.Area arr : removeMap)
				map.remove(arr);
		}
	}

	protected void act() {
		time = agent.time;

		worldModel = agent.getWorldModel();
		// if (agent.noCommi || agent.lowCommi) {
		// log("nocommi");
		// return;
		// }
		if (agent.lowCommi)
			for (AmbulanceTeam amb : worldModel.ambulanceTeams)
				// chon momkene feedback nadan
				amb.isAlive = true;
		// ambaye zende ke mishe task dad
		for (AmbulanceTeam amb : worldModel.ambulanceTeams)
			if (amb.isAlive && !allAmbs.contains(amb)) {
				allAmbs.add(amb);
				unreachableAreas.put(amb,
						new HashMap<worldModel.Area, Integer>());
			}
		if (allAmbs.size() == 0)
			return;

		clearFeedbacks();

		hears();
		sort();
		// printTasks();
		setOwner();
		// printTasks();
		if (!agent.lowCommi
				|| (agent.lowCommi && (agent.time % 7 == 4 || agent.time % 7 == 5)))
			sendAmbAL();
	}

	void hears() {
		// log("hears");
		if (agent.heard != null && agent.heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : agent.heard) {
				Entity entity = worldModel.getEntity(cmd.getAgentID());
				if (cmd instanceof AKSpeak && entity != null
						&& entity instanceof AmbulanceTeam) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						Package p = new Package(worldModel);
						p.hear(recivedSpeak);
						HeaderType h = p.tellHeader();
						if (h == radar.Package.HeaderType.Info) {
							hearInfo(p, entity);
						}
					}
				}
			}
		}
	}

	// private void printTasks() {
	// log("--------------------------------------------------------");
	// log("alltasks.size: " + allTasks.size());
	//
	// for (AmbulanceTeam amb : allAmbs)
	// log("amb: " + amb.getID().getValue() + " isReach: "
	// + amb.isReachable + " isfree: " + amb.isFree
	// + " ismozakhraf" + amb.isMozakhraf + " lastsett: "
	// + lastSetTime.get(amb));
	// for (TaskAssign t : allTasks)
	// log(" task: " + t.human + "  pos:  " + t.human.getPosition()
	// + "    " + t.ambs + " " + time);
	// log("--------------------------------------------------------");
	// }

	private void clearFeedbacks() {
		for (AmbulanceTeam amb : worldModel.ambulanceTeams) {
			amb.feedback = false;
			amb.isReachable = true;
			amb.isMozakhraf = false;
		}
	}

	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.AMBULANCE_CENTRE);
	}

}
