package agent;

import java.util.ArrayList;
import java.util.HashSet;

import radar.Package;
import radar.Package.HeaderType;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.StandardEntity;
import utilities.Logger;
import worldModel.Area;
import worldModel.Building;
import worldModel.Entrance;
import worldModel.Road;
import worldModel.WorldModel;

public class CenterForFire {
	public ThinkAgent<StandardEntity> agent = null;
	protected WorldModel worldModel = null;

	private FireZoneBuilder fireZoneBuilder = null;

	private Logger logger = null;

	public CenterForFire(ThinkAgent<StandardEntity> agent, Logger logger) {
		this.agent = agent;
		this.worldModel = agent.getWorldModel();
		fireZoneBuilder = new FireZoneBuilder(worldModel, agent.bigMap,
				agent.averageOfSurfaces, logger, agent.noCommi);
		this.logger = logger;
	}

	private void log(String msg) {
		// logger.log(msg);
	}

	protected void act() {

		if (agent.bigMap)
			simulate();

		// Zone...
		fireZoneBuilder.executeModelZones(null);

		setWorkersNum();
		setHibitors();

		// FireGroup...
		setFireGroup();
		setVisionBuilding();
		setFireAgentsForFireZones();

		// Task Assign
		// Set zone
		if (fireZoneBuilder.zoneChange()) {
			for (FireBrigade fireBrigade : worldModel.fireBrigades) {
				fireBrigade.zone = null;
			}
		} else {
			updateFireBrigadesZones();
		}

		for (FireBrigade fireBrigade : worldModel.fireBrigades)
			if (fireBrigade.zone == null)
				setAimNull(fireBrigade);
		setZoneForFireBrigades();

		// Set Aim
		checkAim();
		setAim();

		if (!agent.lowCommi
				|| (agent.lowCommi && (agent.time % 7 == 1 || agent.time % 7 == 2)))
			sendFireAL();
	}

	private void simulate() {
		for (Building b : worldModel.buildings) {
			if (b.stFire > 0
					&& (((agent.time - b.timeOfStartOfFire) > 10 && !b.isSimulatedForFire))) {
				for (Building bu : b.nearBuildings) {
					if ((bu.getLTS() < 0 || agent.time - bu.getLTS() > 20 * (bu.counterForSimulate + 1))
							&& !bu.isNotIgnitable()
							&& !bu.isSimulatedForFire
							&& bu.stFire == 0
							&& (!bu.isFirenessDefined() || bu.getFieryness() != 8)) {
						bu.stFire = 1;
						bu.LTS = agent.time;
						bu.isSimulatedForFire = true;
						bu.counterForSimulate++;
					}
				}
			}

			if (b.serverBuilding().isOnFire()
					&& b.serverArea instanceof GasStation && !b.itWasSimulated) {
				b.itWasSimulated = true;
				for (StandardEntity bui : worldModel.getAreasInRange(b, 50000)) {
					if (bui instanceof rescuecore2.standard.entities.Building) {
						worldModel.Building bu = ((worldModel.Building) (((rescuecore2.standard.entities.Building) bui).area));
						if ((bu.getLTS() < 0)
								&& !bu.isNotIgnitable()
								&& !bu.isSimulatedForFire
								&& bu.stFire == 0
								&& (!bu.isFirenessDefined() || bu
										.getFieryness() != 8)) {
							bu.stFire = 1;
							bu.LTS = agent.time;
							bu.isSimulatedForFire = true;
							bu.counterForSimulate++;
						}
					}
				}
			}
		}
	}

	private void setAim() {
		if (fireZoneBuilder.modelZones.size() == 0)
			return;

		HashSet<FireBrigade> fireAgents = new HashSet<FireBrigade>();
		for (FireBrigade fireBrigade : worldModel.fireBrigades)
			if (isOk(fireBrigade) && fireBrigade.zone != null
					&& fireBrigade.aim == null)
				fireAgents.add(fireBrigade);

		while (true) {
			boolean foundAim = false;
			foundAim = findBestAim(fireAgents, 1, true, false);
			if (!foundAim)
				foundAim = findBestAim(fireAgents, 3, true, true);
			if (!foundAim)
				foundAim = findBestAim(fireAgents, 100, false, true);
			if (!foundAim)
				foundAim = findBestAim(fireAgents, 100, false, false);
			if (fireAgents.size() == 0)
				break;
		}
	}

	protected boolean findBestAim(HashSet<FireBrigade> fireAgents, int workers,
			boolean withSetGood, boolean isReal) {
		double minDis = Integer.MAX_VALUE;
		Building minBuild = null;
		FireBrigade fireb = null;

		ArrayList<Building> aims = new ArrayList<Building>();

		for (FireBrigade fireBrigade : fireAgents) {
			if (withSetGood && !isReal)
				aims.addAll(fireBrigade.zone.allAroundsWithSetGood);
			else if (withSetGood && isReal)
				aims.addAll(fireBrigade.zone.realAroundsWithSetGood);
			else if (!withSetGood && isReal)
				aims.addAll(fireBrigade.zone.realArounds);
			else if (!withSetGood && !isReal)
				aims.addAll(fireBrigade.zone.allArounds);
			for (Building building : aims) {
				if (building.assigned < workers) {
					double dist = Math.hypot(
							fireBrigade.getX() - building.getX(),
							fireBrigade.getY() - building.getY());
					if (dist < minDis) {
						minDis = dist;
						minBuild = building;
						fireb = fireBrigade;
					}
				}
			}
		}
		if (minBuild != null) {
			fireb.aim = minBuild;
			minBuild.assigned++;
			fireAgents.remove(fireb);
			return true;
		}
		return false;

	}

	protected void checkAim() {
		for (FireBrigade fireBrigade : worldModel.fireBrigades) {
			if (fireBrigade.aim == null)
				continue;

			if (!isOk(fireBrigade)) {
				setAimNull(fireBrigade);
				continue;
			}

			if ((fireBrigade.aim.isFirenessDefined() && fireBrigade.aim
					.getFieryness() == 8) || fireBrigade.aim.stFire == 0) {
				setAimNull(fireBrigade);
				continue;
			}

			if (!fireGroups.get(fireBrigade.fgNum).nearTargets
					.contains(fireBrigade.aim)) {
				setAimNull(fireBrigade);
				continue;
			}

			if (!fireBrigade.zone.allArounds.contains(fireBrigade.aim))
				setAimNull(fireBrigade);
		}
	}

	private void setFireAgentsForFireZones() {
		for (FireZone fireZone : fireZoneBuilder.modelZones) {
			for (Building building : fireZone.buildings)
				if (building.fgNum > -1)
					fireZone.fireGroups.add(fireGroups.get(building.fgNum));

			for (GroupForFire gf : fireZone.fireGroups)
				fireZone.fireAgents.addAll(gf.fireAgents);
		}
	}

	private void setFireBrigadeForZone(ArrayList<FireZone> zones,
			boolean checkWorkersNum) {
		boolean finished = false;
		while (!finished) {
			finished = true;
			double mindis = Integer.MAX_VALUE;
			FireBrigade bestFireBrigade = null;
			FireZone minFireZone = null;
			for (FireZone fireZone : zones) {
				if (fireZone.workersNum > 0 || !checkWorkersNum)
					for (FireBrigade fireBrigade : fireZone.fireAgents) {
						if (isOk(fireBrigade) && fireBrigade.zone == null) {
							double dist = Math.hypot(fireZone.centerX
									- fireBrigade.getX(), fireZone.centerY
									- fireBrigade.getY());
							if (mindis > dist) {
								mindis = dist;
								bestFireBrigade = fireBrigade;
								minFireZone = fireZone;
							}
						}
					}
			}

			if (bestFireBrigade != null) {
				bestFireBrigade.zone = minFireZone;
				minFireZone.workersNum--;
				finished = false;
			}
		}
	}

	protected void setZoneForFireBrigades() {
		ArrayList<FireZone> hibitors = new ArrayList<FireZone>();
		ArrayList<FireZone> nothibitor = new ArrayList<FireZone>();
		for (FireZone fireZone : fireZoneBuilder.modelZones)
			if (fireZone.hibitor)
				hibitors.add(fireZone);
			else
				nothibitor.add(fireZone);

		setFireBrigadeForZone(hibitors, true);
		setFireBrigadeForZone(nothibitor, true);
		setFireBrigadeForZone(nothibitor, false);
		setFireBrigadeForZone(hibitors, false);
	}

	public void setAimNull(FireBrigade fb) {
		if (fb.aim != null)
			fb.aim.assigned--;

		fb.aim = null;
	}

	protected void setVisionBuilding() {
		for (FireZone fireZone : fireZoneBuilder.modelZones) {
			for (Building build : fireZone.allArounds) {
				for (Area nearArea : build.nearAreas50000) {
					if (nearArea.fgNum > -1
							&& nearArea.distance(build) <= Agent.maxDistanceExtinguish
									* Agent.PECENTAGE_OF_MAX_DISTANCE) {
						fireZone.fireGroups.add(fireGroups.get(nearArea.fgNum));
						fireGroups.get(nearArea.fgNum).nearTargets.add(build);
					}
				}
			}
		}
	}

	public boolean isInCommon(ArrayList<Building> b1, HashSet<Building> b2) {
		for (Building b : b1)
			if (b2.contains(b))
				return true;
		return false;
	}

	public void updateFireBrigadesZones() {
		for (FireBrigade fireBrigade : worldModel.fireBrigades) {
			if (!isOk(fireBrigade) || fireBrigade.zone == null
					|| fireBrigade.zone.buildings.get(0).fireZoneNum == -1) {
				fireBrigade.zone = null;
				continue;
			}

			FireZone newZone = fireZoneBuilder.modelZones
					.get(fireBrigade.zone.buildings.get(0).fireZoneNum);
			if (isInCommon(newZone.ignitedBuilds,
					fireGroups.get(fireBrigade.fgNum).nearTargets)) {
				fireBrigade.zone = newZone;
			} else {
				fireBrigade.zone = null;
			}
		}
	}

	private GroupForFire bfsForSetFireGroups(FireBrigade fireBrigade) {
		GroupForFire fireGroup = new GroupForFire();
		fireGroup.num = fireGroups.size();

		Area area = worldModel.getArea(fireBrigade.getPosition());
		fireGroup.fireAgents.addAll(area.fireBrigades);

		if (area instanceof Building)
			fireGroup.buildings.add((Building) area);
		else
			fireGroup.roads.add((Road) area);

		HashSet<Entrance> layer = new HashSet<Entrance>();
		layer.addAll(area.entrances);

		while (layer.size() > 0) {
			HashSet<Entrance> newLayer = new HashSet<Entrance>();
			for (Entrance enterance : layer) {
				if (enterance.mark)
					continue;
				enterance.mark = true;
				Entrance neighbourE = enterance.neighbour;
				if (neighbourE.isItConnectedToNeighbour) {
					Area areaOfNeighbourE = neighbourE.area;
					if (areaOfNeighbourE instanceof Building)
						fireGroup.buildings.add((Building) areaOfNeighbourE);
					if (areaOfNeighbourE instanceof Road)
						fireGroup.roads.add((Road) areaOfNeighbourE);

					fireGroup.fireAgents.addAll(neighbourE.area.fireBrigades);
					for (Entrance internal : neighbourE.internalEntrances)
						if (!internal.mark)
							newLayer.add(internal);
					enterance.neighbour.mark = true;
				}
			}
			layer = newLayer;
		}

		for (FireBrigade f : fireGroup.fireAgents)
			f.fgNum = fireGroup.num;
		for (Building build : fireGroup.buildings)
			build.fgNum = fireGroup.num;
		for (Road road : fireGroup.roads)
			road.fgNum = fireGroup.num;

		return fireGroup;
	}

	ArrayList<GroupForFire> fireGroups = new ArrayList<GroupForFire>();

	private boolean isOk(FireBrigade fireBrigade) {
		return (!fireBrigade.hasBuriedness && fireBrigade.getWater() > 0);
	}

	private void setFireGroup() {
		agent.worldModel.clearAreas();

		for (Area a : worldModel.areas) {
			a.fgNum = -1;
			a.fireBrigades.clear();
		}

		fireGroups = new ArrayList<GroupForFire>();
		for (FireBrigade fireBrigade : worldModel.fireBrigades) {
			fireBrigade.fgNum = -1;

			if (!isOk(fireBrigade))
				continue;

			Area area = ((worldModel.getAreaByInt(fireBrigade.getPosition()
					.getValue())));
			area.fireBrigades.add(fireBrigade);
		}

		for (FireBrigade fireBrigade : worldModel.fireBrigades)
			if (fireBrigade.fgNum == -1 && isOk(fireBrigade))
				fireGroups.add(bfsForSetFireGroups(fireBrigade));
	}

	private void setWorkersNum() {
		for (FireZone fireZone : fireZoneBuilder.modelZones) {
			boolean isZoneReal = false;
			for (Building bu : fireZone.buildings)
				if (bu.stFire > 0 && !bu.isSimulatedForFire) {
					isZoneReal = true;
					break;
				}

			if (!isZoneReal) {
				fireZone.workersNum = 1;
			} else {
				fireZone.workersNum = fireZone.simulatedAroundsWithSetGood
						.size();
				if (fireZone.valueForWorkwerNum <= 7)
					fireZone.workersNum = 5;
			}
		}
	}

	protected void setHibitors() {
		for (FireZone fireZone : fireZoneBuilder.modelZones) {
			if (fireZone.ignitedBuilds.size() < 10)
				fireZone.hibitor = true;
		}
	}

	protected void setmessage(Package p, ArrayList<FireBrigade> workers) {
		for (FireBrigade fire : workers) {
			if (fire.aim == null) {
				p.addInt(0, worldModel.buildings.size() + 1);
			} else {
				p.addInt(worldModel.buildings.indexOf(fire.aim) + 1,
						worldModel.buildings.size() + 1);
				int lasttimeseen = agent.time - fire.aim.getLTS();
				if (lasttimeseen >= 32)
					lasttimeseen = 31;
				p.addInt(lasttimeseen, 32);
			}
		}
	}

	protected void sendFireAL() {
		if (agent.time > 2) {
			Package p = new Package(worldModel);
			ArrayList<FireBrigade> workers = new ArrayList<FireBrigade>();
			p.addHeader(HeaderType.FCenter);
			if (agent.lowCommi)
				if (agent.time % 7 == 1)
					for (int i = 0; i < worldModel.fireBrigades.size() / 2; i++)
						workers.add(worldModel.fireBrigades.get(i));
				else
					for (int i = worldModel.fireBrigades.size() / 2; i < worldModel.fireBrigades
							.size(); i++)
						workers.add(worldModel.fireBrigades.get(i));

			workers.addAll(worldModel.fireBrigades);
			setmessage(p, workers);
			agent.speak(agent.time, HeaderType.FCenter, p.send());
		}
	}
}
