package agent;

import geometry.Mathematic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;

import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import worldModel.Area;
import worldModel.Building;
import worldModel.Entrance;
import worldModel.Hydrant;
import worldModel.Refuge;

import com.vividsolutions.jts.geom.Coordinate;

public class FireBrigadeAgent extends ThinkAgent<FireBrigade> {
	protected int maxPower = -1;
	private int maxWater = -1;

	private Building searchAround = null;
	private int startOfSearchAroundTemperature = 0;

	private static final String MAX_POWER_KEY = "fire.extinguish.max-sum";
	private static final String MAX_WATER_KEY = "fire.tank.maximum";
	private HashSet<Building> allArounds = new HashSet<Building>();
	private HashSet<Building> allAroundsWithsetGood = new HashSet<Building>();

	private boolean amISearching = false;
	FireZone lastZone = null;
	private ArrayList<Building> toBeSearched = new ArrayList<Building>();

	private int rw = 0;

	ArrayList<Building> unreachableAndIgnitedBuildings = new ArrayList<Building>();

	private boolean goingToRefuge = false;

	int lastInNoHear = -1;

	protected void postConnect() {
		super.postConnect();
		maxPower = config.getIntValue(MAX_POWER_KEY);
		maxWater = config.getIntValue(MAX_WATER_KEY);
		fbNoHear.agent = this;

	}

	// BFS - START

	public ArrayList<EntityID> BFS(Area a, Area b) {
		HashSet<Entrance> marked = new HashSet<>();
		ArrayList<Entrance> layer = new ArrayList<>();
		layer.addAll(a.entrances);
		ArrayList<Area> path = new ArrayList<>();
		while (path.size() == 0) {
			HashSet<Entrance> al = new HashSet<>();
			if (path.size() > 0) {
				break;
			}
			for (Entrance i : layer) {
				if (path.size() > 0) {
					break;
				}
				if (i.neighbour.isItConnectedToNeighbour) {
					for (Entrance j : i.neighbour.internalEntrances) {
						if (path.size() > 0) {
							break;
						}
						if (!marked.contains(j)) {
							al.add(j);
							j.preEntrance = i;
							marked.add(j);
							if (j.area == b) {
								while (!path.contains(a)) {
									path.add(j.area);
									j = j.preEntrance;
								}
								Collections.reverse(path);
								break;
							}
							if (j.neighbour.area == b) {
								while (!path.contains(a)) {
									path.add(j.area);
									j = j.preEntrance;
								}
								Collections.reverse(path);
								path.add(b);
								break;
							}
						}
					}
				}
			}
			layer.clear();
			layer.addAll(al);
		}
		ArrayList<EntityID> rtrn = new ArrayList<>();
		for (Area e : path) {
			rtrn.add(e.getID());
		}
		return rtrn;
	}

	@SuppressWarnings("unused")
	private void logs1() {
		// log("MeID: " + me().getID().getValue());
		// if (getMyPos() instanceof Road)
		// log("My Position: " + getMyPos() + "  instanceof ROAD");
		// else if (getMyPos() instanceof Building)
		// log("My Position: " + getMyPos() + "  instance of BUILDING");
		// log("isBigMap = " + bigMap);
		// log("MapType = " + worldModel.map);
		// if (me().aim != null) {
		// log("Aim: " + me().aim);
		// } else
		// log("My aim is null");
		// log("My Damage: " + me().getDamage());
		// log("My Buriedness: " + me().getBuriedness());
		// String s = "";
		// String st = "";
		// String onFire = "";
		// for (Building b : worldModel.buildings) {
		// if (b.getFieryness() > 0)
		// s += b.id + " ";
		// if (b.stFire > 0)
		// st += b.id + " ";
		// if (b.isOnFire())
		// onFire += b.id + " ";
		//
		// }
		// log("buildingsWithFireyness: " + s);
		// log("buildingsWithStFire: " + st);
		// log("onFireBuildings: " + onFire);
		// s = "";
		// for (Building b : allArounds)
		// s += b.id + " ";
		// log("arounds: " + s);
		// s = "";
		// for (Building b : allAroundsWithsetGood)
		// s += b.id + " ";
		// log("allAroundsWithSetGood: " + s);
		// s = "";
		// for (Building b : fireZoneBuilder.reachableAndIgnitedBuildings)
		// s += b.id + " ";
		// log("fireZoneBuilder.reachableAndIgnitedBuildings: " + s);
		// // String melina = "";
		// // for (Hydrant h : worldModel.reachableHydrants){
		// // melina += " " + h.getID() + "Power: " + h;
		// // }
		// // for (Refuge r : worldModel.reachableRefuges)
		// // melina += " " + r.getID();
		// // log("log checkwater: " + melina);
		// // log("min: "
		// // + worldModel.getMinPath(getMyPos(),
		// // worldModel.getAreaByInt(35921), true));
		// // log("min: " +
		// isReachable(worldModel.getAreaByInt(35921).getID()));
	}

	@SuppressWarnings("unused")
	private void logs2() {
		// -Xss10m -XX:+AggressiveHeap -Xmx3G -Xms3G
		// log("LOGS: -->");
		// log("================================================================");
		// String str1 = "";
		// for (Building i : worldModel.buildings) {
		// if (i.stFire == 1)
		// str1 = str1 + i.getID().getValue() + "  ";
		// }
		// log("buildings with stFire == 1   " + str1);
		// double x1 = me().getX(), y1 = me().getY();
		// log("MyZone");
		// if (me().myZone != null) {
		// double x2 = me().myZone.centerX, y2 = me().myZone.centerY;
		// log("    circle(" + (int) x1 + ", " + (int) y1 + ", 1000) 4; line("
		// + (int) x1 + ", " + (int) y1 + ") (" + (int) x2 + ", "
		// + (int) y2 + ") 2;");
		// log("    zone center: " + "X == " + x2 + "   Y == " + y2);
		// } else
		// log("    no zone.....!!!!!");
		// log("Around Buildings & ModelZones");
		// if (fireZoneBuilder.modelZones == null)
		// log("ModelZones & Around Buildings is null");
		// else if (fireZoneBuilder.modelZones.size() == 0)
		// log("    ModelZones & Around Buildings has nothing");
		// else {
		// String str2 = "";
		// for (FireZone i : fireZoneBuilder.modelZones) {
		// for (Building j : i.buildings) {
		// str2 = str2 + j.getID().getValue() + " ";
		// }
		// str2 = str2 + " ; ";
		// }
		// log("    ModelZones: " + str2);
		// }
		// for (FireZone i : fireZoneBuilder.modelZones) {
		// String aroundswsg = "", arounds = "";
		// for (Building j : i.aroundBuildingsWithSetGood) {
		// aroundswsg += " " + j.getID().getValue();
		// }
		// for (Building j : i.aroundBuildings) {
		// arounds += j;
		// }
		// log("    arounds with set good: " + aroundswsg);
		// log("    arounds wo wsg: " + arounds);
		// }
		// String sss = "";
		// for (Building b : allAroundsWithsetGood)
		// sss += " " + b;
		// log("allAroungWST: " + sss);
		// if (me().aim != null) {
		// log("Aim: " + me().aim);
		// } else
		// log("My aim is null");
		// log("================================================================");
		// log("<-- :LOGS");
	}

	@SuppressWarnings("unused")
	private void logs3() {
		// System.err.println(time);
		// log("my aim: " + me().aim);
		// if (me().aim != null) {
		// log("st: " + me().aim.stFire);
		// log("br: " + !buildingReachability(me().aim));
		// log("arounds: "
		// + (!allArounds.contains(me().aim) && me().aim.heardlastTimeSeen <
		// me().aim
		// .getLTS()));
		// }
		// log("logs for nohear");
		// log("    shenide: " + shenide);
		// if (me().aim != null) {
		// log("    st fire for aim: " + me().aim.stFire);
		// log("    aim isreachable: " + buildingReachability(me().aim));
		// log("    aim is in allaround: "
		// + allAroundsWithsetGood.contains(me().aim)
		// + "    HLTS for aim: " + me().aim.heardlastTimeSeen
		// + "    LTS: " + me().aim.getLTS());
		// } else
		// log("    aim is null");

	}

	protected void emergencyExtinguish() throws ActionCommandException {
		if (me().aim == null)
			return;
		for (FireZone fz : fireZoneBuilder.modelZones) {
			if (fz.buildings.contains(me().aim))
				continue;
			for (Building b : fz.buildings) {
				if (b.stFire > 0
						&& b.timeOfStartOfFire - CenterLastHeardTime <= 6
						&& canExtinguish(b)) {
					extinguish(b.getID(), maxPower);
				}
			}
		}
	}

	protected void searchFire(ArrayList<worldModel.Building> buildings)
			throws ActionCommandException {
		worldModel.Building target = null;
		double minDist = Integer.MAX_VALUE;
		for (worldModel.Building building : buildings) {
			double dist = 0;
			dist = Math.hypot(searchAround.getX() - building.getX(),
					searchAround.getY() - building.getY());
			if (minDist > dist && buildingReachability(building)) {
				minDist = dist;
				target = building;
			}
		}
		if (target == null)
			return;
		((FireBrigadeAgent) this).moveToFireAndSee(target);
	}

	ArrayList<Building> buildingsGoodForSearchAround = new ArrayList<Building>();

	protected void SearchAroundTemperature(Building buil)
			throws ActionCommandException {
		int minLTS;
		if (searchAround == null) {
			buildingsGoodForSearchAround = new ArrayList<Building>();
			buildingsGoodForSearchAround.addAll(buil.nearBuildings);
			searchAround = buil;
		}
		if (startOfSearchAroundTemperature == 0) {
			startOfSearchAroundTemperature = time;
		}
		HashSet<Building> toRemove = new HashSet<Building>();
		for (Building building : buildingsGoodForSearchAround) {
			if (!buildingReachability(building))
				toRemove.add(building);
		}

		buildingsGoodForSearchAround.removeAll(toRemove);

		for (EntityID id : worldModel.changeSet.getChangedEntities()) {
			if (((worldModel.getEntity(id) instanceof rescuecore2.standard.entities.Building))) {
				buildingsGoodForSearchAround.remove(worldModel.getArea(id));
			}
		}
		minLTS = getMinLTS(buildingsGoodForSearchAround);
		if (minLTS < startOfSearchAroundTemperature) {
			searchFire(buildingsGoodForSearchAround);
		}
		if (minLTS >= startOfSearchAroundTemperature) {
			searchAround = null;
			startOfSearchAroundTemperature = 0;
			buildingsGoodForSearchAround = new ArrayList<Building>();
		}
	}

	protected boolean InRangestFire(Building bui) {
		ArrayList<Building> buildings = new ArrayList<Building>();
		buildings = bui.nearBuildings;
		for (Building building : buildings) {
			// log("inrangestfire--> first");
			if (building.fireness() >= 1) {
				searchAround = null;
				startOfSearchAroundTemperature = 0;
				return true;
			}
		}
		return false;
	}

	private void CheckForSearchAroundTemperature()
			throws ActionCommandException {
		if (fireZoneBuilder.modelZones.size() == 0) {
			for (EntityID id : worldModel.changeSet.getChangedEntities()) {
				if (((worldModel.getEntity(id) instanceof rescuecore2.standard.entities.Building)
						&& ((Building) worldModel.getArea(id))
								.isTemperatureDefined()
						&& ((Building) worldModel.getArea(id)).getTemperature() > 0
						&& ((Building) worldModel.getArea(id)).fireness() <= 0 && !InRangestFire((Building) worldModel
							.getArea(id)))
						|| (searchAround != null && !InRangestFire(searchAround))) {
					if (searchAround == null) {
						SearchAroundTemperature(((Building) worldModel
								.getArea(id)));
					}
					if (searchAround != null)
						SearchAroundTemperature(searchAround);
				}
			}
		}
	}

	protected void act() throws ActionCommandException {
		if (time < 3) {
			rest();
		}

		setIamOnFire();
		if (getMyPos() instanceof Refuge || getMyPos() instanceof Hydrant)
			checkWater();

		calcutingStucks();
		if (!amIStucking)
			rw = 0;

		if (me().hasBuriedness) {
			rest();
		}

		checkEmergencySituation();

		counterForEmergencySit = 0;
		isInEmergencySituation = false;

		randomWalkComputations();

		checkWater();

		setZoneHibitors();

		allArounds.clear();
		allAroundsWithsetGood.clear();

		for (FireZone z : fireZoneBuilder.modelZones) {
			allArounds.addAll(z.simulatedArounds);
			allAroundsWithsetGood.addAll(z.simulatedAroundsWithSetGood);
		}
		if (noCommi)
			CheckForSearchAroundTemperature();
		checkForSearchAround();
		// logs3();

		// if (buildingReachability((Building) worldModel.getAreaByInt(19490))
		// && ((Building) worldModel.getAreaByInt(19490)).stFire == 1) {
		// if (worldModel.isInChangeSet(((Building) worldModel
		// .getAreaByInt(19490)).getID()))
		// extinguish(((Building) worldModel.getAreaByInt(19490)).getID(),
		// maxPower);
		// moveToFire(((Building) worldModel.getAreaByInt(22806)));
		// } else if (buildingReachability((Building) worldModel
		// .getAreaByInt(22806))) {
		// if (worldModel.isInChangeSet(((Building) worldModel
		// .getAreaByInt(22806)).getID()))
		// extinguish(((Building) worldModel.getAreaByInt(19490)).getID(),
		// maxPower);
		// moveToFire(((Building) worldModel.getAreaByInt(22806)));
		// } else
		// randomWalk(false);
		if (!shenide
				|| me().aim == null
				|| me().aim.stFire == 0
				|| !buildingReachability(me().aim)
				|| (!allArounds/* WithsetGood */.contains(me().aim) && me().aim.heardlastTimeSeen < me().aim
						.getLTS())) {
			noHear();
		} else {
			if (lowCommi) {
				emergencyExtinguish();
			}
			moveAndExtinguish(me().aim);
		}
	}

	int hydrantStayTime = 0;
	int refugeStayTime = 0;

	private void checkWater() throws ActionCommandException {
		Area targetRefuge = null;
		Area target = null;
		Area targetHydrant = null;
		if (me().getWater() == 0)
			goingToRefuge = true;
		if (me().getWater() >= maxWater)
			goingToRefuge = false;
		if (goingToRefuge) {
			if ((me().zone != null && me().zone.hibitor)
					|| worldModel.reachableRefuges.size() == 0) {
				if (worldModel.reachableHydrants != null
						&& worldModel.reachableHydrants.size() != 0) {
					for (int i = 0; i < worldModel.reachableHydrants.size(); i++) {
						for (FireBrigade f : worldModel.fireBrigades) {
							if (f.getPosition().getValue() == ((Area) worldModel.reachableHydrants
									.get(i)).getID().getValue()
									&& f.getID().getValue() != me().getID()
											.getValue()
									&& me().getID().getValue() > f.getID()
											.getValue()) {
								worldModel.reachableHydrants
										.remove(worldModel.reachableHydrants
												.get(i));
								i--;
								break;
							}
						}
					}
					if (worldModel.reachableHydrants.size() != 0)
						targetHydrant = worldModel.reachableHydrants.get(0);
				}
			}
			if (worldModel.reachableRefuges.size() != 0) {
				targetRefuge = worldModel.reachableRefuges.get(0);
			}

			if (targetRefuge != null && targetHydrant != null)
				if (targetHydrant.distanceFromSelf < targetRefuge.distanceFromSelf * 0.7)
					target = targetHydrant;
				else
					target = targetRefuge;
			if (targetRefuge == null)
				target = targetHydrant;
			if (targetHydrant == null)
				target = targetRefuge;
			if (target == null)
				search();
			/*
			 * if (target instanceof Refuge && me().getWater() >= 30 *
			 * refillRateRefuge) { goingToRefuge = false; return; } else if
			 * (me().getWater() >= 10 * refillRateHydrant) { goingToRefuge =
			 * false; return; }
			 */
			if ((getMyPos() instanceof Hydrant && hydrantStayTime == 10)
					|| ((getMyPos() instanceof Refuge && refugeStayTime == 30))) {
				hydrantStayTime = 0;
				refugeStayTime = 0;
				goingToRefuge = false;
				return;
			} else if ((getMyPos() instanceof Hydrant && hydrantStayTime < 10)
					|| (getMyPos() instanceof Refuge && refugeStayTime < 30)) {
				hydrantStayTime++;
				refugeStayTime++;
			}
			move(target.getID());
		}
		refugeStayTime = 0;
		hydrantStayTime = 0;

	}

	private void checkForSearchAround() throws ActionCommandException {
		if (lastZone != null && lastZone.buildings.get(0).fireZoneNum == -1) {
			if (!amISearching)
				setFireBrigadeForSearch();
			for (EntityID id : worldModel.changeSet.getChangedEntities())
				if (worldModel.getEntity(id) instanceof rescuecore2.standard.entities.Building) {
					toBeSearched.remove(worldModel.getArea(id));
					Building b = (Building) worldModel.getArea(id);
					if (b.stFire > 0) {
						amISearching = false;
						break;
					}
				}

			if (amISearching) {
				for (Building building : toBeSearched)
					if (building.stFire > 0) {
						amISearching = false;
						break;
					}
			}
			if (amISearching) {
				Building searchTarget = searchAroundZone();
				if (searchTarget != null) {
					moveToFireAndSee(searchTarget);
				}
			}
		}
		amISearching = false;
		lastZone = null;
	}

	private void noHear() throws ActionCommandException {
		if (me().aim != null && me().aim.stFire > 0
				&& buildingReachability(me().aim)
				&& allArounds.contains(me().aim)) {
			if (lowCommi) {
				emergencyExtinguish();
			}
			moveAndExtinguish(me().aim);
		}
		if (me().aim != null && me().aim.isOnFire()
				&& !buildingReachability(me().aim)
				&& allArounds.contains(me().aim)
				&& hasBeenRecentlySeen(me().aim)) {

			if (canExtinguish(me().aim) && me().aim.stFire > 0 && !IamOnfire) {

				extinguish(me().aim.getID(), maxPower);
			} else {
				ArrayList<EntityID> unEmploymentSituationPath = new ArrayList<EntityID>();
				unEmploymentSituationPath = areaToEntityID(unEmploymentSituationPath(me().aim));
				if (unEmploymentSituationPath != null
						&& unEmploymentSituationPath.size() != 0) {
					Area unemploymentTarget = worldModel
							.getArea(unEmploymentSituationPath
									.get(unEmploymentSituationPath.size() - 1));
					double dist = Math.hypot(
							(unemploymentTarget.getX() - me().aim.getX()),
							(unemploymentTarget.getY() - me().aim.getY()));
					if (lowCommi)
						emergencyExtinguish();
					if (dist <= maxDistanceExtinguish) {
						move(unEmploymentSituationPath,
								unemploymentTarget.getX(),
								unemploymentTarget.getY(), false);
					}
				}
			}
		}

		if (me().aim != null) {
			Building target = null;
			int minDis = Integer.MAX_VALUE;
			int minDisForUnEmployment = Integer.MAX_VALUE;
			if (me().aim.fireZoneNum != -1) {
				ArrayList<Area> unemploymentSituationPath = new ArrayList<Area>();
				Area unEmploymentTarget = null;
				ArrayList<Area> path = new ArrayList<Area>();
				for (Building build : fireZoneBuilder.modelZones
						.get(me().aim.fireZoneNum).simulatedAroundsWithSetGood) {
					if (buildingReachability(build)) {
						int dist = getDistToFire(build);
						if (dist < minDis) {
							minDis = dist;
							target = build;
						}
					} else if (build.isOnFire() && hasBeenRecentlySeen(build)) {
						path = new ArrayList<Area>();
						path = unEmploymentSituationPath(build);
						if (path != null
								&& (path.size() != 0 || canExtinguish(build))
								&& path.size() < minDisForUnEmployment) {
							minDisForUnEmployment = path.size();
							unEmploymentTarget = build;
							unemploymentSituationPath = path;
						}
					}
				}
				if (target != null) {
					moveAndExtinguish(target);
				} else if (unEmploymentTarget != null) {
					checkEmergencySituation();
					checkandExtinguish((Building) unEmploymentTarget);
					if (unemploymentSituationPath != null
							&& unemploymentSituationPath.size() >= 1) {
						Area way = unemploymentSituationPath
								.get((unemploymentSituationPath.size() - 1));
						// move(way);
						move(areaToEntityID(unemploymentSituationPath),
								way.getX(), way.getY(), false);
					}
				}
			} else {
				minDis = Integer.MAX_VALUE;
				for (Building build : allAroundsWithsetGood)
					if (buildingReachability(build)) {
						int dist = getDistToFire(build);
						if (dist < minDis) {
							minDis = dist;
							target = build;
						}
					}
				if (target != null) {
					moveAndExtinguish(target);
				}
			}
		}

		if (fireZoneBuilder.modelZones.size() != 0) {
			fbNoHear.act();
		}
		search();
	}

	@SuppressWarnings("unused")
	private Building chooseMinLtsAndNearest(ArrayList<Building> buildings) {
		double minDist = Integer.MAX_VALUE;
		Building nearestBuilding = null;

		int minlts = Integer.MAX_VALUE;
		for (Building b : buildings)
			if (b.getLTS() < minlts)
				minlts = b.getLTS();

		ArrayList<Building> minltsBuildings = new ArrayList<Building>();
		for (Building b : buildings)
			if (b.getLTS() == minlts)
				minltsBuildings.add(b);

		for (Building b : minltsBuildings) {
			if (b.distanceFromSelf < minDist && buildingReachability(b)) {
				minDist = b.distanceFromSelf;
				nearestBuilding = b;
			}
		}

		return nearestBuilding;
	}

	private void setZoneHibitors() {
		for (FireZone fireZone : fireZoneBuilder.modelZones)
			if (fireZone.ignitedBuilds.size() < 10)
				fireZone.hibitor = true;
	}

	Building lastSearchTarget = null;

	private Building ChoosNearsetToLastTargetExtenguished(
			ArrayList<Building> toSearch) {
		Building Target = null;
		Building lastTarget = null;
		if (extinguishedTarg != null
				&& lastZone.buildings.contains((Building) (worldModel
						.getArea(extinguishedTarg)))) {
			lastTarget = (Building) (worldModel.getArea(extinguishedTarg));
		} else {
			double min = Integer.MAX_VALUE;
			for (EntityID id : worldModel.changeSet.getChangedEntities()) {
				if (((worldModel.getEntity(id) instanceof rescuecore2.standard.entities.Building))) {
					Building building = (Building) worldModel.getArea(id);
					if (building.fireness() > 3 && building.fireness() < 8) {
						if (building.distanceFromSelf < min
								&& buildingReachability(building)) {
							min = building.distanceFromSelf;
							lastTarget = building;
						}
					}
				}

			}
		}
		if (lastTarget != null) {
			double minDist = Integer.MAX_VALUE;
			for (Building build : toSearch) {
				if (Math.hypot(build.getX() - lastTarget.getX(), build.getY()
						- lastTarget.getY()) < minDist) {
					Target = build;
					minDist = Math.hypot(build.getX() - lastTarget.getX(),
							build.getY() - lastTarget.getY());
				}
			}
		}
		return Target;

	}

	private Building searchAroundZone() {
		for (EntityID id : worldModel.changeSet.getChangedEntities())
			if (worldModel.getEntity(id) instanceof rescuecore2.standard.entities.Building
					&& (toBeSearched.contains(worldModel.getArea(id))))
				toBeSearched.remove(worldModel.getArea(id));

		HashSet<Building> notReachable = new HashSet<Building>();
		for (Building building : toBeSearched)
			if (!buildingReachability(building))
				notReachable.add(building);
		toBeSearched.removeAll(notReachable);

		if (lastSearchTarget == null
				|| !toBeSearched.contains(lastSearchTarget))
			lastSearchTarget = ChoosNearsetToLastTargetExtenguished(toBeSearched);
		return lastSearchTarget;
	}

	private double getZoneRad(FireZone fz) {
		double maxDist = -1;
		for (Building b : fz.simulatedArounds)
			for (Coordinate p : b.shape.getCoordinates()) {
				double dist = Math.hypot(p.x - fz.centerX, p.y - fz.centerY);
				if (maxDist < dist)
					maxDist = dist;
			}

		return maxDist;
	}

	private void setFireBrigadeForSearch() {
		HashSet<FireBrigade> nearFireBrigades = new HashSet<FireBrigade>();
		double zoneCenterToAround = getZoneRad(lastZone)
				+ worldModel.maxDistanceView;
		for (Human h : hearAGPosition.keySet())
			if (h instanceof FireBrigade && !h.hasBuriedness)
				if (Math.hypot(h.getX() - lastZone.centerX, h.getY()
						- lastZone.centerY) < zoneCenterToAround)
					nearFireBrigades.add((FireBrigade) h);

		// String s = new String();
		// for (FireBrigade fb : nearFireBrigades)
		// s = s + fb.getID().getValue() + " ";
		// ("nearFire " + s);

		for (int i = 0; i < 2; i++) {
			int id = Integer.MAX_VALUE;
			FireBrigade fire = null;
			for (FireBrigade f : nearFireBrigades)
				if (f.getID().getValue() < id) {
					id = f.getID().getValue();
					fire = f;
				}

			if (fire != null
					&& me().getID().getValue() == fire.getID().getValue()) {
				// log("daram amISearchingo true mikonam");
				amISearching = true;
				break;
			}

			nearFireBrigades.remove(fire);
		}
	}

	@SuppressWarnings("unchecked")
	void beforeExt(Building target) {
		lastSearchTarget = null;
		lastZone = fireZoneBuilder.modelZones.get(target.fireZoneNum);
		toBeSearched = (ArrayList<Building>) lastZone.buildings.clone();
		for (Building b : lastZone.buildings)
			for (Building bu : b.nearBuildings)
				if (!toBeSearched.contains(bu) && bu.getFieryness() != 8)
					toBeSearched.add(bu);

		ArrayList<Building> toBeRemoved = new ArrayList<Building>();
		for (Building building : toBeSearched)
			if (!building.isIgnitable())
				toBeRemoved.add(building);

		toBeSearched.removeAll(toBeRemoved);
	}

	public void moveAndExtinguish(Building fireTarget)
			throws ActionCommandException {
		if (!noCommi)
			beforeExt(fireTarget);
		if (worldModel.isInChangeSet(fireTarget.getID())
				&& canExtinguish(fireTarget) && !IamOnfire) {
			if (noCommi)
				beforeExt(fireTarget);
			extinguish(fireTarget.getID(), maxPower);
		}
		moveToFire(fireTarget);
	}

	protected boolean canExtinguish(worldModel.Building target) {
		return Mathematic.getDistance(target.getX(), target.getY(), getX(),
				getY()) < maxDistanceExtinguish;
	}

	int counterForEmergencySit = 0;

	private void checkEmergencySituation() throws ActionCommandException {
		if (!amIStucking || me().getWater() == 0)
			return;
		if (isInEmergencySituation && counterForEmergencySit % 10 == 0)
			rw = 2;
		if (rw < 3) {
			rw++;
			return;
		}
		counterForEmergencySit++;
		isInEmergencySituation = true;
		if (worldModel.reachableAreas.size() <= 6) {
			for (FireZone fz : fireZoneBuilder.modelZones)
				for (Building b : fz.simulatedAroundsWithSetGood) {
					if (b.isOnFire() && canExtinguish(b))
						extinguish(b.getID(), maxPower);
				}

			for (FireZone fz : fireZoneBuilder.modelZones)
				for (Building b : fz.simulatedAroundsWithSetGood) {
					if (b.stFire > 0 && canExtinguish(b))
						extinguish(b.getID(), maxPower);
				}

			for (Building build : worldModel.buildings)
				if (build.stFire > 0 && canExtinguish(build)) {
					extinguish(build.getID(), maxPower);
				}
		}
	}

	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}
}
