package agent;

import java.util.ArrayList;

import agent.Agent.MapType;
import utilities.Logger;
import worldModel.Area;
import worldModel.Building;
import worldModel.Road;
import worldModel.WorldModel;

public class FireZoneBuilder {

	public ArrayList<FireZone> modelZones = new ArrayList<FireZone>();
	private WorldModel worldModel = null;
	private ArrayList<Building> ignitedBuildings = new ArrayList<Building>();
	private FireZone zone = null;
	private int groundMidst = 0;
	private int numMidst = 0;
	private boolean bigMap = true;
	private boolean noCommi = false;
	private double averageOfSurfaces = 0;
	ArrayList<FireZone> lastModelZones = new ArrayList<FireZone>();
	public ArrayList<Building> reachableAndIgnitedBuildings = new ArrayList<Building>();
	Logger logger = null;
	MapType map = MapType.Unknown;

	public FireZoneBuilder(WorldModel wm, boolean bm, double aos,
			Logger logger, boolean noCommi) {
		worldModel = wm;
		bigMap = bm;
		setMidsts();
		averageOfSurfaces = aos;
		this.logger = logger;
		this.noCommi = noCommi;
		this.map = worldModel.map;
	}

	public void executeModelZones(FireBrigadeAgent fireBrigade) {
		lastModelZones.clear();
		lastModelZones.addAll(modelZones);

		initialBuildings(fireBrigade);
		zoning(fireBrigade);
		setAroundForFireZone();
		zoneMatching();

		for (FireZone fz : modelZones) {
			fz.setGoodArounds(worldModel, numMidst, groundMidst);
		}

		for (FireZone fz : modelZones) {
			fz.setZoneCenter();
			fz.setZoneSurface(averageOfSurfaces);
		}

		// defining real arounds ...
		if (fireBrigade == null) {
			for (FireZone fireZone : modelZones) {
				FireZone realFireZone = new FireZone();
				for (Building building : fireZone.ignitedBuilds)
					if (!building.isSimulatedForFire)
						realFireZone.ignitedBuilds.add(building);
				realFireZone.reconstructedBuildings();
				realFireZone.setGoodArounds(worldModel, numMidst, groundMidst);

				fireZone.realArounds.addAll(realFireZone.simulatedArounds);
				fireZone.realAroundsWithSetGood
						.addAll(realFireZone.simulatedAroundsWithSetGood);
				fireZone.allArounds.addAll(fireZone.simulatedArounds);
				fireZone.allArounds.addAll(fireZone.realArounds);
				fireZone.allAroundsWithSetGood
						.addAll(fireZone.simulatedAroundsWithSetGood);
				fireZone.allAroundsWithSetGood
						.addAll(fireZone.realAroundsWithSetGood);
			}
		}
	}

	private void setMidsts() {
		int n = 0;
		int s = 0;
		Building max = worldModel.buildings.get(0);
		for (Building bui : worldModel.buildings) {
			for (int i : bui.areasInRange) {
				Area e = worldModel.areas.get(i);
				Building bu = null;
				if (e instanceof Building) {
					bu = ((Building) e);
					n++;
					s += bu.getGroundArea();
					if (bu.getGroundArea() > max.getGroundArea())
						max = bu;
				}
			}
		}
		groundMidst = s / (worldModel.buildings.size());
		numMidst = n / (worldModel.buildings.size());
	}

	private void initialBuildings(FireBrigadeAgent fireBrigade) {
		ignitedBuildings.clear();
		reachableAndIgnitedBuildings.clear();

		for (Building build : worldModel.buildings)
			if (build.stFire > 0)
				ignitedBuildings.add(build);

		if (fireBrigade != null)
			for (Building build : ignitedBuildings)
				if (fireBrigade.buildingReachability(build)) {
					reachableAndIgnitedBuildings.add(build);
				}
	}

	@SuppressWarnings("static-access")
	private void zoning(FireBrigadeAgent fireBrigade) {
		modelZones = new ArrayList<FireZone>();

		for (Building building : worldModel.buildings) {
			building.fireZoneNum = -1;
			building.haveZone = false;
		}

		ArrayList<Building> fireds = new ArrayList<Building>();
		if (fireBrigade == null)
			fireds.addAll(ignitedBuildings);
		else
			fireds.addAll(reachableAndIgnitedBuildings);

		for (Building building : fireds) {
			if (!building.haveZone) {
				zone = new FireZone(modelZones.size());
				checkNeighboursForZoning(building);
				modelZones.add(zone);
			}
		}

		if (fireBrigade != null) {
			fireBrigade.unreachableAndIgnitedBuildings = new ArrayList<Building>();
			for (Building building : worldModel.buildings) {
				if (building.stFire > 0
						&& fireBrigade.hasBeenRecentlySeen(building)
						&& !fireBrigade.buildingReachability(building)) {
					for (Area area : building.nearAreas50000) {
						if (area instanceof Road
								&& area.isReachable
								&& Math.hypot(area.getX() - building.getX(),
										area.getY() - building.getY()) < fireBrigade.maxDistanceExtinguish) {
							fireBrigade.unreachableAndIgnitedBuildings
									.add(building);
							break;
						}
					}
				}
			}

			for (Building isCheckingBuilding : fireBrigade.unreachableAndIgnitedBuildings) {
				if (!isCheckingBuilding.haveZone) {
					zone = new FireZone(modelZones.size());
					zone.emergency = true;
					checkNeighboursForZoning(isCheckingBuilding);
					modelZones.add(zone);
				}
			}
		}
	}

	private void checkNeighboursForZoning(Building building) {
		building.fireZoneNum = zone.fireZoneNum;
		building.haveZone = true;

		zone.buildings.add(building);

		if (building.stFire > 0)
			zone.ignitedBuilds.add(building);

		for (Building neighbour : building.nearBuildings)
			if (((neighbour.isFirenessDefined() && neighbour.getFieryness() != 0) || neighbour.stFire > 0)
					&& !neighbour.haveZone)
				checkNeighboursForZoning(neighbour);
	}

	private void setAroundForFireZone() {
		for (FireZone fz : modelZones)
			fz.reconstructedBuildings();
	}

	private void zoneMatching() {
		if (!bigMap && !noCommi)
			return;

		if (modelZones.size() <= 1)
			return;

		if (ignitedBuildings.size() > 250)
			return;

		while (true) {
			FireZone toBeRemoved = null;
			FireZone topZone = null;

			done: for (FireZone fireZone : modelZones) {
				for (Building building : fireZone.simulatedArounds) {
					for (int i : building.areasInRange) {
						Area area = worldModel.areas.get(i);
						if (area instanceof Building) {
							Building b = ((Building) area);
							if (b.stFire > 0 && b.haveZone
									&& b.fireZoneNum != building.fireZoneNum) {
								toBeRemoved = modelZones.get(b.fireZoneNum);
								zoneMatched(fireZone, toBeRemoved);
								topZone = fireZone;
								break done;
							}
						}
					}
				}
			}

			if (toBeRemoved == null)
				break;

			topZone.reconstructedBuildings();
			modelZones.remove(toBeRemoved);

			int i = 0;
			for (FireZone fz : modelZones) {
				fz.fireZoneNum = i;
				for (Building b : fz.buildings)
					b.fireZoneNum = i;
				i++;
			}
		}
	}

	private void zoneMatched(FireZone asli, FireZone farE) {
		for (Building build : farE.buildings)
			build.fireZoneNum = asli.fireZoneNum;
		asli.buildings.addAll(farE.buildings);
		asli.ignitedBuilds.addAll(farE.ignitedBuilds);
	}

	public boolean zoneChange() {
		if (lastModelZones.size() == 0 || modelZones.size() == 0)
			return true;

		if (lastModelZones.size() != modelZones.size())
			return true;

		if (lastModelZones.size() == modelZones.size())
			for (FireZone last : lastModelZones) {
				for (FireZone model : modelZones) {
					if (model.buildings.contains(last.buildings.get(0))
							&& model.hibitor != last.hibitor)
						return true;
				}
			}
		for (FireZone fireZone : lastModelZones)
			if (fireZone.fireAgents.size() != 0
					&& fireZone.buildings.get(0).fireZoneNum != -1
					&& modelZones.get(fireZone.buildings.get(0).fireZoneNum).fireAgents
							.size() == 0)
				return true;
		// for (FireZone fireZone : lastModelZones) {
		// if (fireZone.fireAgents.size() != 0
		// && fireZone.buildings.get(0).fireZoneNum != -1
		// && modelZones.get(fireZone.buildings.get(0).fireZoneNum).hibitor !=
		// fireZone.hibitor)
		// return true;
		// }
		return false;
	}
	// public void log(String msg) {
	// logger.log(msg);
	// }
}
