/* 
 * KMeans.java ; Cluster.java ; Point.java
 *
 * Solution implemented by DataOnFocus
 * www.dataonfocus.com
 * 2015
 *
 */
package agent;

import java.util.ArrayList;

import worldModel.Area;
import worldModel.Building;
import worldModel.WorldModel;

public class KMeans {

	// Number of Clusters. This metric should be related to the number of points
	private int NUM_CLUSTERS = 3;
	// Number of Points
	// Min and Max X and Y
	private int minX;
	private int minY;
	private int maxX;
	private int maxY;

	private utilities.Logger logger;
	private ArrayList<Point> points;
	private ArrayList<Cluster> clusters;
	WorldModel worldModel;

	public KMeans(WorldModel worldModel, int numberOfClusters,
			utilities.Logger logger) {
		this.worldModel = worldModel;
		findMaxX();
		findMaxY();
		findMinX();
		findMinY();
		this.logger = logger;
		this.NUM_CLUSTERS = numberOfClusters;
		this.points = new ArrayList<Point>();
		this.clusters = new ArrayList<Cluster>();
	}

	public void log(String msg) {
		// logger.log(msg);
	}

	private void findMinX() {
		minX = Integer.MAX_VALUE;
		for (Area a : worldModel.buildings)
			if (minX > a.getX())
				minX = a.getX();
	}

	private void findMinY() {
		minY = Integer.MAX_VALUE;
		for (Area a : worldModel.buildings)
			if (minY > a.getY())
				minY = a.getY();
	}

	private void findMaxX() {
		maxX = Integer.MIN_VALUE;
		for (Area a : worldModel.buildings)
			if (maxX < a.getX())
				maxX = a.getX();
	}

	private void findMaxY() {
		maxY = Integer.MIN_VALUE;
		for (Area a : worldModel.buildings)
			if (maxY < a.getY())
				maxY = a.getY();
	}

	// Initializes the process
	private void setPoints() {
		for (Building b : worldModel.buildings) {
			Point p = new Point(b.getX(), b.getY());
			if (!points.contains(p))
				points.add(p);
			p.building = b;

		}
	}

	ArrayList<Point> centroids = new ArrayList<Point>();

	public void init() {
		// Create Points
		setPoints();

		// Create Clusters
		// Set Random Centroids
		ArrayList<Point> centerPoint = findPoints();
		for (int i = 0; i < NUM_CLUSTERS; i++) {
			Cluster cluster = new Cluster(i);
			Point centroid = centerPoint.get(i);
			cluster.setCentroid(centroid);
			centroids.add(cluster.getCentroid());
			clusters.add(cluster);
		}
		// Print Initial state
		// plotClusters();
	}

	private ArrayList<Point> findPoints() {
		int maxX = Integer.MIN_VALUE, maxY = Integer.MIN_VALUE, minX = Integer.MAX_VALUE, minY = Integer.MAX_VALUE;
		for (Area a : worldModel.areas) {
			if (a.getX() > maxX)
				maxX = a.getX();
			if (a.getY() > maxY)
				maxY = a.getY();
			if (a.getX() < minX)
				minX = a.getX();
			if (a.getY() < minY)
				minY = a.getY();
		}
		int num = NUM_CLUSTERS;
		int i = 1;
		ArrayList<Point> points = new ArrayList<Point>();
		ArrayList<SectorZone> tempClusters = new ArrayList<SectorZone>();
		while (tempClusters.size() < num) {
			tempClusters.clear();
			int partNum = (int) Math.sqrt(num) + i;
			int xPart = Math.round((maxX - minX) / partNum);
			int yPart = Math.round((maxY - minY) / partNum);
			yektaZone(xPart, yPart, tempClusters);
			i++;
		}
		while (tempClusters.size() != num) {
			int min = Integer.MAX_VALUE;
			SectorZone hadafCluster = null;
			for (SectorZone sz : tempClusters)
				if (sz.buildings.size() < min) {
					min = sz.buildings.size();
					hadafCluster = sz;
				}
			tempClusters.remove(hadafCluster);
		}
		for (SectorZone sz : tempClusters)
			points.add(new Point(sz.center.x, sz.center.y));
		return points;
	}

	public void yektaZone(int xPart, int yPart,
			ArrayList<SectorZone> tempClusters) {
		int index = 0;
		for (int i = minX; i <= maxX; i += xPart) {
			if (index % 2 == 0)
				for (int j = minY; j <= maxY; j += yPart) {
					SectorZone cl = new SectorZone();
					for (Building aar : worldModel.buildings) {
						if (aar.getX() >= i && aar.getX() <= i + xPart
								&& aar.getY() >= j && aar.getY() <= j + yPart) {
							cl.buildings.add(aar);
						}
					}
					if (cl.buildings.size() > 0) {
						cl.setCenter();
						tempClusters.add(cl);
					}
				}
			if (index % 2 == 1)
				for (int j = maxY; j >= minY; j -= yPart) {
					SectorZone cl = new SectorZone();
					for (Building aar : worldModel.buildings) {
						if (aar.getX() >= i && aar.getX() <= i + xPart
								&& aar.getY() >= j && aar.getY() <= j + yPart) {
							cl.buildings.add(aar);
						}
					}
					if (cl.buildings.size() > 0) {
						cl.setCenter();
						tempClusters.add(cl);
					}
				}
			index++;
		}
	}

	@SuppressWarnings("unused")
	private void plotClusters() {
		for (int i = 0; i < NUM_CLUSTERS; i++) {
			Cluster c = clusters.get(i);
			c.plotCluster();
		}
	}

	// The process to calculate the K Means, with iterating method.
	public ArrayList<Cluster> calculate() {
		boolean finish = false;
		// int iteration = 0;

		// Add in new data, one at a time, recalculating centroids with each new
		// one.
		while (!finish) {
			// Clear cluster state
			clearClusters();

			ArrayList<Point> lastCentroids = getCentroids();

			// Assign points to the closer cluster
			assignCluster();

			// Calculate new centroids.
			calculateCentroids();
			// iteration++;
			// log("iteration: " + iteration);
			ArrayList<Point> currentCentroids = getCentroids();

			// Calculates total distance between new and old Centroids
			int distance = 0;
			// log("lastCentroid.size: " + lastCentroids.size());
			for (int i = 0; i < lastCentroids.size(); i++) {
				distance += Point.distance(lastCentroids.get(i),
						currentCentroids.get(i));

			}
			// System.out.println("#################");
			// System.out.println("Iteration: " + iteration);
			// System.out.println("Centroid distances: " + distance);
			// plotClusters();
			// log("distance: " + distance);
			if (distance == 0)
				finish = true;

		}
		// String s = "";
		// for (Cluster c : clusters) {
		// for (Point p : c.points) {
		// s += p.building.id + " ";
		// }
		// s += ";";
		// }
		return clusters;
	}

	private void clearClusters() {
		for (Cluster cluster : clusters) {
			cluster.clear();
		}
	}

	private ArrayList<Point> getCentroids() {
		ArrayList<Point> centroids = new ArrayList<Point>(NUM_CLUSTERS);
		for (Cluster cluster : clusters) {
			Point aux = cluster.getCentroid();
			Point point = new Point(aux.getX(), aux.getY());
			centroids.add(point);
		}
		return centroids;
	}

	private void assignCluster() {
		int cluster = 0;
		int distance;
		for (Point point : points) {
			int min = Integer.MAX_VALUE;
			for (int i = 0; i < NUM_CLUSTERS; i++) {
				Cluster c = clusters.get(i);
				distance = (int) Point.distance(point, c.getCentroid());
				if (distance < min) {
					min = distance;
					cluster = i;
				}
			}
			point.setCluster(cluster);
			clusters.get(cluster).addPoint(point);
		}
	}

	private void calculateCentroids() {
		for (Cluster cluster : clusters) {

			double sumX = 0;
			double sumY = 0;
			ArrayList<Point> list = cluster.getPoints();
			int n_points = list.size();

			for (Point point : list) {
				sumX += point.getX();
				sumY += point.getY();
			}

			Point centroid = cluster.getCentroid();
			if (n_points > 0) {
				double newX = sumX / n_points;
				double newY = sumY / n_points;
				centroid.setX(newX);
				centroid.setY(newY);
			}
		}
	}
}