package agent;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.TreeSet;

import radar.Calulations;
import radar.Package;
import radar.Package.AreaType;
import radar.Package.HeaderType;
import radar.PackageSchedule;
import rescuecore2.messages.Command;
import rescuecore2.registry.Registry;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import worldModel.Area;
import worldModel.Blockade;
import worldModel.Building;
import worldModel.Entrance;
import worldModel.GeometryTools;
import worldModel.Refuge;
import worldModel.Road;

import com.vividsolutions.jts.geom.Coordinate;

public abstract class RadarAgent<E extends StandardEntity> extends Agent<E> {
	public Boolean checkAllBlockades = false;
	int stuckedValue = 5;
	int SBNum = -1;
	public int channel = 1, otherChannel = 2, channel0Bytes = 0,
			agentChannelCount = 0, centerChannelCount = 0,
			channelByteCount = -1, otherChannelByteCount = -1;
	int maxAmountOfValue = 8;
	private int lastPosHeardTime = -1;
	int fireNum = 8;
	private int lastFiredsTimeLimit = 4;
	private int lastPacksTimeLimit = 2;
	private int beginOfFakeCVID = 0;
	private HashSet<Building> lastBuildings = new HashSet<>();

	public int counterForHelp = 0, counterForStucks = 0, counterForUnHelp = 0,
			counterForUnStuck = 0, timeLimitForBeingTrapped = 0;
	private int stuckLimit = -1, stuckIndex = 0;
	boolean shenide = false;

	protected ArrayList<Area> stuckedAreas = new ArrayList<Area>();
	private ArrayList<Package> packsForSay = new ArrayList<Package>();
	private HashSet<Civilian> CVsForSay = new HashSet<Civilian>();
	protected HashMap<Human, Area> hearAGPosition = null;
	public HashMap<Integer, Integer> trapped = new HashMap<Integer, Integer>();
	protected HashMap<Building, Integer> fireds = new HashMap<Building, Integer>();
	protected HashMap<Integer, Integer> helps = new HashMap<>();

	public HashMap<HeaderType, HashSet<Integer>> channels = new HashMap<>();
	public ArrayList<Integer> subscribes = new ArrayList<>();

	public ArrayList<Building> lastCycleHasFireds = new ArrayList<Building>();

	byte[] lastCyclesTasks = null;

	@Override
	protected void postConnect() {
		super.postConnect();
		selectChannel();
		setChannels();
		if (noCommi)
			timeForNewWayMemory = 30;
		if (lowCommi)
			timeForNewWayMemory = 50;
		if (!lowCommi && !noCommi)
			timeForNewWayMemory = 100;
		selectCenters();
		if (lowCommi)
			setByteTook();
	}

	private void selectCenters() {
		selectCenterForAmb();
		selectCenterForFire();
		selectCenterForPolice();
	}

	protected void process() throws ActionCommandException {
		try {
			radarWorks();
		} catch (Exception e) {
			System.err.println("int time " + time + " from agent "
					+ me().getID().getValue() + " in Radar!");
			e.printStackTrace();
			if (logger.getWriter() != null)
				e.printStackTrace(logger.getWriter());
		}

		ActionCommandException ace = null;
		try {
			decide();
		} catch (ActionCommandException ac) {
			ace = ac;
		}

		if (myHum() != null) {
			checkRechablePolicesAvailability();
		}

		sendPackageSchedules();
		if (ace != null)
			throw ace;
	}

	private void selectChannel() {
		int tedad = config.getIntValue("comms.channels.count");
		tedad = config.getIntValue("comms.channels.count");
		channel0Bytes = config.getIntValue("comms.channels.0.messages.size");
		agentChannelCount = config.getIntValue("comms.channels.max.platoon");
		centerChannelCount = config.getIntValue("comms.channels.max.centre");
		if (tedad <= 1) {
			noCommi = true;
			channel = 0;
		}
		if (!noCommi) {
			int tedadByteBehtarinChannal1 = 0;
			int tedadByteBehtarinChannal2 = 0;
			if (tedad == 2) {
				channel = 1;
				otherChannel = 0;
			} else {
				for (int i = 2; i < tedad; i++) {
					int currentChannelByte = config
							.getIntValue("comms.channels." + i + ".bandwidth");
					tedadByteBehtarinChannal1 = config
							.getIntValue("comms.channels." + channel
									+ ".bandwidth");
					tedadByteBehtarinChannal2 = config
							.getIntValue("comms.channels." + otherChannel
									+ ".bandwidth");
					if (currentChannelByte > tedadByteBehtarinChannal1) {
						otherChannel = channel;
						channel = i;
					} else if (currentChannelByte >= tedadByteBehtarinChannal2)
						otherChannel = i;
				}
			}
			channelByteCount = config.getIntValue("comms.channels." + channel
					+ ".bandwidth");
			if (otherChannel > 0)
				otherChannelByteCount = config.getIntValue("comms.channels."
						+ otherChannel + ".bandwidth");
			else
				otherChannelByteCount = channel0Bytes;

			if (channelByteCount < 64)
				noCommi = true;
			else if (channelByteCount <= 256)
				lowCommi = true;
			else if (channelByteCount > 512)
				sendAll = true;
		}
	}

	protected void setChannels() {
		if (channel != 0 && !subscribes.contains(channel))
			subscribes.add(channel);
		for (int i = 0; i < HeaderType.values().length; i++) {
			channels.put(HeaderType.values()[i], new HashSet<Integer>());
			channels.get(HeaderType.values()[i]).add(0);
		}
		if (channelByteCount >= 800) {
			for (int i = 0; i < HeaderType.values().length; i++)
				channels.get(HeaderType.values()[i]).add(channel);
			return;
		} else if (channelByteCount > 210) {
			if (agentChannelCount > 1 && centerChannelCount > 1
					&& otherChannel != 0) {
				if (!subscribes.contains(otherChannel))
					subscribes.add(otherChannel);
				channels.get(HeaderType.FCenter).add(otherChannel);
				channels.get(HeaderType.ACenter).add(otherChannel);
				channels.get(HeaderType.PCenter).add(otherChannel);
				for (int i = 0; i < HeaderType.values().length; i++)
					if (HeaderType.values()[i] != HeaderType.PCenter
							&& HeaderType.values()[i] != HeaderType.FCenter
							&& HeaderType.values()[i] != HeaderType.ACenter)
						channels.get(HeaderType.values()[i]).add(channel);
			} else
				for (int i = 0; i < HeaderType.values().length; i++)
					channels.get(HeaderType.values()[i]).add(channel);
		} else {
			if (agentChannelCount > 1 && centerChannelCount > 1
					&& otherChannel != 0) {
				if (!subscribes.contains(otherChannel))
					subscribes.add(otherChannel);
				channels.get(HeaderType.Way).add(otherChannel);
				channels.get(HeaderType.NoWay).add(otherChannel);
				channels.get(HeaderType.Clearer).add(otherChannel);
				for (int i = 0; i < HeaderType.values().length; i++)
					if (HeaderType.values()[i] != HeaderType.Way
							&& HeaderType.values()[i] != HeaderType.NoWay
							&& HeaderType.values()[i] != HeaderType.Clearer)
						channels.get(HeaderType.values()[i]).add(channel);
			} else {
				for (int i = 0; i < HeaderType.values().length; i++)
					// if (HeaderType.values()[i] != HeaderType.PCentre
					// && HeaderType.values()[i] != HeaderType.FCentre
					// && HeaderType.values()[i] != HeaderType.ACentre)
					channels.get(HeaderType.values()[i]).add(channel);
			}
		}
	}

	protected void preliminaries() {
		if (hearAGPosition == null) {
			hearAGPosition = new HashMap<Human, Area>();
			for (Human h : worldModel.agents)
				hearAGPosition.put(h, worldModel.getArea(h.getPosition()));
		}
		if (time == 1 && noCommi)
			fillingLongsForNoCommi();
		if (noCommi)
			for (Human h : worldModel.agents)
				setIsOK(h);
		// if (lowCommi)
		// lastFiredsTimeLimit = 3;
		if (lowCommi)
			lastPacksTimeLimit = 4;
		shenide = false;
		setSTFire();
	}

	protected void fillingLongsForNoCommi() {
		// log("in filling");
		if (me() instanceof PoliceForce)
			for (Human hum : worldModel.agents) {
				if (!(hum instanceof PoliceForce)
						|| worldModel.getArea(hum.getPosition()) instanceof Building)
					trapped.put(hum.getID().getValue(), hum.getPosition()
							.getValue());
			}
		// else if (me() instanceof AmbulanceTeam)
		// for (Human hum : worldModel.agents) {
		// if (worldModel.getArea(hum.getPosition()) instanceof Building) {
		// if (!(worldModel.getArea(hum.getPosition()) instanceof Refuge)) {
		// // log("hum  :  " + hum.getID().getValue());
		// hum.deadtime = 502;
		// hum.setBuriedness(2);
		// hum.hasBuriedness = true;
		// // log("true shod!!");
		// }
		// }
		// }
	}

	protected void subscribe(int time) {
		int[] subToArray = new int[subscribes.size()];
		for (int i = 0; i < subscribes.size(); i++)
			subToArray[i] = subscribes.get(i);
		sendSubscribe(time, subToArray);
	}

	public void radarWorks() throws Exception {
		for (Human h : worldModel.agents)
			h.isOK = true;
		preliminaries();
		hears();
		for (Entry<Building, Integer> t : lastSeenBy.entrySet()) {
			if (trapped.keySet().contains(t.getValue()))
				if (t.getKey().getHLTS() < time - 1 && !noCommi)
					t.getKey().setHLTS(time - 1, me().getID().getValue(),
							lastSeenBy, false);

		}
		// if (time == 8)
		// decideNoisyMode();
		worldModel.updateReachableAreas();
		subscribe(time);
		if (time > 2)
			sends();
	}

	boolean sendAll = false;

	// boolean saySeprate = false;
	// int cyclesToSay = 0;
	// private void findAliveSituations() {
	// if (channelByteCount > 500) {
	// sendAll = true;
	// cyclesToSay = 6;
	// }
	// if (channelByteCount < worldModel.agents.size()) {
	// saySeprate = true;
	// cyclesToSay = 2;
	// }
	// }
	// protected void firstStuffs() {
	// findAliveSituations();
	// }

	protected void sendHelloWorld() {
		if (myHum() == null)
			return;
		Package p = new Package(worldModel);
		if (sendAll)
			p.addHeader(HeaderType.Pos);

		if (!myHum().isBuriednessDefined())
			p.addInt(0, 128);
		else if (myHum().getBuriedness() > 127)
			p.addInt(127, 128);
		else
			p.addInt(myHum().getBuriedness(), 128);
		speak(time, HeaderType.Pos, p.send());
		if (channel != 0)
			speak(time, HeaderType.Pos, p.send());
	}

	protected void hearHelloWorld(Package p, EntityID agentID,
			boolean isSelfHeard) {
		int bNum = p.getInt(128);
		Human h = ((Human) worldModel.getEntity(agentID));
		h.isOK = true;
		if (worldModel.isInChangeSet(agentID) || isSelfHeard)
			return;
		h.setBuriedness(bNum);
		if (bNum == 0)
			h.hasBuriedness = false;
		else
			h.hasBuriedness = true;
	}

	@SuppressWarnings("unused")
	private void sendNoisyMode() {
		int agentsCount = worldModel.agents.size()
				+ worldModel.ambulanceCenters.size()
				+ worldModel.fireStations.size()
				+ worldModel.policeOffices.size();
		Package p = new Package(worldModel);

		int numToSayChannel = channelByteCount / agentsCount;
		if (numToSayChannel == 0 && me() instanceof Human
				&& worldModel.agents.indexOf(myHum()) < channelByteCount)
			sendSpeak(time, channel, p.send());
		else
			for (int i = 0; i < numToSayChannel; i++)
				sendSpeak(time, channel, p.send());

		int numToSayOtherChannel = otherChannelByteCount / agentsCount;
		if (numToSayOtherChannel == 0 && me() instanceof Human
				&& worldModel.agents.indexOf(myHum()) < otherChannelByteCount)
			sendSpeak(time, channel, p.send());
		else
			for (int i = 0; i < numToSayOtherChannel; i++)
				sendSpeak(time, otherChannel, p.send());
	}

	int numOfHeardNoisyModesChannel = 0;
	int numOfHeardNoisyModesOtherChannel = 0;

	@SuppressWarnings("unused")
	private void decideNoisyMode() {
		double noisyChannelPercentage = (double) numOfHeardNoisyModesChannel
				/ (double) channelByteCount;
		double noisyOtherChannelPercentage = (double) numOfHeardNoisyModesOtherChannel
				/ (double) otherChannelByteCount;
		// System.err.println("noisy channel percentage: "
		// + noisyChannelPercentage);
		// System.err.println("noisy other percentage: "
		// + noisyOtherChannelPercentage);
		// System.exit(0);
	}

	protected void sends() {
		// TODO handel kardane age kase dg e mige in sycle man nagam masalan to
		// fire & cv...age to change set mardom bood man nagam - be inke momkene
		// noise dame dahane man biyofte ya oon yeki degat she!

		// TODO Noisy Mode
		// if (time == 7 && !noCommi)
		// sendNoisyMode();

		if (myHum() != null) {
			checkBurriedness();
		}
		// TODO time>=10 ?!
		if (!noCommi && (time >= 10 || sendAll))
			sendLastPacks();
		if (!noCommi)
			sendFiredBuildings();
		if (!lowCommi && !noCommi)
			sendCVl();
		if (lowCommi)
			sendCVlForLowCommi();
		if (time > 10) {
			// TODO no && low doros she!
			if (noCommi)
				sendCVlForNoCommi();
			if (!lowCommi && !noCommi)
				sendWays();
			if (!lowCommi && !noCommi)
				clearers();
			if (myHum() != null && !noCommi)
				sendMyPos();
		} else if (time == 9 && me() instanceof PoliceForce && !lowCommi
				&& !noCommi)
			sendMyPos();
		if (noCommi) {
			setSTFire();
		}
	}

	HashSet<Building> lastCycleSawFireds = new HashSet<Building>();

	// int r = 0;
	protected void sendFiredBuildings() {
		HashSet<Building> hasFireds = new HashSet<Building>();
		for (EntityID eID : worldModel.changeSet.getChangedEntities()) {
			Entity e = worldModel.getEntity(eID);
			if (e instanceof rescuecore2.standard.entities.Building) {
				Building build = (Building) ((rescuecore2.standard.entities.Area) e).area;
				int rf = calculatedRF(build);
				if (build.shouldSend
						|| ((rf > 0 || rf != build.lastFireness) && (!fireds
								.containsKey(build) || fireds.get(build) != rf))) {
					hasFireds.add(build);
				}
			}
		}

		HashSet<Building> notInNears = new HashSet<Building>();
		HashSet<Building> thisAreaNears = new HashSet<Building>();

		if (lowCommi) {
			if (time % 7 == 4 || time % 7 == 5) {
				lastCycleHasFireds.addAll(hasFireds);
				return;
			}
			for (Building building : lastCycleHasFireds)
				if (!hasFireds.contains(building))
					hasFireds.add(building);
			lastCycleHasFireds.clear();
		}

		for (Building building : lastCycleSawFireds)
			if (!fireds.containsKey(building))
				hasFireds.add(building);
		lastCycleSawFireds.clear();
		lastCycleSawFireds.addAll(hasFireds);

		for (Building building : hasFireds) {
			building.saidTime = time;
			if (getMyPos().nearBuildingsIncludeSelf.contains(building)) {
				if (!thisAreaNears.contains(building)) {
					thisAreaNears.add(building);
				}
			} else {
				if (!notInNears.contains(building))
					notInNears.add(building);
			}
		}
		// notInNearsSize[time] += notInNears.size();

		if (thisAreaNears.size() > 0) {
			Package fired = new Package(worldModel);
			fired.addHeader(HeaderType.Fired);
			fired.addArea(getMyPos(), AreaType.Area);
			for (Building build : thisAreaNears) {
				addFiredBuildings(build, getMyPos(), fired);
			}
			fired.setEnd();
			speak(time, HeaderType.Fired, fired.send());
		}
		if (notInNears.size() > 0) {
			Package outFired = new Package(worldModel);
			outFired.addHeader(HeaderType.OutFired);
			for (Building b : notInNears) {
				outFired.addArea(b, AreaType.Building);
				outFired.addInt(calculatedRF(b), fireNum);
			}
			outFired.addInt(0, lastFiredsTimeLimit);
			outFired.setEnd();
			speak(time, HeaderType.OutFired, outFired.send());
		}
		// notInNearsSize[time] += notInNears.size();

		boolean sentOne = false;
		ArrayList<Building> lastBuildsToSay = new ArrayList<>();
		HashSet<Building> al = new HashSet<>();
		for (Building building : lastBuildings) {
			if (time - building.saidTime >= lastFiredsTimeLimit) {
				al.add(building);
				continue;
			}
			if ((!sentOne || building.stFire == 0)
					&& !hasFireds.contains(building)) {
				lastBuildsToSay.add(building);
				if (building.stFire == 1)
					sentOne = true;
			}
		}
		lastBuildings.removeAll(al);

		// lastBuildsToSaySize[time] += lastBuildsToSay.size();

		if (lastBuildsToSay.size() > 0) {
			Package outFired = new Package(worldModel);
			outFired.addHeader(HeaderType.OutFired);
			int minLTS = Integer.MAX_VALUE;
			for (Building building : lastBuildsToSay) {
				if (building.getLTS() < minLTS)
					minLTS = building.getLTS();
				outFired.addArea(building, AreaType.Building);
				outFired.addInt(calculatedRF(building), fireNum);
			}
			int t = time - minLTS;
			if (t > 3)
				t = 3;
			// if (time - minLTS > 3) {
			// System.err.println("Bad min LTS in send fire buildings! Time: "
			// + time + " Agent: " + me().getID().getValue());
			// }
			outFired.addInt(t, lastFiredsTimeLimit);
			outFired.setEnd();
			speak(time, HeaderType.OutFired, outFired.send());
		}
		lastBuildings.addAll(hasFireds);
	}

	private void addFiredBuildings(Building b, Area source, Package fired) {
		fired.addInt(source.nearBuildingsIncludeSelf.indexOf(b),
				source.nearBuildingsIncludeSelf.size());
		fired.addInt(calculatedRF(b), fireNum);
	}

	protected void hearFiredBuildings(Package fired, EntityID agentID,
			boolean isSelfHeard, Command cmd, int saidChannel, byte[] rec) {
		Area a = fired.getArea(AreaType.Area, worldModel);
		setAgentPos(a, agentID);
		int seenTime = time - 1;
		while (fired.hearingPointer < fired.findLastBit()) {
			Building b = (Building) a.nearBuildingsIncludeSelf.get(fired
					.getInt(a.nearBuildingsIncludeSelf.size()));
			int fireness = fired.getInt(fireNum);
			if (!worldModel.isInChangeSet(b.getID()) && seenTime >= b.getLTS()) {
				if (saidChannel != 0)
					fireds.put(b, fireness);
				if (!isSelfHeard)
					setFireness(b, fireness, seenTime, agentID.getValue());
			}
		}
	}

	protected void hearOutFiredBuildings(Package fired, boolean isSelfHeard,
			int saidChannel, EntityID agentID) {
		int seenTime = time - 1 - fired.getTime(lastFiredsTimeLimit);
		while (fired.hearingPointer < fired.findLastBit()
				- Calulations.findPow(lastFiredsTimeLimit)) {
			Building b = (Building) fired
					.getArea(AreaType.Building, worldModel);
			int fireness = fired.getInt(fireNum);
			if (!worldModel.isInChangeSet(b.getID()) && seenTime >= b.getLTS()) {
				if (saidChannel != 0)
					fireds.put(b, fireness);
				if (!isSelfHeard)
					setFireness(b, fireness, seenTime, agentID.getValue());
			}
		}
	}

	public void setFireness(Building b, int fireness, int seenTime, int agentID) {
		hearRF(b, fireness);
		b.setHLTS(seenTime, agentID, lastSeenBy);
		b.isSimulatedForFire = false;
	}

	private void sendNoCommiFireds(HashSet<Building> say, int LTSLimit,
			HeaderType ht) {
		Package says = new Package(worldModel);
		says.addHeader(ht);
		for (Building b : say) {
			says.addArea(b, AreaType.Building);
			says.addInt(time - b.getLTS(), LTSLimit);

		}
		says.setEnd();
		speak(time, ht, says.send());
	}

	private void hearNoCommiFireds(Package p, int stFire, int LTSLimit,
			int agentID) {
		while (p.hearingPointer < p.findLastBit()) {
			Building fired = (Building) p
					.getArea(AreaType.Building, worldModel);
			int lastTimeSeen = p.getInt(LTSLimit);
			if (fired.getLTS() <= time - 1 - lastTimeSeen/*
														 * fired.getHLTS() <=
														 * time - 1 -
														 * lastTimeSeen
														 */) {
				fired.setHLTS(time - 1 - lastTimeSeen, agentID, lastSeenBy);
				// if (fired.getLTS() <= time - 1 - lastTimeSeen)
				fired.stFire = stFire;
				if (stFire == 0) {
					shouldSayExtinguished.add(fired);
					fired.setFireness(4);
				}
			}
		}
	}

	protected void sendFiredBuildingsForNoCommi() {
		sendNoCommiFireds(shouldSayFireds, 32, HeaderType.Fired);
	}

	private void hearFiredBuildingsForNoCommi(Package shouldRead, int agentID) {
		hearNoCommiFireds(shouldRead, 1, 32, agentID);
	}

	protected void sendOutFiredBuildingsForNoCommi() {
		sendNoCommiFireds(shouldSayOutFireds, 64, HeaderType.OutFired);
	}

	private void hearOutFiredBuildingsForNoCommi(Package shouldRead, int agentID) {
		hearNoCommiFireds(shouldRead, 1, 64, agentID);
	}

	protected void sendExtinguishedBuildings() {
		// log("should say exting: " + shouldSayExtinguished);
		ArrayList<Building> temp = new ArrayList<Building>();
		for (Building b : shouldSayExtinguished)
			if (time - b.getLTS() > 64 || b.stFire == 0)
				temp.add(b);
		shouldSayExtinguished.removeAll(temp);
		if (shouldSayExtinguished.size() != 0)
			sendNoCommiFireds(shouldSayExtinguished, 64, HeaderType.Refuge);
	}

	private void hearExtinguishedBuildings(Package shouldRead, int agentID) {
		hearNoCommiFireds(shouldRead, 0, 64, agentID);
	}

	// TODO dorst kar mikone?
	protected void clearers() {
		if (lastSaidByteArray != null && lastSaidByteArray.length != 0) {
			speak(time, HeaderType.Clearer, lastSaidByteArray);
			if (channel != 0)
				speak(time, HeaderType.Clearer, lastSaidByteArray);
			lastSaidByteArray = null;
		}
		sendClearer();

		for (byte[] b : lastByteArrays)
			if (b != null && b.length > 0) {
				speak(time, HeaderType.Clearer, b);
				if (channel != 0)
					speak(time, HeaderType.Clearer, b);
			}
		sendClearerArea();
	}

	protected void setIsAlive(Entity ent) {
		if (ent != null && ent instanceof AmbulanceTeam)
			((AmbulanceTeam) ent).isAlive = true;
	}

	protected void setIsOK(Entity ent) {
		if (ent != null && ent instanceof Human)
			((Human) ent).isOK = true;
	}

	protected void sendMyPos() {
		int policeCount = 2, fireCount = 3, ambCount = 20;
		if (lowCommi) {
			policeCount = 4;
			fireCount = 6;
			ambCount = 25;
		}
		if ((lowCommi && time % 7 != 0 && time % 7 != 1)
				|| (me() instanceof FireBrigade && ((time - lastPosHeardTime)
						% fireCount != worldModel.fireBrigades.indexOf(myHum())
						% fireCount || time == lastPosHeardTime))
				|| (me() instanceof PoliceForce
						&& ((time - lastPosHeardTime) % policeCount != worldModel.policeForces
								.indexOf(myHum()) % policeCount) || time == lastPosHeardTime)
				|| (me() instanceof AmbulanceTeam && ((time - lastPosHeardTime)
						% ambCount != worldModel.ambulanceTeams
						.indexOf(myHum()) % ambCount || time == lastPosHeardTime))
				|| (myHum().lastUpdatedPos != null && getMyPos().id == myHum().lastUpdatedPos.id)
				|| myHum().hasBuriedness
				|| trapped.containsKey(me().getID().getValue()))
			return;

		Package p = new Package(worldModel);
		p.addHeader(HeaderType.Pos);
		p.addArea(getMyPos(), AreaType.Area);
		speak(time, HeaderType.Pos, p.send());
	}

	protected void hearSOPos(Package p, EntityID agentID) {
		Area pos = p.getArea(AreaType.Area, worldModel);
		setAgentPos(pos, agentID);
	}

	protected void sendLastPacks() {
		int timeForSay = 2;
		if (lowCommi)
			timeForSay = 4;

		for (int i = 0; i < packsForSay.size(); i++) {
			Package temp = new Package(worldModel);
			if (time - packsForSay.get(i).timeSaid >= timeForSay) {
				packsForSay.remove(i);
				i--;
				continue;
			} else if (time - packsForSay.get(i).timeSaid == 0)
				continue;
			temp.setBits(packsForSay.get(i)
					.getBits(0, packsForSay.get(i).start),
					packsForSay.get(i).start);
			temp.addInt(time - packsForSay.get(i).timeSaid, lastPacksTimeLimit);
			temp.setEnd();

			// lastPacksSize[time] += temp.send().length;

			speak(time, temp.tellHeader(), temp.send());
		}
	}

	boolean isUnStuckTrueOnce = false;

	private boolean isStuck() {
		if (isUnStuckTrueOnce && myHum().hasBuriedness)
			return false;
		Coordinate myPos = new Coordinate(myHum().getX(), myHum().getY());
		boolean checkIAmInBlock = false;
		for (Blockade b : getBlocksInRange(worldModel.getTrafficSimulator().agentRad))
			if (b.circumstance.contains(GeometryTools.coordToGeometry(myPos,
					worldModel.getTrafficSimulator().geom))) {
				checkIAmInBlock = true;
				break;
			}
		if (worldModel.myEntrances.size() == 0 && !checkIAmInBlock
				&& worldModel.getTrafficSimulator().myShape == null)
			return false;
		for (PoliceForce police : worldModel.policeForces) {
			if (isReachable(police.getPosition())
					&& police.getID().getValue() != me().getID().getValue()
					&& !police.hasBuriedness && time > 2) {
				isUnStuckTrueOnce = true;
				return false;
			}
		}
		return true;
	}

	protected void sendUnStuck() {
		Package p = new Package(worldModel);
		p.addHeader(HeaderType.UStuck);
		speak(time, HeaderType.UStuck, p.send());
	}

	protected void hearUnStuck(Package p, EntityID agentID) {
		if (trapped.containsKey(agentID.getValue()))
			trapped.remove(agentID.getValue());
	}

	public int RRT = 0;
	public int lastRRT = 0;
	public int RRTpathSize = -1;
	public int lastRRTpathSize = -1;
	public int akharinStuckTime = 0;
	public boolean shouldNotSendUnStuck = false;
	public boolean isInEmergencySituation = false;

	protected void calcutingStucks() {
		if (me() instanceof Human) {
			Human m = (Human) me();
			amIStucking = false;
			shouldNotSendUnStuck = false;

			if (akharinStuckTime != time - 1
					&& (lastRRT < RRTpathSize || lastRRT == 0))
				stuckToolKeshid = 0;

			// log("MyPOS: "
			// + GeometryTools.printPoint(new Coordinate(m.getX(), m
			// .getY())));
			// log("Last POS: "
			// + GeometryTools.printPoint(new Coordinate(xLastCycle,
			// yLastCycle)));

			double hypotMeWithLastMove = Math.hypot(m.getX() - xRealMove,
					m.getY() - yRealMove);

			double hypotMeWithLastCycle = Math.hypot(m.getX() - xLastCycle,
					m.getY() - yLastCycle);

			// if (Math.hypot(getX() - getMyPos().getX(), getY()
			// - getMyPos().getY()) == 0
			// && hypotMeWithLastCycle == 0 && lastMoveCycle == time - 1) {
			// double maxDist = 0;
			// Coordinate myCoordinate = new Coordinate(getX(), getY());
			// Coordinate target = null;
			// for (Coordinate c : getMyPos().shape.getCoordinates()) {
			// if (c.distance(myCoordinate) > maxDist) {
			// target = c;
			// maxDist = c.distance(myCoordinate);
			// }
			// }
			// move(getMyPos(), (int) target.x, (int) target.y, true);
			// }
			// log("xRealMove "+xRealMove+" xLastCycle "+xLastCycle);
			// log("yRealMove "+yRealMove+" yLastCycle "+yLastCycle);

			// log("hypotMeWithLastMove " + hypotMeWithLastMove);
			// log("hypotMeWithLastCycle " + hypotMeWithLastCycle);

			if (((hypotMeWithLastMove > 500 && hypotMeWithLastCycle < 3000 && Math
					.abs(hypotMeWithLastCycle - hypotMeWithLastMove) > 500))// ||
																			// amIGettingObserbedToBlockade())
					&& time > 3
					&& (lastMoveCycle == time - 1 || isInEmergencySituation)
					&& !isInMoveToFire && (RRT < RRTpathSize || RRT == 0)) { // NWWEWEWEW

				akharinStuckTime = time;
				stuckToolKeshid++;
				// if (stuckToolKeshid >= 2)
				amIStucking = true;
			}

			xLastCycle = m.getX();
			yLastCycle = m.getY();
			isInMoveToFire = false;
		}
		lastRRT = RRT;
		lastRRTpathSize = RRTpathSize;
		if (RRT >= RRTpathSize && RRT != 0) {
			RRT = 0;
			RRTpathSize = -1;
			shouldNotSendUnStuck = true;
		}
		// log("stuckToolKeshid222 " + stuckToolKeshid);
	}

	// private boolean amIGettingObserbedToBlockade() {
	// //TODO:
	// return false;
	// }

	protected void checkRechablePolicesAvailability() {
		if (!(me() instanceof PoliceForce) || ((Human) me()).hasBuriedness) {
			boolean isStuck = isStuck() || stuckToolKeshid >= 3;
			// log("IS STUCK - 1: " + isStuck());
			if (isStuck) {
				counterForUnStuck = 0;
				shouldNotSendUnStuck = true;
			} else {
				counterForStucks = 0;
				timeLimitForBeingTrapped = 0;
			}
			if (counterForUnStuck == 5)
				counterForUnStuck = 0;

			// log("1: " + trapped.containsKey(me().getID().getValue()));
			// log("2: " + !isStuck);
			// log("3: " + counterForUnStuck);
			// log("4: " + !shouldNotSendUnStuck);
			if (((trapped.containsKey(me().getID().getValue()) && !isStuck) || (counterForUnStuck > 0 && counterForUnStuck < 5))
					&& !shouldNotSendUnStuck) {
				counterForUnStuck++;
				sendUnStuck();
			} else if (isStuck && worldModel.policeForces.size() != 0) {
				timeLimitForBeingTrapped++;
				if ((time < 5 || timeLimitForBeingTrapped > 2) && time > 2)
					sendStuck();
			}
		}
	}

	protected void sendStuck() {
		if (lowCommi) {
			boolean notMyTurn = false;
			int index = worldModel.agents.indexOf(worldModel.getEntity(me()
					.getID()));
			setStuckLimit();
			if (!(index >= stuckIndex && index < stuckIndex + stuckLimit))
				notMyTurn = true;
			stuckIndex += stuckLimit;
			if (stuckIndex >= worldModel.agents.size()) {
				stuckIndex %= worldModel.agents.size();
				if (index >= 0 && index < stuckIndex)
					notMyTurn = false;
			}
			if (notMyTurn)
				return;
		}
		if ((!trapped.containsKey(me().getID().getValue()))// || !trapped
				// .containsValue(getMyPos().getID().getValue())
				|| (noCommi && !(getMyPos().getID().getValue() == myHum().firstPosition
						.getID().getValue())) || counterForStucks < 5) {
			Package p = new Package(worldModel);
			p.addHeader(HeaderType.Stuck);
			if (time < 20
					&& myHum().firstPosition.nearAreasIncludeSelf
							.contains(getMyPos()))
				p.addInt(myHum().firstPosition.nearAreasIncludeSelf
						.indexOf(getMyPos()),
						myHum().firstPosition.nearAreasIncludeSelf.size());
			else if (time >= 20)
				p.addArea(getMyPos(), AreaType.Area);
			speak(time, HeaderType.Stuck, p.send());
		}
	}

	private int headerTypeByte = 4;
	private int ambCenterByteTook;
	private int fireCenterByteTook;
	private int stuckByteTook;

	private void setByteTook() {
		ambCenterByteTook = (int) Math.ceil((headerTypeByte + 1 + (Calulations
				.findPow(worldModel.areas.size() + 1) + Calulations
				.findPow(cvRecognizer))
				* (int) Math.ceil(worldModel.ambulanceTeams.size() / 2)) / 8);
		fireCenterByteTook = (int) Math.ceil((headerTypeByte + (Calulations
				.findPow(worldModel.buildings.size() + 1) + Calulations
				.findPow(32))
				* (int) Math.ceil(worldModel.fireBrigades.size() / 2)) / 8);
		stuckByteTook = (int) Math.ceil((headerTypeByte + Calulations
				.findPow(worldModel.areas.size())) / 8);
	}

	private void setStuckLimit() {
		int count = channelByteCount;
		if (time % 7 == 4 || time % 7 == 5)
			count -= ambCenterByteTook;
		else if (time % 7 == 1 || time % 7 == 2)
			count -= fireCenterByteTook;
		else
			count = channelByteCount / 2;

		if (count < 0)
			count = 0;

		stuckLimit = (int) (count / stuckByteTook);
	}

	int counterHeardFromSelfStuck = 0;

	protected void hearStuck(Package p, EntityID agentID,
			boolean isHeardFromSelf) {
		Area a = null;
		if (time > 20) {
			a = p.getArea(AreaType.Area, worldModel);
		} else {
			Human h = (Human) worldModel.getEntity(agentID);
			a = h.firstPosition.nearAreasIncludeSelf.get(p
					.getInt(h.firstPosition.nearAreasIncludeSelf.size()));
		}
		trapped.put(agentID.getValue(), a.getID().getValue());
		if (isHeardFromSelf) {
			// log ("value: hearStuck " +
			// trapped.containsValue(a.getID().getValue()));
			counterHeardFromSelfStuck++;
			counterForStucks++;
			if (counterHeardFromSelfStuck == 2)
				counterForStucks--;
		}
		setAgentPos(a, agentID);
	}

	protected void sendRefuge(Refuge r) {
		Package p = new Package(worldModel);
		p.addHeader(HeaderType.Refuge);
		speak(time, HeaderType.Refuge, p.send());
	}

	protected void hearRefuge(Package p, EntityID agentID,
			boolean isHeardFromSelf) {
		((PoliceForce) worldModel.getEntity(agentID)).refugeTarget = null;
	}

	protected void sendTask() {
		Package tasks = new Package(worldModel);
		tasks.addHeader(HeaderType.PCenter);
		ArrayList<FireTask> saidFireTasks = new ArrayList<FireTask>();
		for (PoliceForce pf : worldModel.policeForces) {

			if (pf.hasBuriedness) {
				tasks.addInt(0, 6);
			} else if (pf.stuckTarget == null && pf.refugeTarget == null
					&& pf.fireTask == null) {
				tasks.addInt(1, 6);
			} else if (pf.refugeTarget != null) {
				tasks.addInt(4, 6);
				tasks.addInt(worldModel.refuges.indexOf(pf.refugeTarget),
						worldModel.refuges.size());
				// log ("center pf: " + pf.getID() + "  refuge: " +
				// pf.refugeTarget);
			} else if (pf.fireTask != null
					&& !pf.fireTask.isForSearchAroundZone) {
				tasks.addInt(3, 6);
				boolean contains = false;
				for (FireTask ft : saidFireTasks)
					if (ft.agentPart.getID().getValue() == pf.fireTask.agentPart
							.getID().getValue()
							&& ft.firedBuildingPart.getID().getValue() == pf.fireTask.firedBuildingPart
									.getID().getValue()) {
						contains = true;
						break;
					}
				if (contains) {
					tasks.addInt(1, 2);
					tasks.addInt(saidFireTasks.indexOf(pf.fireTask),
							saidFireTasks.size());
				} else {
					tasks.addInt(0, 2);
					tasks.addInt(pf.fireTask.headOrTail, 2);
					tasks.addArea(pf.fireTask.agentPart, AreaType.Area);
					tasks.addArea(pf.fireTask.firedBuildingPart,
							AreaType.Building);
					saidFireTasks.add(pf.fireTask);
				}
			} else if (pf.fireTask != null && pf.fireTask.isForSearchAroundZone) {
				tasks.addInt(5, 6);
				tasks.addArea(pf.fireTask.agentPart, AreaType.Area);
			} else if (pf.stuckTarget != null) {
				tasks.addInt(2, 6);
				tasks.addArea(pf.stuckTarget, AreaType.Area);
			}
		}
		if (lastCyclesTasks == null)
			lastCyclesTasks = tasks.send();
		boolean similarToLastCycle = (lastCyclesTasks.length == tasks.bits
				.length());
		if (similarToLastCycle)
			for (int i = 0; i < tasks.bits.length(); i++)
				if (tasks.send()[i] != lastCyclesTasks[i]) {
					similarToLastCycle = false;
					break;
				}
		if (!similarToLastCycle) {
			lastCyclesTasks = tasks.send();
			speak(time, HeaderType.PCenter, tasks.send());
			// addToPackageSchedules("PoliceCenterTasks", tasks, 5, 1);
		}

	}

	protected void hearTask(Package tasks) {
		ArrayList<FireTask> heardFireTasks = new ArrayList<FireTask>();
		for (PoliceForce pf : worldModel.policeForces) {

			pf.stuckTarget = null;
			pf.refugeTarget = null;
			pf.fireTask = null;

			int job = tasks.getInt(6);
			if (job == 0) {
				if (pf.getID().getValue() != me().getID().getValue())
					pf.hasBuriedness = true;
				else if (!((PoliceForce) me()).hasBuriedness) {
					sendUnHelp();
				}
			} else if (job == 4) {
				pf.refugeTarget = worldModel.refuges.get(tasks
						.getInt(worldModel.refuges.size()));
				// log ("job == 4 " + pf.getID() + "  " + pf.refugeTarget);
			} else if (job == 3) {
				pf.fireTask = new FireTask();
				int offSetNew = tasks.getInt(2);
				pf.fireTask.isForSearchAroundZone = false;
				if (offSetNew == 1) {
					int pointer = tasks.getInt(heardFireTasks.size());
					FireTask herFT = heardFireTasks.get(pointer);
					pf.fireTask.agentPart = herFT.agentPart;
					pf.fireTask.firedBuildingPart = herFT.firedBuildingPart;
					if (herFT.headOrTail == 0)
						pf.fireTask.headOrTail = 1;
					else
						pf.fireTask.headOrTail = 0;
				} else {
					int headOrTail = tasks.getInt(2);
					pf.fireTask = new FireTask(tasks.getArea(AreaType.Area,
							worldModel), tasks.getArea(AreaType.Building,
							worldModel), false);
					pf.fireTask.headOrTail = headOrTail;
					heardFireTasks.add(pf.fireTask);
				}
				pf.realFireTask = new FireTask(pf.fireTask.agentPart,
						pf.fireTask.firedBuildingPart, pf.fireTask.headOrTail,
						pf.fireTask.isForSearchAroundZone);
			} else if (job == 5) {
				pf.fireTask = new FireTask();
				pf.fireTask.isForSearchAroundZone = true;
				pf.fireTask.agentPart = tasks
						.getArea(AreaType.Area, worldModel);
				pf.fireTask.firedBuildingPart = pf.fireTask.agentPart;
				pf.searchAroundZoneBuilds.clear();
				for (Area a : pf.fireTask.firedBuildingPart.nearAreas50000)
					if (a instanceof Building)
						pf.searchAroundZoneBuilds.add((Building) a);

			} else if (job == 2)
				pf.stuckTarget = tasks.getArea(AreaType.Area, worldModel);

		}
	}

	protected void checkBurriedness() {
		if (me() instanceof Human) {
			if (myHum().isBuriednessDefined() && myHum().getBuriedness() > 0
					&& worldModel.ambulanceTeams.size() != 0)
				sendHelp();
			else {
				counterForHelp = 0;
			}
			if (counterForUnHelp == 5)
				counterForUnHelp = 0;
			if ((helps.containsKey(me().getID().getValue()) && myHum().hasBuriedness == false)
					|| (counterForUnHelp > 0 && counterForUnHelp < 5)) {
				counterForHelp = 0;
				counterForUnHelp++;
				sendUnHelp();
			}
		}
	}

	protected void sendUnHelp() {
		Package p = new Package(worldModel);
		p.addHeader(HeaderType.UHelp);
		speak(time, HeaderType.UHelp, p.send());
	}

	protected void hearUnHelp(Package p, EntityID agentID) {
		if (helps.containsKey(agentID.getValue())) {
			helps.remove(agentID.getValue());
			if (agentID.getValue() != me().getID().getValue())
				((Human) worldModel.getEntity(agentID)).hasBuriedness = false;
		}
	}

	protected void sendHelp() {
		int id = me().getID().getValue();
		if (!helps.containsKey(id) || counterForHelp < 5) {
			Package p = new Package(worldModel);
			p.addHeader(HeaderType.Help);
			p.addArea(getMyPos(), AreaType.Area);
			p.addInt(myHum().deadtime, 502);
			if (myHum().getBuriedness() > 127)
				p.addInt(127, 128);
			else
				p.addInt(myHum().getBuriedness(), 128);
			speak(time, HeaderType.Help, p.send());
		}
	}

	int counterHeardFromSelfHelp = 0;

	protected void hearHelp(Package p, EntityID agentID, boolean isSelfHeard) {
		Area pos = p.getArea(AreaType.Area, worldModel);
		if (!helps.containsKey(agentID.getValue()))
			helps.put(agentID.getValue(), pos.getID().getValue());
		if (isSelfHeard) {
			counterHeardFromSelfHelp++;
			counterForHelp++;
			if (counterHeardFromSelfHelp == 2)
				counterForHelp--;
			return;
		}
		setAgentPos(pos, agentID);
		Human him = ((Human) worldModel.getEntity(agentID));
		him.deadtime = p.getInt(502);
		him.setBuriedness(p.getInt(128));
		if (him.isBuriednessDefined() && him.getBuriedness() > 0)
			him.hasBuriedness = true;
	}

	private void setAgentPos(Area newPos, EntityID id) {
		if (!(worldModel.getEntity(id) instanceof Human))
			return;
		Human h = ((Human) worldModel.getEntity(id));
		if (me().getID().getValue() == id.getValue()) {
			h.lastUpdatedPos = newPos;
			return;
		} else if (worldModel.isInChangeSet(h.getID())
				|| worldModel.isInChangeSet(newPos.getID())) {
			if (time - h.lastPosUpdatedTime < 10) {
				updateWays(worldModel.getArea(h.getPosition()), newPos,
						id.getValue());
				h.lastUpdatedPos = newPos;
			}
			return;
		}
		// if (h instanceof PoliceForce
		// && !(me() instanceof PoliceForce || me() instanceof PoliceOffice))
		// return;
		if (h.lastPosUpdatedTime == time - 1)
			return;
		newPos.setHLTS(time - 1, id.getValue(), lastSeenBy);
		newPos.HLTV = time - 1;
		if (time - h.lastPosUpdatedTime < 10) {
			updateWays(worldModel.getArea(h.getPosition()), newPos,
					id.getValue());
			h.lastUpdatedPos = newPos;
		}
		h.lastPosUpdatedTime = time - 1;
		h.setPosition(newPos.getID());
		h.setX(newPos.getX());
		h.setY(newPos.getY());
		hearAGPosition.put(h, newPos);
	}

	protected void sendCVl() {
		for (EntityID eID : worldModel.changeSet.getChangedEntities()) {
			Entity e = worldModel.getEntity(eID);
			if (!(e instanceof Civilian))
				continue;
			Civilian cv = (Civilian) e;
			if (isGoodCVToSay(cv, false))
				CVsForSay.add(cv);
		}
		if (time <= 10)
			return;
		if (CVsForSay.size() > 0) {
			HashSet<Civilian> CVs = new HashSet<Civilian>();
			for (Civilian c : CVsForSay)
				if (getMyPos().nearAreasIncludeSelf.contains(worldModel
						.getArea(c.getPosition())))
					CVs.add(c);
			if (CVs.size() > 0)
				CVlCalculations(CVs);
		}
		CVsForSay.clear();
	}

	// TODO ED : no refuge situation >> say !hasBuriedness && !hasDamage >>
	// load to the dame dar!!
	private boolean isGoodCVToSay(Civilian cv, boolean generateFakeID) {
		if (!cv.isPositionDefined())
			return false;

		// TODO in shart check besh ba code 2013
		// TODO age gass station betereke cv damage migire vali buridness nadare
		// pas kool kardan shayad bekhad ==> check beshe
		if (!cv.hasBuriedness && worldModel.refuges.size() != 0)
			return false;

		if ((worldModel.refuges.size() == 0
				&& worldModel.getEntity(cv.getPosition()) instanceof rescuecore2.standard.entities.Road && !cv.hasBuriedness))
			return false;

		if (worldModel.ambulanceTeams.contains(worldModel.getEntity(cv
				.getPosition())))
			return false;

		int id = cv.getID().getValue();
		if (generateFakeID)
			id = generateFakeID(cv.getID().getValue() % cvRecognizer,
					worldModel.areas.indexOf(worldModel.getArea(cv
							.getPosition())));

		boolean goodCVToSay = false;
		if (!heardCVs.containsKey(id)
				|| !heardCVs.get(id).equals(cv.getPosition().getValue()))
			goodCVToSay = true;

		// TODO ambul ke balaye sare cv e nage
		else if (cv.lastDeadTimeSaid < 0
				|| Math.abs(cv.lastDeadTimeSaid - cv.deadtime) >= 20)
			goodCVToSay = true;

		return goodCVToSay;
	}

	protected void CVlCalculations(HashSet<Civilian> cvs) {
		Package CV = new Package(worldModel);
		Package temp = new Package(worldModel);
		CV.addHeader(HeaderType.CV);
		CV.addArea(getMyPos(), AreaType.Area);
		for (Civilian myCV : cvs) {
			CV.addID(myCV.getID().getValue());
			CV.addInt(getMyPos().nearAreasIncludeSelf.indexOf(worldModel
					.getArea(myCV.getPosition())),
					getMyPos().nearAreasIncludeSelf.size());
			myCV.lastDeadTimeSaid = myCV.deadtime;
			CV.addInt(myCV.deadtime / 10, 51);
			if (myCV.isBuriednessDefined()) {
				if (myCV.getBuriedness() > 127)
					CV.addInt(127, 128);
				else
					CV.addInt(myCV.getBuriedness(), 128);
			} else
				CV.addInt(0, 128);
			// CVsSended.add(myCV.getID().getValue());
		}
		temp.setBits(CV.getBits(0, CV.start), CV.start);
		temp.timeSaid = time;
		packsForSay.add(temp);
		CV.addInt(0, lastPacksTimeLimit);
		CV.setEnd();
		speak(time, HeaderType.CV, CV.send());
	}

	protected void sendCVlForLowCommi() {
		for (EntityID eID : worldModel.changeSet.getChangedEntities()) {
			Entity e = worldModel.getEntity(eID);
			if (!(e instanceof Civilian))
				continue;
			Civilian cv = (Civilian) e;
			if (isGoodCVToSay(cv, true))
				CVsForSay.add(cv);
		}
		if ((time % 7 != 0 && time % 7 != 3) || time <= 10)
			return;
		HashSet<Civilian> toBeRemoved = new HashSet<Civilian>();
		for (Civilian cv : CVsForSay)
			if (!isGoodCVToSay(cv, true))
				toBeRemoved.add(cv);
		CVsForSay.removeAll(toBeRemoved);
		if (CVsForSay.size() > 0)
			CVlCalculationsForLowCommi(CVsForSay);
		CVsForSay.clear();
	}

	protected void CVlCalculationsForLowCommi(HashSet<Civilian> cvs) {
		Package CV = new Package(worldModel);
		CV.addHeader(HeaderType.CV);
		for (Civilian myCV : cvs) {
			CV.addInt(myCV.getID().getValue() % cvRecognizer, cvRecognizer);
			CV.addArea(worldModel.getArea(myCV.getPosition()), AreaType.Area);
			heardCVs.put(
					generateFakeID(myCV.getID().getValue() % cvRecognizer,
							worldModel.areaToIndex.get(worldModel.getArea(myCV
									.getPosition()))), myCV.getPosition()
							.getValue());
			myCV.lastDeadTimeSaid = myCV.deadtime;
			CV.addInt(myCV.deadtime / 10, 51);
			if (myCV.isBuriednessDefined()) {
				if (myCV.getBuriedness() > 127)
					CV.addInt(127, 128);
				else
					CV.addInt(myCV.getBuriedness(), 128);
			} else
				CV.addInt(0, 128);
			// CVsSended.add(myCV.getID().getValue());
		}
		CV.setEnd();
		speak(time, HeaderType.CV, CV.send());
	}

	// TODO check this
	HashSet<Civilian> noCommiCVs = new HashSet<Civilian>();

	public void sendCVlForNoCommi() {
		HashSet<Civilian> tmp = new HashSet<Civilian>();
		for (Civilian cv : noCommiCVs)
			if (!isGoodCVToSayForNoCommi(cv))
				tmp.add(cv);
		noCommiCVs.removeAll(tmp);
		for (EntityID eID : worldModel.changeSet.getChangedEntities()) {
			Entity e = worldModel.getEntity(eID);
			if (!(e instanceof Civilian))
				continue;
			Civilian cv = (Civilian) e;
			if (isGoodCVToSayForNoCommi(cv))
				noCommiCVs.add(cv);
		}
		if (noCommiCVs.size() > 0)
			CVlCalculationsForLowCommi(noCommiCVs);
	}

	private boolean isGoodCVToSayForNoCommi(Civilian cv) {
		if (!cv.isPositionDefined())
			return false;
		if (!cv.hasBuriedness
				&& ((worldModel.refuges.size() == 0 && worldModel.getEntity(cv
						.getPosition()) instanceof rescuecore2.standard.entities.Road) || worldModel.refuges
						.size() != 0))
			return false;
		if (!(worldModel.getEntity(cv.getPosition()) instanceof rescuecore2.standard.entities.Area))
			return false;
		if (worldModel.getArea(cv.getPosition()) instanceof Refuge)
			return false;
		if (worldModel.ambulanceTeams.contains(worldModel.getEntity(cv
				.getPosition())))
			return false;
		if (cv.deadtime <= time + 2)
			return false;
		// TODO:
		// if (cv.getPosition().getValue() != getMyPos().getID().getValue())
		// return false;
		if (worldModel.getEntity(cv.getPosition()) instanceof Building
				&& ((rescuecore2.standard.entities.Building) worldModel
						.getEntity(cv.getPosition())).isOnFire())
			return false;
		return true;
	}

	private int generateFakeID(int cvRec, int a) {
		if (beginOfFakeCVID == 0)
			beginOfFakeCVID = Integer.parseInt(String.valueOf((int) (21 * Math
					.pow(10, 6 - Integer.toString(worldModel.areas.size())
							.length()))));

		int fakeCVID = Integer.parseInt(String
				.valueOf((int) (beginOfFakeCVID * Math.pow(10, Integer
						.toString(worldModel.areas.size()).length()
						- Integer.toString(a).length())))
				+ String.valueOf(a)) * 100;
		fakeCVID += (((cvRecognizer - (fakeCVID % cvRecognizer)) + cvRec) % cvRecognizer);
		return fakeCVID;
	}

	public HashMap<Integer, Integer> heardCVs = new HashMap<>();

	protected void hearCVl(Package CV, EntityID agentID, int saidChannel,
			boolean isSelfHeard) {
		Area firstPos = CV.getArea(AreaType.Area, worldModel);
		setAgentPos(firstPos, agentID);
		int visitedTime = time - 1 - CV.getTime(lastPacksTimeLimit);
		while (CV.hearingPointer < CV.findLastBit()
				- Calulations.findPow(lastPacksTimeLimit)) {
			int id = CV.getID();
			int index = CV.getInt(firstPos.nearAreasIncludeSelf.size());
			Area a = firstPos.nearAreasIncludeSelf.get(index);
			int deadTime = CV.getInt(51) * 10;
			int bn = CV.getInt(128);
			if (a.getLTV() <= visitedTime) {
				if (saidChannel != 0)
					heardCVs.put(id, a.getID().getValue());
				if (!isSelfHeard)
					addCVlToModel(id, a, deadTime, bn, visitedTime,
							agentID.getValue());
				Civilian civilian = (Civilian) worldModel.getEntityByInt(id);
				civilian.lastDeadTimeSaid = deadTime;

			}
		}
	}

	protected void hearCVlForLowCommi(Package CV, EntityID agentID,
			int saidChannel, boolean isSelfHeard) {

		ArrayList<Integer> cvRecAL = new ArrayList<Integer>();
		while (CV.hearingPointer < CV.findLastBit()) {
			int cvRec = CV.getInt(cvRecognizer);
			cvRecAL.add(cvRec);
			Area a = CV.getArea(AreaType.Area, worldModel);
			int deadTime = CV.getInt(51) * 10;
			int bn = CV.getInt(128);
			int id = generateFakeID(cvRec, worldModel.areaToIndex.get(a));

			// TODO age chand ta boodan chi??vaghti code zirie bood har dafe ye
			// ID jadid misakht:/
			/*
			 * int frequency = Collections.frequency(cvRecAL, cvRec); if
			 * (frequency > 1 && frequency <= 3) id += cvRecognizer * frequency;
			 */
			// System.out.println(me().getID() + "  cvl heared  :  " + id);

			// TODO ye hash Map<Area, ArrayList(Civilian)> ke for nazanim ro kol
			// cv haye shahr!
			for (Civilian yaroo : worldModel.getCivilians())
				if (a.getID().getValue() == yaroo.getPosition().getValue()
						&& yaroo.getID().getValue() != id
						&& yaroo.getID().getValue() % cvRecognizer == cvRec)
					id = yaroo.getID().getValue();
			if (a.getLTV() != time) {
				if (saidChannel != 0)
					heardCVs.put(id, a.getID().getValue());
				if (!isSelfHeard)
					addCVlToModel(id, a, deadTime, bn, time - 1,
							agentID.getValue());

			}
		}
	}

	// protected void hearCVlForNoCommi(Package CV, EntityID agentID,
	// int saidChannel, boolean isSelfHeard) {
	// while (CV.hearingPointer < CV.findLastBit()) {
	// int id = CV.getID();
	// Area a = CV.getArea(AreaType.Area, worldModel);
	// int deadTime = CV.getInt(51) * 10;
	// int bn = CV.getInt(128);
	// if (!worldModel.isInChangeSet(a.getID()) && a.getLTS() <= time - 1) {
	// if (saidChannel != 0)
	// heardCVs.put(id, a.getID().getValue());
	// if (!isSelfHeard)
	// addCVlToModel(id, a, deadTime, bn, time - 1,
	// agentID.getValue());
	// }
	// }
	// }

	protected void addCVlToModel(int num, Area pos, int deadTime, int BN,
			int seenTime, int agentID) {
		Entity existingEntity = worldModel.getEntityByInt(num);
		if (existingEntity == null
				|| !worldModel.isInChangeSet(existingEntity.getID())) {
			EntityID eID = new EntityID(num);
			if (existingEntity == null) {
				existingEntity = Registry.getCurrentRegistry().createEntity(
						StandardEntityURN.CIVILIAN.toString(), eID);
				model.addEntity(existingEntity);
			}

			if (noCommi && pos instanceof Building) {
				Building b = (Building) pos;
				if (b.isOnFire() || b.stFire == 1)
					b.setFireness(4);
				else
					b.setFireness(0);
				b.stFire = 0;
			}
			pos.setHLTS(seenTime, agentID, lastSeenBy);
			pos.HLTV = seenTime;
			Civilian cv = ((Civilian) existingEntity);
			cv.setPosition(pos.getID());
			cv.deadtime = deadTime;
			cv.setBuriedness(BN);
			if (BN != 0)
				cv.hasBuriedness = true;
			else
				cv.hasBuriedness = false;
			cv.setX(pos.getX());
			cv.setY(pos.getY());
		}
	}

	byte[] lastSaidByteArray = null;
	private ArrayList<Road> notClearedRoads = new ArrayList<Road>();
	private ArrayList<byte[]> lastByteArrays = new ArrayList<byte[]>();

	protected void sendClearerArea() {
		if (!(this instanceof PoliceForceAgent))
			return;
		ArrayList<Road> clearedRoad = new ArrayList<Road>();
		for (EntityID ent : worldModel.changeSet.getChangedEntities()) {
			Entity entity = model.getEntity(ent);
			if (entity instanceof rescuecore2.standard.entities.Road) {
				Area area = ((rescuecore2.standard.entities.Area) entity).area;
				if (area.entrances.size() == 0
						|| area.entrances.get(0).internalEntrances.size() == area.entrances
								.size() - 1)
					clearedRoad.add((Road) area);
			}
		}
		lastByteArrays = new ArrayList<byte[]>();
		for (Road road : notClearedRoads)
			if (clearedRoad.contains(road)) {
				Package p = new Package(worldModel);
				p.addHeader(HeaderType.Clearer);
				p.addClearer(road);
				byte[] copyByteArray = p.send();
				lastByteArrays.add(copyByteArray.clone());
				speak(time, HeaderType.Clearer, p.send());
			}

		notClearedRoads = new ArrayList<Road>();
		for (EntityID ent : worldModel.changeSet.getChangedEntities()) {
			Entity entity = model.getEntity(ent);
			if (entity instanceof rescuecore2.standard.entities.Road) {
				Area area = ((rescuecore2.standard.entities.Area) entity).area;
				if (area.entrances.size() > 0
						&& area.entrances.get(0).internalEntrances.size() != area.entrances
								.size() - 1)
					notClearedRoads.add((Road) area);
			}
		}
	}

	protected TreeSet<Integer> heardclearers = new TreeSet<Integer>();

	protected void sendClearer() {
		if (this instanceof PoliceForceAgent) {
			if (time > 3) {
				if (commandHistory.get(time - 1).getAction()
						.equals(StandardMessageURN.AK_CLEAR)) {
					int lastClearedTarget = commandHistory.get(time - 1).lastClearedTarget
							.getValue();
					if (!heardclearers.contains(lastClearedTarget)) {
						Entity entit = worldModel
								.getEntityByInt(lastClearedTarget);
						if (entit instanceof rescuecore2.standard.entities.Road) {
							Area area = ((rescuecore2.standard.entities.Area) entit).area;
							boolean isItCleanEnough = true;
							Road other = (Road) ((rescuecore2.standard.entities.Road) entit).area;
							for (Entrance e : area.entrances)
								if (e.internalEntrances.size() != area.entrances
										.size() - 1)
									isItCleanEnough = false;
							if (isItCleanEnough == true) {
								Package p = new Package(worldModel);
								p.addHeader(HeaderType.Clearer);
								p.addClearer(other);
								lastSaidByteArray = p.send();
								speak(time, HeaderType.Clearer, p.send());
							}
						}
					}
				}
			}
		}
	}

	protected void hearClearer(Package p, boolean isSelfHeard, int agentID,
			int saidChannel) {
		int[] array = p.decodeForClearer();
		int pos = array[0];
		Area area = ((rescuecore2.standard.entities.Area) worldModel
				.getEntityByInt(pos)).area;
		if (saidChannel != 0)
			heardclearers.add(pos);
		if (!worldModel.changeSet.getChangedEntities().contains(area.getID())
				&& !isSelfHeard) {
			area.setHLTS(time - 1, agentID, lastSeenBy);
			for (Blockade b : area.blockades)
				worldModel.getTrafficSimulator().removeBlockade(b);
			area.blockades.clear();
			worldModel.getTrafficSimulator().updateAreaConnectivity(area);
			for (Entrance e : area.entrances) {
				e.resetShapeInArea();
				e.internalEntrances.clear();
				for (Entrance f : area.entrances)
					if (f != e)
						e.internalEntrances.add(f);
			}
			for (int i = 1; i < array.length; i++)
				if (!worldModel.changeSet.getChangedEntities().contains(
						area.entrances.get(i - 1).neighbour.area.getID())) {
					area.entrances.get(i - 1).isItConnectedToNeighbour = area.entrances
							.get(i - 1).neighbour.isItConnectedToNeighbour = (array[i] == 1);
					if (array[i] == 1)
						area.entrances.get(i - 1).neighbour.resetShapeInArea();
				}
		}
	}

	protected void sendWays() {
		Package noWaysP = new Package(worldModel);
		Package waysP = new Package(worldModel);
		noWaysP.addHeader(HeaderType.NoWay);
		waysP.addHeader(HeaderType.Way);
		noWaysP.addArea(getMyPos(), AreaType.Area);
		waysP.addArea(getMyPos(), AreaType.Area);
		Area myArea = getMyPos();
		for (EntityID entityID : worldModel.changeSet.getChangedEntities()) {
			Entity e = worldModel.getEntity(entityID);
			if (e instanceof rescuecore2.standard.entities.Road) {
				worldModel.Road road = ((Road) ((rescuecore2.standard.entities.Road) e).area);
				if (myArea.nearRoadsIncludeSelf.contains(road)
						&& !sentBlocks.contains(road)) {
					int num = myArea.nearRoadsIncludeSelf.indexOf(road);
					sendBlockades(road, num, waysP, noWaysP,
							myArea.nearRoadsIncludeSelf.size());
				}
			}
		}
		if (noWaysP.start > noWaysP.headerPow + noWaysP.areaPow) {
			noWaysP.setEnd();
			speak(time, HeaderType.NoWay, noWaysP.send());
		}
		if (waysP.start > waysP.headerPow + waysP.areaPow) {
			waysP.setEnd();
			speak(time, HeaderType.Way, waysP.send());
		}
	}

	private void sendBlockades(worldModel.Road road, int num, Package ways,
			Package noWays, int amount) {
		int entranceNum = road.entrances.size();
		int[][] array = new int[entranceNum][entranceNum];
		for (Entrance e : road.entrances)
			for (Entrance f : e.internalEntrances)
				array[e.id][f.id] = 1;
		for (Entrance e : road.entrances) {
			if (e.isItConnectedToNeighbour)
				array[e.id][e.id] = 1;
			else
				array[e.id][e.id] = 0;
		}
		blocksCalculations(array, num, amount, ways, noWays);
	}

	private void blocksCalculations(int[][] array, int num, int amount,
			Package ways, Package noWays) {
		mark(array);
		int home = 0;
		BitSet bits = new BitSet();
		for (int i = 0; i < array.length; i++)
			for (int j = i + 1; j < array.length; j++)
				if (array[i][j] != 2) {
					if (array[i][j] == 1)
						bits.set(home);
					home++;
				}
		for (int i = 0; i < array.length; i++) {
			if (array[i][i] == 1)
				bits.set(home);
			home++;
		}

		int amountOf1s = 0;
		for (int i = 0; i < home; i++)
			if (bits.get(i))
				amountOf1s++;

		if (amountOf1s == home)
			return;
		else {
			if (amountOf1s == 0) {
				noWays.addInt(num, amount);
			} else {
				ways.addInt(num, amount);
				ways.setBits(bits, home);
			}
		}
	}

	private void mark(int[][] array) {
		for (int j = 0; j < array.length; j++)
			for (int a = j + 1; a < array.length; a++)
				if (array[j][a] == 1)
					for (int i = a + 1; i < array.length; i++)
						array[a][i] = 2;
				else if (array[j][a] == 0)
					for (int i = a + 1; i < array.length; i++)
						if (array[j][i] == 1)
							array[a][i] = 2;
	}

	protected void hearWays(Package p, EntityID agentID, int saidChannel,
			boolean isSelfHeard) {
		Area hisArea = p.getArea(AreaType.Area, worldModel);
		Entity e = worldModel.getEntity(agentID);
		if (e instanceof Human)
			setAgentPos(hisArea, agentID);
		while (p.hearingPointer < p.findLastBit()) {
			Area area = hisArea.nearRoadsIncludeSelf.get(p
					.getInt(hisArea.nearRoadsIncludeSelf.size()));
			int[][] array = unCalcForBlockade(p, area, hisArea);
			if (saidChannel != 0)
				sentBlocks.add((worldModel.Road) area);
			if (worldModel.isInChangeSet(area.getID())
					|| time - 1 <= area.getLTS() || isSelfHeard)
				continue;
			area.setHLTS(time - 1, agentID.getValue(), lastSeenBy);
			for (int i = 0; i < array.length; i++)
				area.entrances.get(i).internalEntrances.clear();
			for (int i = 0; i < array.length; i++)
				for (int j = i + 1; j < array.length; j++) {
					if (array[i][j] == 1) {
						area.entrances.get(i).internalEntrances
								.add(area.entrances.get(j));
						area.entrances.get(j).internalEntrances
								.add(area.entrances.get(i));
					}
				}
			for (int a = 0; a < array.length; a++) {
				area.entrances.get(a).resetShapeInArea();
				boolean connectivity = (array[a][a] == 1);
				area.entrances.get(a).isItConnectedToNeighbour = connectivity;
				area.entrances.get(a).neighbour.isItConnectedToNeighbour = connectivity;
				if (connectivity)
					area.entrances.get(a).neighbour.resetShapeInArea();
			}
		}
	}

	public void sendingAUndifinededPositionCV(Civilian cv) {
		if (worldModel.refuges.size() > 0) {
			Package CV = new Package(worldModel);
			CV.addHeader(HeaderType.CV);
			CV.addArea(worldModel.refuges.get(0), AreaType.Area);
			CV.addInt(worldModel.refuges.get(0).nearAreasIncludeSelf
					.indexOf(worldModel.refuges.get(0)), worldModel.refuges
					.get(0).nearAreasIncludeSelf.size());
			CV.addInt(cv.deadtime, 502);
			CV.addInt(cv.getBuriedness(), 128);
			CV.setEnd();
			speak(time, HeaderType.CV, CV.send());
		}
	}

	public int[][] unCalcForBlockade(Package p, Area area, Area hisArea) {
		int entrancesNum = area.entrances.size();
		int[][] array = new int[entrancesNum][entrancesNum];
		for (int i = 0; i < array.length; i++)
			for (int j = i; j < array.length; j++)
				array[i][j] = 2;
		int home = 0;
		for (int j = 0; j < entrancesNum; j++) {
			for (int i = j + 1; i < array.length; i++)
				if (array[j][i] == 2) {
					if (p.bits.get(p.hearingPointer + home))
						array[j][i] = 1;
					else
						array[j][i] = 0;
					home++;
				}
			for (int a = j + 1; a < entrancesNum; a++) {
				if (array[j][a] == 1)
					for (int i = a + 1; i < array.length; i++)
						array[a][i] = array[j][i];
				else
					for (int r = a + 1; r < array.length; r++)
						if (array[j][r] == 1)
							array[a][r] = 0;
			}
		}
		for (int a = 0; a < entrancesNum; a++) {
			if (p.bits.get(p.hearingPointer + home) == true)
				array[a][a] = 1;
			else
				array[a][a] = 0;
			home++;
		}
		p.hearingPointer += home;
		return array;
	}

	protected void hearNoWays(Package p, EntityID agentID, int saidChannel,
			boolean isSelfHeard) {
		Area hisArea = p.getArea(AreaType.Area, worldModel);
		Entity e = worldModel.getEntity(agentID);
		if (e instanceof Human)
			setAgentPos(hisArea, agentID);
		while (p.hearingPointer < p.findLastBit()) {
			Area area = hisArea.nearRoadsIncludeSelf.get(p
					.getInt(hisArea.nearRoadsIncludeSelf.size()));
			if (saidChannel != 0)
				sentBlocks.add((worldModel.Road) area);
			if (worldModel.isInChangeSet(area.getID())
					|| time - 1 <= area.getLTS() || isSelfHeard)
				continue;
			area.setHLTS(time - 1, agentID.getValue(), lastSeenBy);

			for (Entrance entrance : area.entrances) {
				entrance.shapeInArea = null;
				entrance.internalEntrances.clear();
				entrance.isItConnectedToNeighbour = entrance.neighbour.isItConnectedToNeighbour = false;
			}
		}
	}

	public void hearFire(Package p, Command cmd) {
		CenterLastHeardTime = time;
		ArrayList<FireBrigade> workers = new ArrayList<FireBrigade>();
		if (lowCommi)
			if (time % 7 == 2)
				for (int i = 0; i < worldModel.fireBrigades.size() / 2; i++) {
					workers.add(worldModel.fireBrigades.get(i));
					if (me().getID().getValue() == worldModel.fireBrigades
							.get(i).getID().getValue())
						shenide = true;
				}
			else
				for (int i = worldModel.fireBrigades.size() / 2; i < worldModel.fireBrigades
						.size(); i++) {
					workers.add(worldModel.fireBrigades.get(i));
					if (me().getID().getValue() == worldModel.fireBrigades
							.get(i).getID().getValue())
						shenide = true;
				}
		else {
			workers.addAll(worldModel.fireBrigades);
			shenide = true;
		}
		for (int i = 0; i < workers.size(); i++) {
			int target = p.getInt(worldModel.buildings.size() + 1);
			if (target != 0) {
				Building aim2 = worldModel.buildings.get(target - 1);
				workers.get(i).aim = aim2;
				int lastTimeSeen = p.getInt(32);
				workers.get(i).aim.heardlastTimeSeen = cmd.getTime()
						- lastTimeSeen;
				// TODO shouldSend
				if (aim2.LTS == time && aim2.stFire == 0) {
					aim2.shouldSend = true;
				}
				if (aim2.getLTS() <= cmd.getTime() - lastTimeSeen
						&& workers.get(i).getID().getValue() == me().getID()
								.getValue()) {
					aim2.stFire = 1;
					aim2.setHLTS(cmd.getTime() - lastTimeSeen, cmd.getAgentID()
							.getValue(), lastSeenBy, false);
					aim2.lastFireness = 1;
					fireds.put(aim2, 1);
				}
			}
		}
	}

	ArrayList<PackageSchedule> packageSchedules = new ArrayList<PackageSchedule>();

	protected void addToPackageSchedules(String key, Package pack, int cycle,
			int num) {
		if (key != null)
			for (PackageSchedule ps : packageSchedules)
				if (ps.key.equals(key)) {
					ps.setData(pack, cycle, num);
					return;
				}
		PackageSchedule ps = new PackageSchedule(key, pack, cycle, num);
		packageSchedules.add(ps);
	}

	protected void removePackageSchedule(String key) {
		for (int i = 0; i < packageSchedules.size(); i++) {
			if (packageSchedules.get(i).key.equals(key)) {
				packageSchedules.remove(i);
				i--;
			}
		}
	}

	private void sendPackageSchedules() {
		ArrayList<PackageSchedule> toBeRemoved = new ArrayList<PackageSchedule>();
		for (PackageSchedule ps : packageSchedules) {
			for (int i = 0; i < ps.num; i++) {
				speak(time, ps.pack.tellHeader(), ps.pack.send());
			}
			ps.cycle--;
			if (ps.cycle == 0)
				toBeRemoved.add(ps);
		}
		packageSchedules.removeAll(toBeRemoved);
	}

	int CenterLastHeardTime = -1;

	private void hears() {
		counterHeardFromSelfHelp = 0;
		counterHeardFromSelfStuck = 0;
		if (heard != null && heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = worldModel.getEntity(cmd.getAgentID());
				setIsAlive(entity);
				setIsOK(entity);
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						// TODO Noisy Mode
						// if (time == 8) {
						// if (as.getChannel() == channel)
						// numOfHeardNoisyModesChannel++;
						// else if (as.getChannel() == otherChannel)
						// numOfHeardNoisyModesOtherChannel++;
						// else
						// System.out
						// .println(":||||||||||||||||||||||||||||||||||||");
						// continue;
						// }

						boolean isSelfHeard;
						if (myHum() == null)
							isSelfHeard = (me().getID().getValue() == cmd
									.getAgentID().getValue());
						else
							isSelfHeard = (myHum().getID().getValue() == cmd
									.getAgentID().getValue());
						Package p = new Package(worldModel);
						p.hear(recivedSpeak);
						HeaderType h = p.tellHeader();
						if (h == radar.Package.HeaderType.Help) {
							if (isSelfHeard)
								lastPosHeardTime = time;
							hearHelp(p, cmd.getAgentID(), isSelfHeard);
						} else if (h == radar.Package.HeaderType.Fired) {
							if (isSelfHeard)
								lastPosHeardTime = time;
							if (noCommi) {
								hearFiredBuildingsForNoCommi(p, cmd
										.getAgentID().getValue());
							} else {
								hearFiredBuildings(p, cmd.getAgentID(),
										isSelfHeard, cmd, as.getChannel(),
										recivedSpeak);
							}
						} else if (h == radar.Package.HeaderType.OutFired) {
							if (noCommi) {
								hearOutFiredBuildingsForNoCommi(p, cmd
										.getAgentID().getValue());
							} else
								hearOutFiredBuildings(p, isSelfHeard,
										as.getChannel(), cmd.getAgentID());
						} else if (h == radar.Package.HeaderType.CV) {
							if (isSelfHeard)
								lastPosHeardTime = time;
							if (lowCommi || noCommi)
								hearCVlForLowCommi(p, cmd.getAgentID(),
										as.getChannel(), isSelfHeard);
							else
								hearCVl(p, cmd.getAgentID(), as.getChannel(),
										isSelfHeard);
						} else if (h == radar.Package.HeaderType.Way) {
							if (isSelfHeard) {
								lastPosHeardTime = time;
							}
							try {
								hearWays(p, cmd.getAgentID(), as.getChannel(),
										isSelfHeard);
							} catch (Exception e) {
								e.printStackTrace();
								if (logger.getWriter() != null)
									e.printStackTrace(logger.getWriter());
							}
						} else if (h == radar.Package.HeaderType.NoWay) {
							if (isSelfHeard)
								lastPosHeardTime = time;
							hearNoWays(p, cmd.getAgentID(), as.getChannel(),
									isSelfHeard);
						} else if (h == radar.Package.HeaderType.Pos) {
							if (time > 10) {
								if (isSelfHeard)
									lastPosHeardTime = time;
								else
									hearSOPos(p, cmd.getAgentID());
							}
						} else if (h == HeaderType.FCenter
								&& me() instanceof FireBrigade) {
							hearFire(p, cmd);
						} else if (h == HeaderType.UHelp)
							hearUnHelp(p, cmd.getAgentID());
						else if (h == HeaderType.UStuck)
							hearUnStuck(p, cmd.getAgentID());
						else if (h == HeaderType.PCenter
								&& me() instanceof PoliceForce) {
							hearTask(p);
						} else if (h == HeaderType.Stuck) {
							if (isSelfHeard)
								lastPosHeardTime = time;
							hearStuck(p, cmd.getAgentID(), isSelfHeard);
						} else if (h == HeaderType.Clearer) {
							hearClearer(p, isSelfHeard, cmd.getAgentID()
									.getValue(), as.getChannel());
						} else if (h == HeaderType.Refuge) {
							if (noCommi) {
								hearExtinguishedBuildings(p, cmd.getAgentID()
										.getValue());
							} else if (me().getID().getValue() == centerForPolice)
								hearRefuge(p, cmd.getAgentID(), isSelfHeard);
						}
					}
				}
			}
		}
	}

	public void speak(int time, HeaderType ht, byte[] data) {
		for (int i : channels.get(ht)) {
			// if (noCommi) {
			// if (ht == HeaderType.Way) {
			// harchiNo[time][1] += data.length;
			// } else if (ht == HeaderType.NoWay) {
			// harchiNo[time][2] += data.length;
			// } else if (ht == HeaderType.CV) {
			// harchiNo[time][3] += data.length;
			// } else if (ht == HeaderType.Fired) {
			// harchiNo[time][4] += data.length;
			// } else if (ht == HeaderType.OutFired) {
			// harchiNo[time][5] += data.length;
			// } else if (ht == HeaderType.Clearer) {
			// harchiNo[time][12] += data.length;
			// } else if (ht == HeaderType.ACenter) {
			// harchiNo[time][13] += data.length;
			// } else if (ht == HeaderType.PCenter) {
			// harchiNo[time][14] += data.length;
			// } else if (ht == HeaderType.FCenter) {
			// harchiNo[time][15] += data.length;
			// } else if (ht == HeaderType.Info) {
			// harchiNo[time][19] += data.length;
			// } else if (ht == HeaderType.Pos) {
			// harchiNo[time][21] += data.length;
			// } else if (ht == HeaderType.Refuge) {
			// harchiNo[time][23] += data.length;
			// } else if (ht == HeaderType.Stuck) {
			// harchiNo[time][26] += data.length;
			// } else if (ht == HeaderType.UStuck) {
			// harchiNo[time][27] += data.length;
			// } else if (ht == HeaderType.Help) {
			// harchiNo[time][28] += data.length;
			// } else if (ht == HeaderType.UHelp) {
			// harchiNo[time][29] += data.length;
			// }
			// harchiNo[time][0] += data.length;
			// } else {
			// if (i == channel) {
			// if (ht == HeaderType.Way) {
			// harchi[time][1] += data.length;
			// } else if (ht == HeaderType.NoWay) {
			// harchi[time][2] += data.length;
			// } else if (ht == HeaderType.CV) {
			// harchi[time][3] += data.length;
			// } else if (ht == HeaderType.Fired) {
			// harchi[time][4] += data.length;
			// } else if (ht == HeaderType.OutFired) {
			// harchi[time][5] += data.length;
			// } else if (ht == HeaderType.Clearer) {
			// harchi[time][12] += data.length;
			// } else if (ht == HeaderType.ACenter) {
			// harchi[time][13] += data.length;
			// } else if (ht == HeaderType.PCenter) {
			// harchi[time][14] += data.length;
			// } else if (ht == HeaderType.FCenter) {
			// harchi[time][15] += data.length;
			// } else if (ht == HeaderType.Info) {
			// harchi[time][19] += data.length;
			// } else if (ht == HeaderType.Pos) {
			// harchi[time][21] += data.length;
			// } else if (ht == HeaderType.Refuge) {
			// harchi[time][23] += data.length;
			// } else if (ht == HeaderType.Stuck) {
			// harchi[time][26] += data.length;
			// } else if (ht == HeaderType.UStuck) {
			// harchi[time][27] += data.length;
			// } else if (ht == HeaderType.Help) {
			// harchi[time][28] += data.length;
			// } else if (ht == HeaderType.UHelp) {
			// harchi[time][29] += data.length;
			// }
			// harchi[time][0] += data.length;
			// } else {
			// if (ht == HeaderType.Way)
			// harchi[time][7] += data.length;
			// else if (ht == HeaderType.NoWay)
			// harchi[time][8] += data.length;
			// else if (ht == HeaderType.CV)
			// harchi[time][9] += data.length;
			// else if (ht == HeaderType.Fired)
			// harchi[time][10] += data.length;
			// else if (ht == HeaderType.OutFired)
			// harchi[time][11] += data.length;
			// else if (ht == HeaderType.ACenter)
			// harchi[time][16] += data.length;
			// else if (ht == HeaderType.PCenter)
			// harchi[time][17] += data.length;
			// else if (ht == HeaderType.FCenter)
			// harchi[time][18] += data.length;
			// else if (ht == HeaderType.Info)
			// harchi[time][20] += data.length;
			// else if (ht == HeaderType.Pos)
			// harchi[time][22] += data.length;
			// else if (ht == HeaderType.Refuge)
			// harchi[time][24] += data.length;
			// else if (ht == HeaderType.Clearer)
			// harchi[time][25] += data.length;
			// else if (ht == HeaderType.Stuck)
			// harchi[time][30] += data.length;
			// else if (ht == HeaderType.UStuck)
			// harchi[time][31] += data.length;
			// else if (ht == HeaderType.Help)
			// harchi[time][32] += data.length;
			// else if (ht == HeaderType.UHelp)
			// harchi[time][33] += data.length;
			// harchi[time][6] += data.length;
			// }
			// }
			// TODO Noisy Mode
			// if (time == 7)
			// return;

			sendSpeak(time, i, data);
		}
	}

	protected abstract void decide() throws ActionCommandException;
}