package agent;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import worldModel.Area;
import worldModel.Blockade;
import worldModel.Entrance;
import worldModel.GeometryTools;
import worldModel.Refuge;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineSegment;

public abstract class ThinkAgent<E extends StandardEntity> extends
		RadarAgent<E> {

	private Coordinate centerOfMap = new Coordinate();

	protected int counterForFeedBack = 0;
	protected int counterFeedBackStuck = 0;
	private static final int RANDOM_WALK_LENGTH = 50;

	public ArrayList<SectorZone> sectorZones = new ArrayList<SectorZone>();
	public ArrayList<SectorZone> noHearSectorZones = new ArrayList<SectorZone>();

	private boolean filledCanBeVisited = false;
	boolean centerOfMapSet = false;

	public FireZoneBuilder fireZoneBuilder = null;

	private void setMapCenter() {
		int minX = Integer.MAX_VALUE, minY = Integer.MAX_VALUE;
		int maxX = Integer.MIN_VALUE, maxY = Integer.MIN_VALUE;
		for (worldModel.Building b : worldModel.buildings) {
			if (b.getX() > maxX)
				maxX = b.getX();
			if (b.getX() < minX)
				minX = b.getX();
			if (b.getY() > maxY)
				maxY = b.getY();
			if (b.getY() < minY)
				minY = b.getY();
		}
		centerOfMap.x = (minX + maxX) / 2;
		centerOfMap.y = (minY + maxY) / 2;
		centerOfMapSet = true;
	}

	protected void postConnect() {
		super.postConnect();
		fireZoneBuilder = new FireZoneBuilder(worldModel, bigMap,
				averageOfSurfaces, logger, noCommi);
	}

	protected void PreSendOnFireBuildings() {
		if (me() instanceof FireBrigade)
			fireZoneBuilder.executeModelZones((FireBrigadeAgent) this);
		else if (me() instanceof PoliceForce)
			fireZoneBuilder.executeModelZones(null);
		if (noCommi)
			doNoCommiFireStuff();
	}

	HashSet<worldModel.Building> iHaveSeenByMyself = new HashSet<worldModel.Building>();

	private void doNoCommiFireStuff() {
		if (me() instanceof AmbulanceTeam)
			fireZoneBuilder.executeModelZones(null);
		shouldSayFireds.clear();
		shouldSayOutFireds.clear();

		for (FireZone fz : fireZoneBuilder.modelZones) {
			boolean iMySelfSawIt = false;
			for (worldModel.Building b : fz.ignitedBuilds) {
				if ((b.getLTS() > b.getHLTS() || iHaveSeenByMyself.contains(b))
						&& time - b.getLTS() < 32) {
					// TODO: b shayad dige tuye ignited ha nabashe (besuze
					// masalan)!
					for (worldModel.Building around : fz.simulatedAroundsWithSetGood)
						if (time - around.getLTS() < 32)
							shouldSayFireds.add(around);
					iMySelfSawIt = true;
					iHaveSeenByMyself.add(b);
					break;
				}
			}
			if (!iMySelfSawIt) {
				for (worldModel.Building b : fz.ignitedBuilds) {
					// TODO jadid tarin buildinge dide shode ro bege
					if (time - b.getLTS() < 64) {
						shouldSayOutFireds.add(b);
						break;
					}
				}
			}
		}

		if (shouldSayFireds.size() != 0)
			sendFiredBuildingsForNoCommi();
		if (shouldSayOutFireds.size() != 0)
			sendOutFiredBuildingsForNoCommi();

	}

	protected void zoningAndNoCommiWorks() {
		if (!noCommi)
			return;

		if (me() instanceof FireBrigade) {
			if (extinguishedTarg != null
					&& extinguishTime == time - 1
					&& ((worldModel.Building) worldModel
							.getArea(extinguishedTarg)).stFire == 0) {
				boolean b = true;
				for (FireZone fz : fireZoneBuilder.modelZones) {
					if (fz.buildings.contains(worldModel
							.getArea(extinguishedTarg))) {
						b = false;
						break;
					}
				}
				if (b == true)
					shouldSayExtinguished.add((worldModel.Building) worldModel
							.getArea(extinguishedTarg));
			}
		}
		for (FireZone fz : fireZoneBuilder.modelZones) {
			fz.maxInfo = Integer.MIN_VALUE;
			for (worldModel.Building b : fz.ignitedBuilds)
				if (b.getLTS() > fz.maxInfo)
					fz.maxInfo = b.getLTS();
		}

		ArrayList<worldModel.Building> shouldRemove = new ArrayList<>();
		for (worldModel.Building b : shouldSayExtinguished) {
			FireZone fireZoneContains = null;
			for (FireZone fz : fireZoneBuilder.modelZones)
				if (fz.buildings.contains(b)) {
					fireZoneContains = fz;
					break;
				}
			if (fireZoneContains == null)
				continue;
			if (fireZoneContains.maxInfo < b.getLTS()) {
				for (worldModel.Building bu : fireZoneContains.ignitedBuilds)
					bu.stFire = 0;
				fireZoneBuilder.modelZones.remove(fireZoneContains);
			} else if (b.stFire == 1)
				shouldRemove.add(b);
		}
		shouldSayExtinguished.removeAll(shouldRemove);
		for (int i = 0; i < fireZoneBuilder.modelZones.size(); i++) {
			fireZoneBuilder.modelZones.get(i).fireZoneNum = i;
			for (worldModel.Building b : fireZoneBuilder.modelZones.get(i).buildings)
				b.fireZoneNum = i;
		}
		sendExtinguishedBuildings();
	}

	private void assignToSectorZones(ArrayList<SectorZone> sectorZones,
			ArrayList<Human> agents, boolean nocommiAssigning) {
		if (!nocommiAssigning) {
			ArrayList<String> positions = setPositions(agents);
			ArrayList<Human> policeForces = new ArrayList<Human>();
			ArrayList<Human> otherType = new ArrayList<Human>();
			ArrayList<SectorZone> tempPoliceZones = new ArrayList<SectorZone>();
			ArrayList<SectorZone> tempOtherZones = new ArrayList<SectorZone>();

			for (Human h : agents) {
				if (h instanceof PoliceForce)
					policeForces.add(h);
				else
					otherType.add(h);
			}

			for (int i = 0; i < sectorZones.size(); i++) {
				sectorZones.get(i).num = i;
				if (positions.get(i) == "p")
					tempPoliceZones.add(sectorZones.get(i));
				else if (positions.get(i) == "o")
					tempOtherZones.add(sectorZones.get(i));
			}

			// log("tempPoliceZones " + tempPoliceZones.size() +
			// " tempOtherZones "
			// + tempOtherZones.size());
			if (policeForces.size() > 0) {
				HungarianAlgorithm ha = new HungarianAlgorithm();
				ha.executeSectorZones(tempPoliceZones, policeForces);
				for (int i = 0; i < policeForces.size(); i++) {
					if (ha.getMatchedNum(i) != -1) {
						policeForces.get(i).sectorZone = tempPoliceZones.get(ha
								.getMatchedNum(i));
						policeForces.get(i).sectorZoneNum = policeForces.get(i).sectorZone.num;
						policeForces.get(i).originalSectorZone = tempPoliceZones
								.get(ha.getMatchedNum(i));
					}
				}
			}

			if (otherType.size() > 0) {
				HungarianAlgorithm ha = new HungarianAlgorithm();
				ha.executeSectorZones(tempOtherZones, otherType);
				for (int i = 0; i < otherType.size(); i++) {
					if (ha.getMatchedNum(i) != -1) {
						otherType.get(i).sectorZone = tempOtherZones.get(ha
								.getMatchedNum(i));
						otherType.get(i).sectorZoneNum = otherType.get(i).sectorZone.num;
						otherType.get(i).originalSectorZone = tempOtherZones
								.get(ha.getMatchedNum(i));
					}
				}
			}

		} else {
			if (agents.size() > 0) {
				HungarianAlgorithm ha = new HungarianAlgorithm();
				ha.executeSectorZones(sectorZones, agents);
				for (int i = 0; i < agents.size(); i++) {
					if (ha.getMatchedNum(i) != -1) {
						agents.get(i).noCommiSectorZone = sectorZones.get(ha
								.getMatchedNum(i));
						agents.get(i).noCommisectorZoneNum = ha
								.getMatchedNum(i);
						agents.get(i).noCommioriginalSectorZone = sectorZones
								.get(ha.getMatchedNum(i));

					}
				}
			} else {
				// System.err.println("khaak bar saram O___O");
			}
		}
	}

	private ArrayList<String> setPositions(ArrayList<Human> agents) {
		ArrayList<Human> policeForces = new ArrayList<Human>();
		ArrayList<Human> otherType = new ArrayList<Human>();
		ArrayList<String> positions = new ArrayList<String>();

		for (Human h : agents) {
			if (h instanceof PoliceForce)
				policeForces.add(h);
			else
				otherType.add(h);
		}

		int pfs = policeForces.size();
		int ots = otherType.size();
		if (pfs == 0) {
			for (int i = 0; i < agents.size(); i++)
				positions.add("o");
			return positions;
		}
		int floor = (int) Math.floor(((double) ots / pfs));// kamtarin tedad
															// othertype haye
															// beyne police ha
		int ceil = (int) Math.ceil(((double) ots / pfs));// bishtarin tedad
															// othertype haye
															// beyne police ha
		int mod = ots % pfs; // tedad bishtarin other type haye beyne police ha
		int modDif = pfs - mod;// tedad kamtarin other type haye beyne police ha
		int temp = 0;
		// ArrayList<SectorZone> tempSectorZone = new ArrayList<SectorZone>();
		for (int i = 0; i < mod; i++) {
			for (int j = 0; j < ceil; j++) {
				// tempSectorZone
				// .add(otherType.get(otherType.size() - 1).sectorZone);
				positions.add("o");
				otherType.get(otherType.size() - 1).sectorZoneNum = temp;
				otherType.remove(otherType.size() - 1);
				temp++;
			}
			// tempSectorZone
			// .add(policeForces.get(policeForces.size() - 1).sectorZone);
			positions.add("p");
			policeForces.get(policeForces.size() - 1).sectorZoneNum = temp;
			policeForces.remove(policeForces.size() - 1);
			temp++;
		}
		for (int i = 0; i < modDif; i++) {
			for (int j = 0; j < floor; j++) {
				// tempSectorZone
				// .add(otherType.get(otherType.size() - 1).sectorZone);
				positions.add("o");
				otherType.get(otherType.size() - 1).sectorZoneNum = temp;
				otherType.remove(otherType.size() - 1);
				temp++;
			}
			// tempSectorZone
			// .add(policeForces.get(policeForces.size() - 1).sectorZone);
			positions.add("p");
			policeForces.get(policeForces.size() - 1).sectorZoneNum = temp;
			policeForces.remove(policeForces.size() - 1);
			temp++;
		}
		// String str = "";
		// for (SectorZone sz : sectorZones) {
		// for (Human h : agents) {
		// if (h.sectorZone == sz) {
		// if (h instanceof PoliceForce)
		// str += "p ";
		// else
		// str += "a ";
		// }
		// }
		// }
		// for (String s : positions)
		// str += s + " ";
		// log("numberOfPolices: " + pfs + " numberOfOtherType: " + ots + str);

		if (positions.size() > 0)
			return positions;
		return null;
	}

	// private SectorZone nearestSectorZone(SectorZone sectorZone,
	// Collection<SectorZone> sectorZones2) {
	// int minDist = Integer.MAX_VALUE;
	// SectorZone bestSectorZone = null;
	// for (SectorZone sZ : sectorZones2) {
	// int dist = (int) Math.hypot(sZ.center.x - sectorZone.center.x,
	// sZ.center.y - sectorZone.center.y);
	// if (dist < minDist) {
	// minDist = dist;
	// bestSectorZone = sZ;
	// }
	// }
	// return bestSectorZone;
	// }

	protected void decide() throws ActionCommandException {
		if (lowCommi)
			cvRecognizer = 8;

		if (!filledCanBeVisited) {
			for (Area area : worldModel.areas) {
				if (area.getNeighbours().size() < 2
						&& GeometryTools
								.createPolygonFromApexes(
										area.serverArea.getApexList())
								.buffer(-500).getArea() <= 0)
					area.canBeVisited = false;

			}
			if (worldModel.map == MapType.Mecxico)
				worldModel.spitOfMexico();
			filledCanBeVisited = true;
		}

		try {
			centers();
			PreSendOnFireBuildings();
			updateCVsInfo();
			if (me() instanceof Human) {
				if (me() instanceof PoliceForce) {
					calcutingStucks();
					randomWalkComputations();
				}
				mikhadBemire();
			}
		} catch (Exception e) {
			if (e instanceof ActionCommandException)
				throw e;
			else {
				// System.err.println("int time " + time + " from agent "
				// + me().getID().getValue() + " in ThinkAgent!");
				e.printStackTrace();
				if (logger.getWriter() != null)
					e.printStackTrace(logger.getWriter());
			}
		}
		if (myHum() != null)
			amIInSearch = false;
		zoning();
		zoningAndNoCommiWorks();

		act();
	}

	private void updateCVsInfo() {
		for (Civilian cv : worldModel.getCivilians()) {
			if (cv.isPositionDefined()
					&& getMyPos().getID().getValue() == cv.getPosition()
							.getValue()) {
				boolean fakeIDSituation = false;
				if (lowCommi || noCommi) {
					for (EntityID eID : worldModel.changeSet
							.getChangedEntities()) {
						Entity e = worldModel.getEntity(eID);
						if (!(e instanceof Civilian))
							continue;
						Civilian realCv = (Civilian) e;
						// TODO tof tarin, same cv rec situation!
						if (realCv.getID().getValue() % cvRecognizer == cv
								.getID().getValue() % cvRecognizer
								&& realCv.getPosition().getValue() == cv
										.getPosition().getValue()
								&& realCv.getID() != cv.getID()) {
							if (me() instanceof AmbulanceTeam) {
								AmbulanceTeamAgent mane = (AmbulanceTeamAgent) this;
								if (mane.myTarget != null)
									if (mane.myIndex == cv.getID().getValue()
											% cvRecognizer) {
										if (mane.goodEnough(realCv))
											mane.setTarget(realCv);
										else
											mane.setTarget(null);
									}
							}
							fakeIDSituation = true;
							model.removeEntity(cv.getID());
						}
					}

				}
				if (!worldModel.isInChangeSet(cv.getID())
						&& ((!noCommi && !lowCommi) || (!fakeIDSituation && (noCommi || lowCommi))))
					if (worldModel.refuges.size() > 0)
						cv.setPosition(worldModel.refuges.get(0).getID());
					else
						cv.setPosition(worldModel.roads.get(0).getID());
			}
		}
	}

	private void zoningForNoCommiComputation() {
		if (sectorZones.size() == 0
				&& (me() instanceof PoliceForce
						|| me() instanceof AmbulanceTeam || me() instanceof FireBrigade)) {
			noCommiSearchZoning();
		}
		if (me() instanceof PoliceForce && noHearSectorZones.size() == 0) {
			policeTasksZoningForNoHear();
		}

	}

	private void zoningForLowCommiComputation() {
		if (sectorZones.size() == 0) {
			Clustering(worldModel.agents.size(), sectorZones);
			assignToSectorZones(sectorZones, worldModel.agents, false);
		}

		if (me() instanceof PoliceForce && noHearSectorZones.size() == 0)
			policeTasksZoningForNoHear();

	}

	private void zoningForFullCommiComputation() {
		if (sectorZones.size() > 0)
			return;

		ArrayList<Human> agents = new ArrayList<Human>();
		for (Human h : worldModel.agents)
			agents.add(h);

		Clustering(agents.size(), sectorZones);
		assignToSectorZones(sectorZones, agents, false);
	}

	protected void Clustering(int numberOfZones,
			ArrayList<SectorZone> sectorzones) {
		KMeans kmeans = new KMeans(worldModel, numberOfZones, logger);
		kmeans.init();
		ArrayList<Cluster> clusters = kmeans.calculate();
		for (Cluster cl : clusters) {
			SectorZone sz = new SectorZone();
			for (Point p : cl.points) {
				sz.buildings.add(p.building);
			}
			sz.setCenter();
			sectorzones.add(sz);
		}
		for (SectorZone sZ : sectorzones)
			for (worldModel.Building b : sZ.buildings)
				if (b instanceof Refuge)
					sZ.refuges.add((Refuge) b);
		// String s = "";
		// for (int i = 0; i < sectorZones.size(); i++) {
		// if (i != sectorZones.size() - 1)
		// s += GeometryTools.printLine(sectorZones.get(i).center,
		// sectorZones.get((i + 1) % sectorZones.size()).center);
		// }
		// System.err.println("lines before sort: " + s);
		// if (!noHearAssigning)
		// sortZonesByDistance(1000);
		// else
		// sortZonesByDistanceGreedy(sectorzones);
	}

	// private void sortZonesByDistance(int N) {
	// SectorZoneList best = generateRandomState(1, Double.MAX_VALUE);
	//
	// System.out.println("MAXDIST: " + best.maxDist);
	// int m = 3;
	// for (int i = 0; i < N; i++) {
	// SectorZoneList sZL = generateRandomState(m, best.maxDist);
	// if (sZL == null)
	// continue;
	// if (best == null || sZL.maxDist < best.maxDist)
	// best = sZL;
	// }
	// System.out.println("FINAL MAXDIST: " + best.maxDist);
	// sectorZones = best.sectorZones;
	// }
	//
	// class SectorZoneList {
	// ArrayList<SectorZone> sectorZones = new ArrayList<SectorZone>();
	// double maxDist = 0;
	// }
	//
	// private Random sectorZonesRandom = new Random(1357);
	//
	// private SectorZoneList generateRandomState(int m, double maxDist) {
	// SectorZoneList result = new SectorZoneList();
	// ArrayList<SectorZone> allSectorZones = (ArrayList<SectorZone>)
	// sectorZones
	// .clone();
	//
	// // SectorZone first = allSectorZones.get(0);
	// SectorZone first = allSectorZones.get(sectorZonesRandom
	// .nextInt(sectorZones.size()));
	// allSectorZones.remove(first);
	// result.sectorZones.add(first);
	// while (allSectorZones.size() > 0) {
	// ArrayList<SectorZone> nearests = new ArrayList<SectorZone>();
	// nearests.addAll(findMNearestZones(m, allSectorZones, first));
	//
	// SectorZone next = nearests.get(sectorZonesRandom.nextInt(nearests
	// .size()));
	//
	// double dist = Mathematic.getDistance(next.center, first.center) / 1000.0;
	// result.maxDist = Math.max(result.maxDist, dist);
	// // result.maxDist += dist;
	// if (result.maxDist > maxDist)
	// return null;
	//
	// first = next;
	// result.sectorZones.add(first);
	// allSectorZones.remove(first);
	// }
	//
	// return result;
	// }
	//
	// private Collection<SectorZone> findMNearestZones(int m,
	// Collection<SectorZone> sectorZones, SectorZone sectorZone) {
	// TreeMap<Double, SectorZone> distance = new TreeMap<Double, SectorZone>();
	// for (SectorZone sZ : sectorZones) {
	// distance.put(Mathematic.getDistance(sZ.center, sectorZone.center),
	// sZ);
	// if (distance.size() > m)
	// distance.pollLastEntry();
	// }
	// return distance.values();
	// }

	private void zoning() {
		if (!centerOfMapSet)
			setMapCenter();
		if (noCommi) {
			zoningForNoCommiComputation();
		} else if (lowCommi) {
			zoningForLowCommiComputation();
		} else {
			zoningForFullCommiComputation();
		}
	}

	private void policeTasksZoningForNoHear() {
		int inBuildPolice = 0;
		for (PoliceForce pf : worldModel.policeForces) {
			if (pf.firstPosition instanceof worldModel.Building)
				inBuildPolice++;
		}
		if (inBuildPolice >= worldModel.policeForces.size() / 2)
			strategy = NoCommiStrategy.TURNING;
		else
			strategy = NoCommiStrategy.NOTTURNING;
		ArrayList<Human> agents = new ArrayList<Human>();

		if (strategy == NoCommiStrategy.NOTTURNING) {
			for (PoliceForce pf : worldModel.policeForces)
				if (!(pf.firstPosition instanceof worldModel.Building))
					agents.add(pf);
		} else
			agents.addAll(worldModel.policeForces);

		if (agents.size() > 0) {

			Clustering(agents.size(), noHearSectorZones);
			assignToSectorZones(noHearSectorZones, agents, true);

			if (me() instanceof PoliceForce)
				for (Human pf : agents)
					pf.noCommiSectorZone.originalOwner = pf;

			if (strategy == NoCommiStrategy.NOTTURNING) {
				agents.clear();
				for (PoliceForce pf : worldModel.policeForces)
					if (pf.noCommiSectorZone == null)
						agents.add(pf);
				if (agents.size() > 0)
					assignToSectorZones(noHearSectorZones, agents, true);
			}

		}
		/*
		 * for (Area b : myHum().noCommiSectorZone.areas) if (b instanceof Road)
		 * myHum().noCommiSectorZone.roads.add((Road)b);
		 */

	}

	public enum NoCommiStrategy {
		TURNING, NOTTURNING
	}

	private void noCommiSearchZoning() {
		ArrayList<Human> agents = new ArrayList<Human>();
		if ((me() instanceof PoliceForce && worldModel.policeForces
				.indexOf(me()) < worldModel.policeForces.size() / 2)
				|| me() instanceof FireBrigade) {
			for (int i = 0; i < worldModel.policeForces.size() / 2; i++)
				agents.add(worldModel.policeForces.get(i));
			agents.addAll(worldModel.fireBrigades);
		} else if ((me() instanceof PoliceForce && worldModel.policeForces
				.indexOf(me()) >= worldModel.policeForces.size() / 2)
				|| me() instanceof AmbulanceTeam) {
			for (int i = worldModel.policeForces.size() / 2; i < worldModel.policeForces
					.size(); i++)
				agents.add(worldModel.policeForces.get(i));
			agents.addAll(worldModel.ambulanceTeams);
		}
		if (agents.size() > 0) {
			Clustering(agents.size(), sectorZones);
			assignToSectorZones(sectorZones, agents, false);
		}

	}

	private void moveToNearestMinLTV(ArrayList<Area> areas)
			throws ActionCommandException {
		int minLTV = getMinLTV(areas);
		moveToNearest(areas, minLTV);
	}

	public void moveToNearest(ArrayList<Area> areas, int minLTV)
			throws ActionCommandException {
		int minDist = Integer.MAX_VALUE;
		Area target = null;
		for (Area a : areas) {
			if (a.getLTV() != minLTV
					|| !((a instanceof worldModel.Building))
					|| !a.canBeVisited
					|| (!a.isReachable && !(me() instanceof PoliceForce))
					|| (noCommi && a instanceof worldModel.Building
							&& !((worldModel.Building) a).isFirenessDefined() && ((worldModel.Building) a).fireRisk))
				continue;
			int dist;
			if (me() instanceof PoliceForce)
				dist = a.distFromeSelfWOB;
			else
				dist = a.distanceFromSelf;
			if (minDist > dist) {
				minDist = dist;
				target = a;
			}
		}
		if (target == null)
			return;
		if (me() instanceof PoliceForce)
			((PoliceForceAgent) this).fastClear(target);
		move(target);
	}

	protected int getMinLTS(ArrayList<worldModel.Building> buildings) {
		int minLTS = time + 1;
		for (worldModel.Building b : buildings)
			if (b.getLTS() < minLTS) {
				minLTS = b.getLTS();
			}
		return minLTS;
	}

	private int getMinLTV(ArrayList<Area> areas) {
		int minLTV = time + 1;
		for (Area a : areas)
			if (a.getLTV() < minLTV && a instanceof worldModel.Building
					&& ((worldModel.Building) a).stFire != 1
					&& ((worldModel.Building) a).fireness() <= 0
					&& (a.isReachable || me() instanceof PoliceForce)
					&& a.canBeVisited) {

				minLTV = a.getLTV();
			}
		return minLTV;
	}

	public void moveToSee(ArrayList<worldModel.Building> buildings, int minLTS)
			throws ActionCommandException {
		worldModel.Building target = null;
		int minDist = Integer.MAX_VALUE;
		for (worldModel.Building building : buildings) {
			if (building.getLTS() != minLTS)
				continue;
			int dist = 0;
			if (me() instanceof PoliceForce)
				dist = distToFire(building);
			else
				dist = getDistToFire(building);
			if (minDist > dist) {
				minDist = dist;
				target = building;
			}
		}
		if (target == null)
			return;
		if (me() instanceof PoliceForce)
			((PoliceForceAgent) this).moveToFireForPolice(target);
		if (me() instanceof FireBrigade)
			((FireBrigadeAgent) this).moveToFireAndSee(target);
		((AmbulanceTeamAgent) this).moveToFireAndSee(target);
	}

	private int distToFire(worldModel.Building building) {
		ArrayList<Area> visionAreas = new ArrayList<Area>();
		visionAreas.addAll(building.visionRoads.keySet());
		visionAreas.addAll(building.visionBuildings.keySet());
		int minDist = Integer.MAX_VALUE;
		for (Area a : visionAreas)
			if (a.distFromeSelfWOB < minDist)
				minDist = a.distFromeSelfWOB;

		return minDist;
	}

	public void fireRisk(worldModel.Building building) {
		// TODO It needs improvement(kolan ke tofe-vali tof haye ba deghat tar),
		// just like "predictingFire" in Human
		float interval = time - (building.timeOfStartOfFire);
		float limit = ((Human) me()).mapRecognizer(worldModel);
		float fireRange = (interval / 30f) * limit;
		for (Area area : building.nearAreas50000) {
			if (area instanceof worldModel.Building
					&& building.distance(area) <= fireRange
					&& !((worldModel.Building) area).isFirenessDefined()
					&& !((worldModel.Building) area).fireRisk) {
				((worldModel.Building) area).fireRisk = true;
			}
		}
	}

	public void noCommiSearch(ArrayList<SectorZone> sectorZones)
			throws ActionCommandException {
		if (getMyPos() instanceof worldModel.Building
				&& (((worldModel.Building) getMyPos()).isOnFire() || (((((worldModel.Building) getMyPos())
						.isFirenessDefined()) && ((worldModel.Building) getMyPos())
						.getFieryness() == 8)))) {
			fireRisk((worldModel.Building) getMyPos());
		}
		SectorZone firstZone = myHum().sectorZone;
		ArrayList<worldModel.Building> myBuilds = new ArrayList<worldModel.Building>();
		done: while (true) {
			if (me() instanceof FireBrigade
					|| (me() instanceof PoliceForce && worldModel.policeForces
							.indexOf(me()) < worldModel.policeForces.size() / 2)) {
				myBuilds.addAll(buildingsGoodToSee(myHum().sectorZone.buildings));
				if (myBuilds.size() != 0) {
					int minLTS = getMinLTS(myBuilds);
					if (minLTS < 0)
						moveToSee(myBuilds, minLTS);
				}
				myHum().sectorZoneNum = (myHum().sectorZoneNum + 1)
						% sectorZones.size();
				myHum().sectorZone = sectorZones.get(myHum().sectorZoneNum);
				if (myHum().sectorZone == firstZone)
					break done;

			} else {
				myBuilds = reachableAndNotIgnitedBuildings(myHum().sectorZone.buildings);
				if (myBuilds.size() > 0) {
					ArrayList<Area> myAreas = new ArrayList<Area>();
					myAreas.addAll(myBuilds);
					int minLTV = getMinLTV(myAreas);
					if (minLTV < 0) {
						moveToNearest(myAreas, minLTV);
					}
				}
				myHum().sectorZoneNum = (myHum().sectorZoneNum + 1)
						% sectorZones.size();
				myHum().sectorZone = sectorZones.get(myHum().sectorZoneNum);
				if (myHum().sectorZone == firstZone)
					break done;
			}
		}
		myBuilds.clear();
		myBuilds.addAll(buildingsGoodToSee(worldModel.buildings));
		int minLTS = getMinLTS(myBuilds);
		moveToSee(myBuilds, minLTS);
	}

	public void greedySearch(ArrayList<SectorZone> sectorZones)
			throws ActionCommandException {
		SectorZone firstZone = myHum().sectorZone;
		ArrayList<worldModel.Building> myBuilds = new ArrayList<worldModel.Building>();
		ArrayList<Area> allBuilds = new ArrayList<Area>();
		allBuilds.addAll(worldModel.buildings);
		done: while (true) {
			// log("myBuilds: " + myBuilds);
			if (!(me() instanceof AmbulanceTeam)
					|| myHum().originalSectorZone == myHum().sectorZone) {
				myBuilds.addAll(buildingsGoodToSee(myHum().sectorZone.buildings));
				if (myBuilds.size() != 0) {
					int minLTS = getMinLTS(myBuilds);
					if (minLTS < 0)
						moveToSee(myBuilds, minLTS);
				}
			}
			myBuilds = reachableAndNotIgnitedBuildings(myHum().sectorZone.buildings);
			if (myBuilds.size() > 0) {
				ArrayList<Area> myAreas = new ArrayList<Area>();
				myAreas.addAll(myBuilds);
				int minLTV = getMinLTV(myAreas);
				if (minLTV == -1)
					moveToNearest(myAreas, minLTV);
			}
			myHum().sectorZoneNum = (myHum().sectorZoneNum + 1)
					% sectorZones.size();
			myHum().sectorZone = sectorZones.get(myHum().sectorZoneNum);
			if (myHum().sectorZone == firstZone)
				break done;
		}
		myBuilds.clear();
		myBuilds.addAll(buildingsGoodToSee(worldModel.buildings));
		int minLTS = getMinLTS(myBuilds);
		moveToSee(myBuilds, minLTS);
	}

	public ArrayList<worldModel.Building> reachableAndNotIgnitedBuildings(
			ArrayList<worldModel.Building> buildings) {
		ArrayList<worldModel.Building> goodBuilds = new ArrayList<worldModel.Building>();
		for (worldModel.Building b : buildings)
			if ((!b.isOnFire()
					&& (!b.isFirenessDefined() || b.getFieryness() == 0)
					&& (me() instanceof PoliceForce || b.isReachable) && !b
						.isNotIgnitable())
					&& b.canBeVisited
					&& (!b.serverBuilding().isBrokennessDefined() || b
							.serverBuilding().getBrokenness() > 0))
				goodBuilds.add(b);
		return goodBuilds;
	}

	public ArrayList<worldModel.Building> buildingsGoodToSee(
			ArrayList<worldModel.Building> myBuilds) {
		ArrayList<worldModel.Building> buildingsWithBuildingReachability = new ArrayList<worldModel.Building>();
		if (me() instanceof PoliceForce) {
			for (worldModel.Building b : myBuilds) {
				if (((PoliceForceAgent) this).policeBuildingReachability(b)
						&& (!b.isFirenessDefined() || b.getFieryness() != 8))
					buildingsWithBuildingReachability.add(b);
			}
		} else {
			for (worldModel.Building b : myBuilds) {
				if (me() instanceof FireBrigade)
					if (((FireBrigadeAgent) this).buildingReachability(b)
							&& (!b.isFirenessDefined() || b.getFieryness() != 8)) {
						buildingsWithBuildingReachability.add(b);
					}
				if (me() instanceof AmbulanceTeam)
					if (((AmbulanceTeamAgent) this).buildingReachability(b)
							&& (!b.isFirenessDefined() || b.getFieryness() != 8))
						buildingsWithBuildingReachability.add(b);
			}
		}
		return buildingsWithBuildingReachability;
	}

	// emergencyMove :
	public boolean isGoodForEmergencySearch(Area area) {
		if ((area instanceof worldModel.Building
				&& !((worldModel.Building) area).isOnFire() && ((worldModel.Building) area)
				.fireness() != 8) || area instanceof worldModel.Road)
			return true;
		return false;
	}

	// emergencySearch :
	public void emergencySearch() throws ActionCommandException {
		// log("emergencySearch");
		if (me() instanceof PoliceForce)
			return;
		// log("reachables  :  " + worldModel.reachableBuildings);
		if (worldModel.reachableBuildings.size() <= 5) {
			counterForFeedBack++;
			ArrayList<worldModel.Building> buildingsWithBuildingReachability = new ArrayList<worldModel.Building>();
			buildingsWithBuildingReachability
					.addAll(buildingsGoodToSee(worldModel.buildings));
			ArrayList<worldModel.Building> toRemove = new ArrayList<worldModel.Building>();
			for (worldModel.Building b : buildingsWithBuildingReachability) {
				if (worldModel.isInChangeSet(b.getID()) || b == getMyPos()) {
					toRemove.add(b);
				}
			}
			buildingsWithBuildingReachability.removeAll(toRemove);
			moveToSee(buildingsWithBuildingReachability,
					getMinLTS(buildingsWithBuildingReachability));
			HashSet<Area> goodAreas = new HashSet<Area>();
			int minLTV = Integer.MAX_VALUE;
			for (Entrance entrance : getMyPos().entrances)
				if (!worldModel.isInChangeSet(entrance.neighbour.area.getID()))
					goodAreas.add(entrance.neighbour.area);
			// goodAreas.addAll(worldModel.reachableAreas);
			goodAreas.remove(getMyPos());
			Area emergencyTarget = null;
			for (Area area : goodAreas)
				if (isGoodForEmergencySearch(area) && area.LTV <= minLTV)
					minLTV = area.LTV;
			int minDist = Integer.MAX_VALUE;
			for (Area area : goodAreas)
				if (isGoodForEmergencySearch(area) && area.LTV == minLTV)
					if (area.distanceFromSelf <= minDist) {
						emergencyTarget = area;
						minDist = area.distanceFromSelf;
					}
			if (emergencyTarget != null) {
				if (emergencyTarget.isReachable)
					move(emergencyTarget);
				else
					moveForPolice(emergencyTarget, emergencyTarget.getX(),
							emergencyTarget.getY(), false);
			}
			ArrayList<Area> reachableBuildings = new ArrayList<Area>();
			reachableBuildings.addAll(worldModel.reachableAreas);
			reachableBuildings.remove(getMyPos());
			moveToNearestMinLTV(reachableBuildings);
			randomWalk(false);
		}
		counterForFeedBack = 0;
	}

	protected void search() throws ActionCommandException {
		emergencySearch();
		if (noCommi)
			noCommiSearch(sectorZones);
		else
			greedySearch(sectorZones);
	}

	@SuppressWarnings("unchecked")
	private void centers() {
		if (time <= 3)
			return;
		if (me().getID().getValue() == centerForAmb && !noCommi) {
			if (pg == null)
				pg = new CenterForAmbulance(this);
			pg.act();
		}
		if (me().getID().getValue() == centerForFire && !noCommi) {
			if (kimpar == null)
				kimpar = new CenterForFire((ThinkAgent<StandardEntity>) this,
						logger);
			kimpar.act();
		}
		if (me().getID().getValue() == centerForPolice && !noCommi && !lowCommi) {
			if (safie == null)
				safie = new CenterForPolice((ThinkAgent<StandardEntity>) this,
						logger);
			safie.act();
		}
	}

	protected void randomWalkComputations() throws ActionCommandException {
		// log("randomWalkComputations");
		// log("RRT: " + RRT);

		if (amIStucking && me() instanceof PoliceForce
				&& ((PoliceForceAgent) this).getBlocksInRange(1000).size() > 0) {
			// for (Blockade b : this.getBlocksInRange(1000))
			// log("b::::::: " + GeometryTools.printShape(b.mainShape));
			((PoliceForceAgent) this).SclearNeareastBlock();
		}

		if (amIStucking == true
				&& Math.hypot(getX() - getMyPos().getX(), getY()
						- getMyPos().getY()) == 0) {
			// log("hyPot = "
			// + Math.hypot(getX() - getMyPos().getX(), getY()
			// - getMyPos().getY()));

			double maxDist = Double.MIN_VALUE;
			Coordinate myCoordinate = new Coordinate(getX(), getY());
			Geometry PosWB = getMyPos().shape;
			for (Blockade b : getMyPos().getBlockades()) {
				PosWB = PosWB.buffer(0).difference(b.mainShape.buffer(0));
			}
			// log("myCord: " + GeometryTools.printPoint(myCoordinate));
			Coordinate target = null;
			// log("Shape: " + GeometryTools.printShape(getMyPos().shape));
			for (Coordinate c : PosWB.getCoordinates()) {
				// log("Cord! " + GeometryTools.printPoint(c));
				// log ("dist? " + c.distance(myCoordinate));
				if (c.distance(myCoordinate) > maxDist) {
					for (Blockade b : getMyPos().getBlockades()) {
						Geometry point = GeometryTools.coordToGeometry(c,
								worldModel.getTrafficSimulator().geom).buffer(
								worldModel.getTrafficSimulator().agentRad);
						if (!b.mainShape.contains(point)
								&& !b.mainShape.intersects(point)) {
							target = c;
							maxDist = c.distance(myCoordinate);
						}
					}
				}
			}

			if (target != null) {
				// log("NextCord: " + GeometryTools.printPoint(target));
				int x = (int) target.x;
				int y = (int) target.y;
				List<EntityID> path = new ArrayList<EntityID>();
				path.add(getMyPos().getID());
				xLastMove = x;
				yLastMove = y;
				lastMoveCycle = time;
				// lastMovePath.clear();
				// lastMovePath.addAll(path);
				// lastMovePathTarget = path.get(path.size() - 1);
				sendMove(time, path, x, y);
				throw new ActionCommandException(StandardMessageURN.AK_MOVE);
				// move(getMyPos(), (int) target.x, (int) target.y, false);
			}
		}

		if ((RRT < RRTpathSize && RRT != 0) || (RRT == 0 && amIStucking)) {
			try {
				RRTComputations();
			} catch (Exception e) {
				if (e instanceof ActionCommandException)
					throw e;
				// System.err.println("ExceptionInRRTTTTTT!!");
				// e.printStackTrace();
				// if (logger.getWriter() != null)
				// e.printStackTrace(logger.getWriter());
				randomWalk(false);
			}
		}

		if (amIStucking)
			randomWalk(false); // it was true

		// if (me() instanceof PoliceForce) {
		// if (((PoliceForceAgent) this).getBlocksInRange(1500).size() > 0)
		// isSafeBlockadesNeededForRandomWalk = true;
		// else {
		// randomWalk(true);
		// }
		// } else {
		// }
		// if (counterForIsSafeBlock == 1) {
		// isSafeBlockadesNeededForRandomWalk = false;
		// counterForIsSafeBlock = 0;
		// }
		// if (isSafeBlockadesNeededForRandomWalk && stuckToolKeshid > 1) {
		// counterForIsSafeBlock++;
		// log("hello?? ");
		// ((PoliceForceAgent) this).SclearNeareastBlock();
		// randomWalk(true);
		// }
		// log("return?");
		return;
	}

	protected int maxX(Geometry shape) {
		int maxX = 0;
		for (Coordinate x : shape.getCoordinates())
			if (x.x > maxX)
				maxX = (int) x.x;
		return maxX;
	}

	protected int maxY(Geometry shape) {
		int maxY = 0;
		for (Coordinate y : shape.getCoordinates())
			if (y.y > maxY)
				maxY = (int) y.y;
		return maxY;
	}

	protected int minX(Geometry shape) {
		int minX = Integer.MAX_VALUE;
		for (Coordinate x : shape.getCoordinates())
			if (x.x < minX)
				minX = (int) x.x;
		return minX;
	}

	protected int minY(Geometry shape) {
		int minY = Integer.MAX_VALUE;
		for (Coordinate y : shape.getCoordinates())
			if (y.y < minY)
				minY = (int) y.y;
		return minY;
	}

	boolean didntSucceed = false;

	public void RRTComputations() throws ActionCommandException {
		didntSucceed = false;
		Coordinate lastTargetCoordinate = new Coordinate(xRealMove, yRealMove);
		Geometry myPosShape = getMyPos().shape;
		Coordinate start = new Coordinate(getX(), getY());
		if (lastMovePathTarget == null
				|| !worldModel.isInChangeSet(getMyPos().getID())) {
			return;
		}

		ArrayList<EntityID> movePath = new ArrayList<EntityID>();
		if (lastRealMovePath.size() > 0)
			movePath.addAll(lastRealMovePath);
		else
			movePath.addAll(lastMovePath);

		if ((lastTargetCoordinate.x != -1 && lastTargetCoordinate.y != -1)
				&& (lastMovePathTarget.getValue() == getMyPos().id || myPosShape
						.contains(GeometryTools.coordToGeometry(
								lastTargetCoordinate,
								worldModel.getTrafficSimulator().geom)))) {
			if (amIReachableToTarget(lastTargetCoordinate, movePath)
					&& !didntSucceed)
				RRTRandomWalk(start, lastTargetCoordinate);
		}
		// else
		if (movePath.contains(getMyPos().getID())) {
			Entrance nextEntrance = null;
			Entrance thisEntrance = null;
			int i = movePath.indexOf(getMyPos().getID());
			if (movePath.size() > i + 1) {
				for (Entrance e : getMyPos().entrances)
					if (e.neighbour.area.id == movePath.get(i + 1).getValue()) {
						// log("e::::: " +
						// GeometryTools.printLine(e.lineSegment));
						thisEntrance = e;
						nextEntrance = e.neighbour;
						break;
					}
				// if (entranceBetween.getPoints().size() >= 2) {
				// Coordinate p0 = entranceBetween.getPoints().get(0)
				// .getCoordinate();
				// Coordinate p1 = entranceBetween.getPoints().get(1)
				// .getCoordinate();
				// lastTargetCoordinate = new Coordinate((p0.x + p1.x) / 2,
				// (p0.y + p1.y) / 2);
				if (thisEntrance != null) {

					lastTargetCoordinate = nextEntrance.lineSegment.midPoint();
					Vector2D v = new Vector2D(lastTargetCoordinate.x - getX(),
							lastTargetCoordinate.y - getY());
					LineSegment l = new LineSegment(start, lastTargetCoordinate);
					v = v.normalised().scale(l.getLength() + 500);
					Vector2D v2 = new Vector2D(getX() + v.getX(), getY()
							+ v.getY());
					lastTargetCoordinate.x = v2.getX();
					lastTargetCoordinate.y = v2.getY();

					if (amIReachableToTarget(lastTargetCoordinate, movePath)
							&& !didntSucceed)
						RRTRandomWalk(start, lastTargetCoordinate);

					// if it couldnt:
					for (Geometry g : nextEntrance.getPoints()) {
						// log("points on ent: "
						// + GeometryTools.printShape(g.buffer(100)));

						lastTargetCoordinate = new Coordinate(
								g.getCoordinate().x, g.getCoordinate().y);

						v = new Vector2D(lastTargetCoordinate.x - getX(),
								lastTargetCoordinate.y - getY());
						l = new LineSegment(start, lastTargetCoordinate);
						v = v.normalised().scale(l.getLength() + 500);
						v2 = new Vector2D(getX() + v.getX(), getY() + v.getY());
						lastTargetCoordinate.x = v2.getX();
						lastTargetCoordinate.y = v2.getY();

						if (amIReachableToTarget(lastTargetCoordinate, movePath)
								&& !didntSucceed)
							RRTRandomWalk(start, lastTargetCoordinate);
					}
				}
			}
		}
		// else{
		// last
		// }

		// List<EntityID> movePath = new ArrayList<EntityID>();
		// movePath.add(getMyPos().getID());
		// if (getMyPos().areaWithOutBlock.getCentroid() == null)
		// return;
		// Coordinate centroid = getMyPos().areaWithOutBlock.getCentroid()
		// .getCoordinate();
		// if (xLastMove == centroid.x && yLastMove == centroid.y)
		// return;
		// log("centroid: " + GeometryTools.printPoint(centroid));
		// sendMove(time, movePath, (int) centroid.x, (int) centroid.y);
		// throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	private boolean amIReachableToTarget(Coordinate c,
			ArrayList<EntityID> movePath) {
		Coordinate myCoord = new Coordinate(getX(), getY());
		// log("mycoord: " + GeometryTools.printPoint(myCoord));
		// log("shape: " + GeometryTools.printShape(myShape));
		// log("c: " + GeometryTools.printPoint(c));

		Geometry myShape = getMyPos().completeShape;
		for (EntityID e : movePath) {
			Area a = worldModel.getArea(e);
			myShape = myShape.union(a.completeShape);
			if (a.completeShape.contains(GeometryTools.coordToGeometry(c,
					worldModel.getTrafficSimulator().geom))) {
				break;
			}

		}

		for (EntityID e : movePath) {
			Area a = worldModel.getArea(e);
			for (Blockade b : a.getBlockades())
				myShape = myShape.buffer(0).difference(b.mainShape.buffer(0));
			if (a.completeShape.contains(GeometryTools.coordToGeometry(c,
					worldModel.getTrafficSimulator().geom))) {
				break;
			}

		}

		ArrayList<Geometry> areaShapes = new ArrayList<>();
		for (int i = 0; i < myShape.getNumGeometries(); i++)
			areaShapes.add(myShape.getGeometryN(i));

		// log("shapeWOB" +
		// GeometryTools.printShape(getMyPos().areaWithOutBlock));
		for (Geometry g : areaShapes) {
			// log("shape-diff : " + GeometryTools.printShape(g));
			if (g.buffer(50).contains(
					GeometryTools.coordToGeometry(myCoord,
							worldModel.getTrafficSimulator().geom))) {
				if (g.buffer(50).contains(
						GeometryTools.coordToGeometry(c,
								worldModel.getTrafficSimulator().geom))) {
					return true;
				}
				break;
			}
		}

		return false;
	}

	protected void RRTRandomWalk(Coordinate start,
			Coordinate lastTargetCoordinate) throws ActionCommandException {
		// log("tagerCord: " + GeometryTools.printPoint(lastTargetCoordinate));

		// log("start: " + GeometryTools.printPoint(start));
		// log("dist: " + lastTargetCoordinate.distance(start) + 2000);
		for (Blockade b : getBlocksInRange((int) lastTargetCoordinate
				.distance(start) + 2000)) {

			// log("block: " +
			// GeometryTools.printShape(b.mainShape.buffer(-50)));
			if (b.mainShape.buffer(-50).contains(
					GeometryTools.coordToGeometry(start,
							worldModel.getTrafficSimulator().geom))
					|| b.mainShape.contains(GeometryTools.coordToGeometry(
							lastTargetCoordinate,
							worldModel.getTrafficSimulator().geom))
					|| b.mainShape.intersects(GeometryTools.coordToGeometry(
							lastTargetCoordinate,
							worldModel.getTrafficSimulator().geom)))
				return;
		}

		ArrayList<Coordinate> targCoordinates = new ArrayList<Coordinate>();
		ArrayList<Coordinate> RRTpath = new ArrayList<Coordinate>();
		Geometry myPosShape = getMyPos().completeShape.buffer(0);
		// log("MyPOSSHAPE: " + GeometryTools.printShape(myPosShape));

		for (Blockade b : getMyPos().getBlockades()) {
			myPosShape = myPosShape.buffer(0).difference(b.mainShape.buffer(0));
			// log("b: " + GeometryTools.printShape(b.mainShape.buffer(0)));
			// log("MP - b: " + GeometryTools.printShape(myPosShape));
		}

		ArrayList<Geometry> areaShapes = new ArrayList<>();
		for (int i = 0; i < myPosShape.getNumGeometries(); i++)
			areaShapes.add(myPosShape.getGeometryN(i));

		for (Geometry g : areaShapes) {
			// log("shape-diff : " + GeometryTools.printShape(g));
			if (g.contains(GeometryTools.coordToGeometry(new Coordinate(getX(),
					getY()), worldModel.getTrafficSimulator().geom))) {
				myPosShape = g.buffer(0);
			}
			// log("MyPOSSHAPE-BLOCKADE: " +
			// GeometryTools.printShape(myPosShape));
		}
		// int minDist = Integer.MAX_VALUE;
		int maxX = maxX(myPosShape);
		int maxY = maxY(myPosShape);
		int minX = minX(myPosShape);
		int minY = minY(myPosShape);
		Random r = new Random(me().getID().getValue() + time);
		Coordinate myCoordinate = start;
		for (int z = 0; z < 100; z++) {
			targCoordinates.clear();
			targCoordinates.add(myCoordinate);
			start = myCoordinate;
			boolean success = false;
			// String sf = " ";
			done: for (int i = 0; i < 500; i++) {
				int randx = r.nextInt(maxX - minX) + minX;
				int randy = r.nextInt(maxY - minY) + minY;
				Coordinate randCoord = new Coordinate(randx, randy);

				try {
					if (!myPosShape.contains(GeometryTools.coordToGeometry(
							randCoord, worldModel.getTrafficSimulator().geom)))
						continue;

				} catch (Exception e) {
					continue;
				}

				LineSegment l = new LineSegment(start.x, start.y, randCoord.x,
						randCoord.y);
				// sf += GeometryTools.printPoint(randCoord);
				for (Blockade b : getMyPos().blockades) {
					// if(b.circumstance.intersects(l.toGeometry(worldModel
					// .getTrafficSimulator().geom)))
					// log("l: " + GeometryTools.printLine(l));
					if (b.mainShape.buffer(-50)
							.intersects(
									l.toGeometry(worldModel
											.getTrafficSimulator().geom)))
						continue done;
				}
				if (getMyPos() instanceof worldModel.Building) {
					for (Edge e : getMyPos().serverArea.getEdges()) {
						if (!e.isPassable()) {
							LineSegment edgeLine = new LineSegment(
									e.getStartX(), e.getStartY(), e.getEndX(),
									e.getEndY());
							if (edgeLine.intersection(l) != null)
								continue done;
						}
					}
				}

				targCoordinates.add(randCoord);
				start = randCoord;
				LineSegment l2 = new LineSegment(start, lastTargetCoordinate);
				for (Blockade b : getMyPos().blockades)
					if (b.mainShape.contains(l2.toGeometry(worldModel
							.getTrafficSimulator().geom))
							|| b.mainShape.intersects(l2.toGeometry(worldModel
									.getTrafficSimulator().geom)))
						continue done;
				targCoordinates.add(lastTargetCoordinate);
				success = true;
				break;
			}
			if (!success)
				continue;
			// int dist = 0;
			// for (int j = 0; j < targCoordinates.size() - 1; j++)
			// dist += Math
			// .hypot(targCoordinates.get(j).x
			// - targCoordinates.get(j + 1).x,
			// targCoordinates.get(j).y
			// - targCoordinates.get(j + 1).y);
			// if (minDist > dist) {
			// minDist = dist;
			// RRTpath.clear();
			// RRTpath.addAll(targCoordinates);
			// }

			if (targCoordinates.size() > 1) {
				RRTpath.clear();
				RRTpath.addAll(targCoordinates);
				break;
			}
		}
		// time11 = System.nanoTime();
		List<EntityID> movePath = new ArrayList<EntityID>();

		if (RRTpath.size() < 2) {
			didntSucceed = true;
			return;
		}

		for (int i = 0; i < RRTpath.size() - 2; i++) {
			done: for (int j = RRTpath.size() - 1; j > i; j--) {
				LineSegment l = new LineSegment(RRTpath.get(j).x,
						RRTpath.get(j).y, RRTpath.get(i).x, RRTpath.get(i).y);
				// log("ll: " + GeometryTools.printLine(l));
				for (Blockade b : getMyPos().blockades)
					if (b.mainShape.contains(l.toGeometry(worldModel
							.getTrafficSimulator().geom))
							|| b.mainShape.intersects(l.toGeometry(worldModel
									.getTrafficSimulator().geom))) {
						// log("LSsss: " + GeometryTools.printLine(l));
						// log("i " + i + " j " + j);
						continue done;
					}

				while (j - 1 > i) {
					RRTpath.remove(j - 1);
					j--;
				}
			}
		}

		RRTpathSize = RRTpath.size();// - 1;
		movePath.add(getMyPos().getID());
		if (lastMovePath.indexOf(getMyPos()) > lastMovePath.size() - 1)
			movePath.add(lastMovePath.get(lastMovePath.indexOf(getMyPos()) + 1));
		// TODO:add khune badi!!!
		xLastMove = (int) RRTpath.get(1).x;
		yLastMove = (int) RRTpath.get(1).y;
		lastMoveCycle = time;
		RRT++;
		sendMove(time, movePath, (int) RRTpath.get(1).x, (int) RRTpath.get(1).y);
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	protected void initMoveToFire() {
		for (Area area : worldModel.areas)
			for (Entrance ents : area.entrances) {
				ents.distForDijk = 1000000000;
				ents.prev = null;
				ents.isTarget = false;
			}
	}

	protected void randomWalk(boolean RRT) throws ActionCommandException {
		if (RRT)
			RRTComputations();
		if ((me() instanceof PoliceForce || me() instanceof AmbulanceTeam)
				&& wasIInSearch)
			amIInSearch = true;
		List<EntityID> result = new ArrayList<EntityID>(RANDOM_WALK_LENGTH);
		Set<EntityID> seen = new HashSet<EntityID>();
		EntityID current = getMyPos().getID();
		for (int i = 0; i < RANDOM_WALK_LENGTH; ++i) {
			result.add(current);
			seen.add(current);
			List<EntityID> toBeRemoved = new ArrayList<EntityID>();
			List<EntityID> possible = new ArrayList<EntityID>(
					neighbours.get(current));
			for (EntityID fired : possible) {
				Entity e = worldModel.getEntity(fired);
				if ((e instanceof Building && ((worldModel.Building) ((Building) e).area)
						.isOnFire())
						&& ((worldModel.Building) ((Building) e).area).isReachable)
					toBeRemoved.add(fired);
			}
			if (toBeRemoved.size() == possible.size()) {
				toBeRemoved = new ArrayList<EntityID>();
				for (EntityID fired : possible) {
					if ((worldModel.getEntity(fired)) instanceof Building
							&& ((Building) worldModel.getEntity(fired))
									.isFierynessDefined()
							&& ((Building) worldModel.getEntity(fired))
									.getFieryness() != 0)
						toBeRemoved.add(fired);
				}
			}
			possible.removeAll(toBeRemoved);
			ArrayList<EntityID> toAddUpWeight = new ArrayList<EntityID>();
			for (EntityID e : possible)
				if (worldModel.getArea(e).isBlockadesDefined()
						&& worldModel.getArea(e).blockades.size() == 0)
					toAddUpWeight.add(e);
			possible.addAll(toAddUpWeight);
			Collections.shuffle(possible, new Random(time
					+ me().getID().getValue()));
			boolean found = false;
			for (EntityID next : possible) {
				if (seen.contains(next)) {
					continue;
				}
				current = next;
				found = true;
				break;
			}
			if (!found) {
				// We reached a dead-end.
				break;
			}
		}

		int x = ((rescuecore2.standard.entities.Area) model.getEntity(result
				.get(result.size() - 1))).getX();
		int y = ((rescuecore2.standard.entities.Area) model.getEntity(result
				.get(result.size() - 1))).getY();
		xLastMove = x;
		yLastMove = y;
		lastMoveCycle = time;
		lastMovePath.clear();
		lastMovePath.addAll(result);
		lastMovePathTarget = result.get(result.size() - 1);
		sendMove(time, result, x, y);
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	protected abstract void act() throws ActionCommandException;
}