package radar;

import java.util.BitSet;

import worldModel.Area;
import worldModel.Building;
import worldModel.Road;
import worldModel.WorldModel;

public class Package {
	public BitSet bits;
	public int start, hNum = Calulations.findPow(HeaderType.values().length);

	// HEADERS
	byte[] POS = { 0, 0, 0, 0 };
	byte[] REFUGE = { 0, 0, 0, 1 };
	byte[] FIRED = { 0, 0, 1, 0 };
	byte[] CV = { 0, 0, 1, 1 };
	byte[] WAY = { 0, 1, 0, 0 };
	byte[] noWAY = { 0, 1, 0, 1 };
	byte[] outFired = { 0, 1, 1, 0 };
	byte[] HELP = { 0, 1, 1, 1 };
	byte[] ACENTER = { 1, 0, 0, 0 };
	byte[] Info = { 1, 0, 0, 1 };
	byte[] PCENTER = { 1, 0, 1, 0 };
	byte[] FCENTER = { 1, 0, 1, 1 };
	byte[] UHELP = { 1, 1, 0, 0 };
	byte[] USTUCK = { 1, 1, 0, 1 };
	byte[] STUCK = { 1, 1, 1, 0 };
	byte[] CLEARER = { 1, 1, 1, 1 };
	int iPOS = 0;
	int iREFUGE = 1;
	int iFIRED = 2;
	int iCV = 3;
	int iWAY = 4;
	int iNOWAY = 5;
	int iOUTFIRED = 6;
	int iHELP = 7;
	int iACENTER = 8;
	int iINFO = 9;
	int iPCENTER = 10;
	int iFCENTER = 11;
	int iUHELP = 12;
	int iUSTUCK = 13;
	int iSTUCK = 14;
	int iCLEARER = 15;

	public int timeSaid = -1;

	public int areaPow, buildingPow, roadPow;

	public int headerPow = Calulations.findPow(HeaderType.values().length);
	public int hearingPointer = 0;
	WorldModel model;

	public static enum HeaderType {
		Help, Fired, CV, NoWay, Way, OutFired, Pos, ACenter, Info, PCenter, FCenter, UHelp, UStuck, Stuck, Clearer, Refuge
	}

	public static enum AreaType {
		Area, Building, Road
	}

	public Package(WorldModel model) {
		bits = new BitSet();
		start = 0;
		hearingPointer = 0;
		this.model = model;
		areaPow = Calulations.findPow(model.areas.size());
		buildingPow = Calulations.findPow(model.buildings.size());
		roadPow = Calulations.findPow(model.roads.size());
	}

	public void addInt(int num, int amount) {
		start = Calulations.calcWithBitSet(bits, num, start,
				Calulations.findPow(amount));
	}

	public int findLastBit() {
		int i = bits.size() - 1;
		while (!bits.get(i))
			i--;
		return i;
	}

	public void addHeader(HeaderType ht) {
		byte[] thisB = null;
		if (ht == HeaderType.Help)
			thisB = HELP;
		else if (ht == HeaderType.Refuge)
			thisB = REFUGE;
		else if (ht == HeaderType.Fired)
			thisB = FIRED;
		else if (ht == HeaderType.CV)
			thisB = CV;
		else if (ht == HeaderType.Way)
			thisB = WAY;
		else if (ht == HeaderType.NoWay)
			thisB = noWAY;
		else if (ht == HeaderType.OutFired)
			thisB = outFired;
		else if (ht == HeaderType.Pos)
			thisB = POS;
		else if (ht == HeaderType.ACenter)
			thisB = ACENTER;
		else if (ht == HeaderType.Info)
			thisB = Info;
		else if (ht == HeaderType.PCenter)
			thisB = PCENTER;
		else if (ht == HeaderType.FCenter)
			thisB = FCENTER;
		else if (ht == HeaderType.UHelp)
			thisB = UHELP;
		else if (ht == HeaderType.UStuck)
			thisB = USTUCK;
		else if (ht == HeaderType.Stuck)
			thisB = STUCK;
		else if (ht == HeaderType.Clearer)
			thisB = CLEARER;
		if (thisB != null) {
			for (int i = 0; i < thisB.length; i++) {
				if (thisB[i] == 1)
					bits.set(i, true);
				else
					bits.set(i, false);
			}
			start = thisB.length;
		}
	}

	public void addArea(Area area, AreaType at) {
		if (at == AreaType.Building)
			start = Calulations.calcWithBitSet(bits,
					model.buildingToIndex.get((Building) area), start,
					buildingPow);
		else if (at == AreaType.Road)
			start = Calulations.calcWithBitSet(bits,
					model.roadToIndex.get((Road) area), start, roadPow);
		else
			start = Calulations.calcWithBitSet(bits,
					model.areaToIndex.get(area), start, areaPow);
	}

	public void addID(int idNum) {
		start = Calulations.calcWithBitSet(bits, idNum, start, 32);
	}

	public void setEnd() {
		start = Calulations.calcWithBitSet(bits, 1, start, 1);
	}

	public byte[] send() {
		return Calulations.toByteArray(bits);
	}

	public void hear(byte[] bytes) {
		bits = Calulations.fromByteArray(bytes);
	}

	public HeaderType tellHeader() {
		int header = Calulations.fromBase2(bits, 0, headerPow);
		hearingPointer += headerPow;
		if (header == iHELP)
			return HeaderType.Help;
		else if (header == iREFUGE)
			return HeaderType.Refuge;
		else if (header == iFIRED)
			return HeaderType.Fired;
		else if (header == iCV)
			return HeaderType.CV;
		else if (header == iWAY)
			return HeaderType.Way;
		else if (header == iNOWAY)
			return HeaderType.NoWay;
		else if (header == iOUTFIRED)
			return HeaderType.OutFired;
		else if (header == iPOS)
			return HeaderType.Pos;
		else if (header == iACENTER)
			return HeaderType.ACenter;
		else if (header == iINFO)
			return HeaderType.Info;
		else if (header == iPCENTER)
			return HeaderType.PCenter;
		else if (header == iFCENTER)
			return HeaderType.FCenter;
		else if (header == iUHELP)
			return HeaderType.UHelp;
		else if (header == iUSTUCK)
			return HeaderType.UStuck;
		else if (header == iSTUCK)
			return HeaderType.Stuck;
		else if (header == iCLEARER)
			return HeaderType.Clearer;
		return null;
	}

	public boolean tellOffSet() {
		int offSet = getInt(2);
		if (offSet == 1)
			return true;
		return false;
	}

	public int getID() {
		int powNum = 32;
		int num = Calulations.fromBase2(bits, hearingPointer, powNum);
		hearingPointer += powNum;
		return num;
	}

	public Area getArea(AreaType at, WorldModel wm) {
		if (at == AreaType.Building) {
			Building b = wm.buildings.get(Calulations.fromBase2(bits,
					hearingPointer, buildingPow));
			hearingPointer += buildingPow;
			return b;
		} else if (at == AreaType.Road) {
			Road r = wm.roads.get(Calulations.fromBase2(bits, hearingPointer,
					roadPow));
			hearingPointer += roadPow;
			return r;
		} else {
			Area a = wm.areas.get(Calulations.fromBase2(bits, hearingPointer,
					areaPow));
			hearingPointer += areaPow;
			return a;
		}
	}

	public int getInt(int amount) {
		int powNum = Calulations.findPow(amount);
		int num = Calulations.fromBase2(bits, hearingPointer, powNum);
		hearingPointer += powNum;
		return num;
	}

	public int getTime(int timeLimit) {
		return Calulations.fromBase2(bits,
				findLastBit() - Calulations.findPow(timeLimit),
				Calulations.findPow(timeLimit));
	}

	public void setBits(BitSet subBits, int end) {
		for (int i = 0; i < end; i++)
			if (subBits.get(i))
				bits.set(start + i);
		start += end;
	}

	public BitSet getBits(int start, int end) {
		BitSet subBits = new BitSet();
		for (int i = start; i < end; i++)
			if (bits.get(i))
				subBits.set(i);
		hearingPointer = end;
		return subBits;
	}

	public int[] decodeForClearer() {
		Area area = getArea(AreaType.Road, model);
		int[] array = new int[area.entrances.size() + 1];
		array[0] = area.getID().getValue();
		for (int i = hearingPointer; i < area.entrances.size() + hearingPointer; i++) {
			if (bits.get(i) == true)
				array[i - hearingPointer + 1] = 1;
			else
				array[i - hearingPointer + 1] = 0;
		}
		return array;
	}

	public void addClearer(Road area) {
		addArea(area, AreaType.Road);
		for (int i = 0; i < area.entrances.size(); i++)
			bits.set(i + start, area.entrances.get(i).isItConnectedToNeighbour);
		start += area.entrances.size();
	}
}
