package rescuecore2.standard.entities;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map.Entry;

import rescuecore2.components.AbstractAgent;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.Property;
import rescuecore2.worldmodel.WorldModel;
import rescuecore2.worldmodel.properties.EntityRefProperty;
import rescuecore2.worldmodel.properties.IntArrayProperty;
import rescuecore2.worldmodel.properties.IntProperty;
import worldModel.Area;
import worldModel.Building;
import worldModel.Entrance;
import worldModel.Road;
import agent.Agent;
import agent.Agent.MapType;
import agent.GroupForPolice;
import agent.SectorZone;

/**
 * Abstract base class for Humans.
 */
public abstract class Human extends StandardEntity {
	private IntProperty x;
	private IntProperty y;
	private EntityRefProperty position;
	private IntArrayProperty positionHistory;
	private IntProperty travelDistance;
	private IntProperty direction;
	private IntProperty stamina;
	private IntProperty hp;
	private IntProperty damage;
	private IntProperty buriedness;
	public int zone = -1;

	public boolean isOK = false;
	public boolean isAlive = false;
	public boolean hasBuriedness = false;
	public HashMap<Area, Integer> distance = new HashMap<Area, Integer>();
	public boolean bfsPoliceMark = false;
	public boolean haveEnoughSupport = false;
	public GroupForPolice lastStuckGroup = null;
	public int pogNum = -1;

	public Area firstPosition = null;
	public int firstX = -1;
	public int firstY = -1;

	public Area lastCyclePos = null;
	public Area lastUpdatedPos = null;
	public int lastPosUpdatedTime = -1;

	public SectorZone sectorZone = null;
	public int sectorZoneNum = 0;
	public SectorZone originalSectorZone = null;
	// 2014
	public EntityID myNextTarget = null;

	// public SearchZone searchZone = null;

	/**
	 * Construct a Human object with entirely undefined property values.
	 * 
	 * @param id
	 *            The ID of this entity.
	 */
	protected Human(EntityID id) {
		super(id);
		x = new IntProperty(StandardPropertyURN.X);
		y = new IntProperty(StandardPropertyURN.Y);
		travelDistance = new IntProperty(StandardPropertyURN.TRAVEL_DISTANCE);
		position = new EntityRefProperty(StandardPropertyURN.POSITION);
		positionHistory = new IntArrayProperty(
				StandardPropertyURN.POSITION_HISTORY);
		direction = new IntProperty(StandardPropertyURN.DIRECTION);
		stamina = new IntProperty(StandardPropertyURN.STAMINA);
		hp = new IntProperty(StandardPropertyURN.HP);
		damage = new IntProperty(StandardPropertyURN.DAMAGE);
		buriedness = new IntProperty(StandardPropertyURN.BURIEDNESS);
		registerProperties(x, y, position, positionHistory, travelDistance,
				direction, stamina, hp, damage, buriedness);
	}

	/**
	 * Human copy constructor.
	 * 
	 * @param other
	 *            The Human to copy.
	 */
	public Human(Human other) {
		super(other);
		x = new IntProperty(other.x);
		y = new IntProperty(other.y);
		travelDistance = new IntProperty(other.travelDistance);
		position = new EntityRefProperty(other.position);
		positionHistory = new IntArrayProperty(other.positionHistory);
		direction = new IntProperty(other.direction);
		stamina = new IntProperty(other.stamina);
		hp = new IntProperty(other.hp);
		damage = new IntProperty(other.damage);
		buriedness = new IntProperty(other.buriedness);
		registerProperties(x, y, position, positionHistory, travelDistance,
				direction, stamina, hp, damage, buriedness);
	}

	public int compareTo(Human other) {
		if (deadtime > other.deadtime)
			return 1;
		else if (deadtime < other.deadtime)
			return -1;
		return 0;
	}

	public int deadtime = 502;
	public int lastDeadTimeSaid = -1;
	public int employedDeadTime = -1;
	public SectorZone noCommiSectorZone;
	public int noCommisectorZoneNum;
	public SectorZone noCommioriginalSectorZone;

	public int getDeadTime() {
		return deadtime;
	}

	public void setDmg(int dmg, int time) {
		this.damage.setValue(dmg);
	}

	public void setHp(int hp, int time) {
		this.hp.setValue(hp);
	}

	public void setBury(int bury) {
		this.buriedness.setValue(bury);
	}

	public boolean isGoodForEstFire(worldModel.WorldModel wm) {
		// sharte ezafe tar vase cv ha
		// bi refuge
		/*
		 * if (this instanceof Civilian) System.out.println("VVVVVV" + this +
		 * "--" + isPositionDefined() + isXDefined() + isYDefined() +
		 * isDamageDefined() + hp.getValue() + "----" + hasDamage);
		 */
		if (isPositionDefined()
				&& isDamageDefined()
				&& hp.getValue() != 0
				&& wm.getEntity(getPosition()) instanceof rescuecore2.standard.entities.Building
				&& !((Building) wm.getArea(getPosition())).isOnFire()
				&& !((Building) wm.getArea(getPosition())).isNotIgnitable()
				&& (this instanceof Civilian || hasBuriedness))
			return true;
		return false;
	}

	public void setDT(int time,
			ArrayList<worldModel.Building> ignitedBuildings,
			worldModel.WorldModel WM, boolean emergency) {
		if (hp.getValue() == 0) {
			deadtime = 0;
			return;
		}
		int two = Integer.MAX_VALUE;
		int one;
		int HP = increaseAccuracy(hp.getValue(), damage.getValue(), time);

		one = newDamage(HP, damage.getValue(), time, ignitedBuildings, WM);

		// +oon isDmgDefined ye chiz dg

		if (ignitedBuildings.size() != 0 && isGoodForEstFire(WM))
			two = predictingFire(time, ignitedBuildings, WM, emergency);
		// if (getID().getValue() == 385666250)
		// WM.simAgent.log("Dead:" + one + "--" + two);
		deadtime = Math.min(one, two);
	}

	int bank = 0;
	int lasthp = 0;
	int lasttime = 0;

	public int increaseAccuracy(int hp, int dmg, int time) {
		bank = hp;
		if ((this instanceof Civilian) && dmg < 95 && dmg > 0 && (lasthp != 0)
				&& time - lasttime == 1)
			hp = ((hp + lasthp) / 2);
		lasthp = bank;
		lasttime = time;
		return hp;

	}

	int lastHp = 0;
	int lastTime = 0;
	int firsTime = 0;

	public int newDamage(int hp, double damage, int newTime,
			ArrayList<worldModel.Building> ignitedBuildings,
			worldModel.WorldModel WM) {
		double bank = 0;
		int agenttime = 1000;
		int count = agenttime - newTime;
		int death = 1000;
		int F = 0;
		int wait = newTime - lastTime;
		lastTime = newTime;
		double accuracy = 0;
		double mean = 0;
		double z = 0;
		double y = damage;
		accuracy = lastHp - hp;
		lastHp = hp;
		if (wait == 1 && accuracy != 0) {
			F = newTime - firsTime;
			bank = firsTime;
			firsTime = newTime;
			if (F != 1 && this instanceof Civilian
					&& ((bank != 0) || (bank == 0 && damage == 0))) {
				if (accuracy % F == 0)
					mean = ((accuracy / F) + (accuracy / (F - 1))) / 2;
				else if (accuracy / (F - 1) == 0)
					mean = ((accuracy / (F + 1)) + (accuracy / (F - 1))) / 2;
				else
					mean = ((accuracy / (F + 1)) + (accuracy / (F - 2))) / 2;

				z = mean;
				for (int i = 0; i < (int) F / 2; i++)
					mean += (0.00025 * mean * mean) + 0.11D;
				damage = mean;
			} else {
				death = estimatedDeathTime(hp, damage, newTime);
				return death;
			}
			/*
			 * System.out.println("oneee" + this + "  time:" + newTime +
			 * "  firstTime" + bank + "  firstDmg:" + y + "  mean" + z +
			 * "  lastDmg:" + damage + "  accuracy" + accuracy);
			 */
			while (count > 0) {
				int time = agenttime - count;
				damage += (0.00025 * damage * damage) + 0.11D;
				count--;
				hp = (int) (hp - damage);

				if (hp <= 0)
					return time;

			}
			return 1000;
		} else {
			death = estimatedDeathTime(hp, damage, newTime);
		}
		return death;
	}

	// ZJU Deadtime codes
	public int estimatedDeathTime(int hp, double dmg, int updatetime) {
		int agenttime = 1000;
		int count = agenttime - updatetime;
		if ((count <= 0) || (dmg == 0.0D)) {
			return hp;
		}
		double kbury = 3.5E-05D;
		double kcollapse = 0.00025D;
		double darsadbury = -0.0014D * updatetime + 0.64D;
		double burydamage = dmg * darsadbury;
		double collapsedamage = dmg - burydamage;

		while (count > 0) {
			int time = agenttime - count;

			burydamage += kbury * burydamage * burydamage + 0.11D;
			collapsedamage += kcollapse * collapsedamage * collapsedamage
					+ 0.11D;
			dmg = burydamage + collapsedamage;
			count--;
			hp = (int) (hp - dmg);

			if (hp <= 0)
				return time;

		}
		return 1000;
	}

	// end of code dozD 2

	public int predictingFire(int time,
			ArrayList<worldModel.Building> ignitedBuildings,
			worldModel.WorldModel WM, boolean emergency) {
		Building nearest = null;
		float min = Integer.MAX_VALUE;
		float distance = 0;
		float layer = 0;
		int deathTime = 1000;
		int limit = 0;
		for (worldModel.Building igBuilding : ignitedBuildings) {
			distance = (float) Math.hypot(WM.getArea(getPosition()).getX()
					- igBuilding.getX(), WM.getArea(getPosition()).getY()
					- igBuilding.getY());
			if (distance < min) {
				min = distance;
				nearest = igBuilding;
			}
		}
		limit = mapRecognizer(WM);
		/*
		 * if (limit != 0) { if (distance % limit == 0) layer = (int) (distance
		 * / limit); else layer = (int) (distance / limit + 1);
		 * 
		 * for (worldModel.Building igBuilding : ignitedBuildings) { distance =
		 * (Double) (layer * limit - Math.hypot(
		 * WM.getArea(getPosition()).getX() - igBuilding.getX(),
		 * WM.getArea(getPosition()).getY() - igBuilding.getY())); if (distance
		 * > 0) { distances.put(igBuilding, distance); if (distance < min) min =
		 * distance; } } bank = min; min = Integer.MAX_VALUE; for
		 * (Entry<Building, Double> dist : distances.entrySet()) { if
		 * (dist.getValue() == bank) { timeOfStartOfFire =
		 * dist.getKey().timeOfStartOfFire; if (timeOfStartOfFire < min) { min =
		 * timeOfStartOfFire; } } }
		 * 
		 * deathTime = (int) (layer * 30 + min);
		 */
		if (emergency)
			deathTime = nearest.timeOfStartOfFire + 10;
		else {
			layer = (float) min / limit;
			deathTime = (int) Math.ceil(layer * 30 + nearest.timeOfStartOfFire);
		}
		/*
		 * } else return Integer.MAX_VALUE;
		 */
		return deathTime;
	}

	public int mapRecognizer(worldModel.WorldModel WM) {
		if (WM.map == MapType.Unknown)
			return 35000;
		if (WM.map == MapType.Kobe)
			return 30000;
		if (WM.map == MapType.VC)
			return 40000;
		if (WM.map == MapType.Test)
			return 10000;
		if (WM.map == MapType.Joao)
			return 80000;
		if (WM.map == MapType.Eindhoven)
			return 140000;
		if (WM.map == MapType.Berlin)
			return 80000;
		if (WM.map == MapType.Paris)
			return 55000;
		if (WM.map == MapType.Istanbul)
			return 70000;
		if (WM.map == MapType.Mecxico)
			return 40000;
		if (WM.map == MapType.NY)
			return 43000;
		return 0;
	}

	@Override
	public Property getProperty(String urn) {
		StandardPropertyURN type;
		try {
			type = StandardPropertyURN.fromString(urn);
		} catch (IllegalArgumentException e) {
			return super.getProperty(urn);
		}
		switch (type) {
		case POSITION:
			return position;
		case POSITION_HISTORY:
			return positionHistory;
		case DIRECTION:
			return direction;
		case STAMINA:
			return stamina;
		case HP:
			return hp;
		case X:
			return x;
		case Y:
			return y;
		case DAMAGE:
			return damage;
		case BURIEDNESS:
			return buriedness;
		case TRAVEL_DISTANCE:
			return travelDistance;
		default:
			return super.getProperty(urn);
		}
	}

	@Override
	public Pair<Integer, Integer> getLocation(
			WorldModel<? extends StandardEntity> world) {
		if (x.isDefined() && y.isDefined()) {
			return new Pair<Integer, Integer>(x.getValue(), y.getValue());
		}
		if (position.isDefined()) {
			EntityID pos = getPosition();
			StandardEntity e = world.getEntity(pos);
			return e.getLocation(world);
		}
		return null;
	}

	/**
	 * Get the position property.
	 * 
	 * @return The position property.
	 */
	public EntityRefProperty getPositionProperty() {
		return position;
	}

	/**
	 * Get the position of this human.
	 * 
	 * @return The position.
	 */
	public EntityID getPosition() {
		return position.getValue();
	}

	/**
	 * Set the position of this human.
	 * 
	 * @param position
	 *            The new position.
	 */
	public void setPosition(EntityID position) {
		this.position.setValue(position);
	}

	/**
	 * Find out if the position property has been defined.
	 * 
	 * @return True if the position property has been defined, false otherwise.
	 */
	public boolean isPositionDefined() {
		return position.isDefined();
	}

	/**
	 * Undefine the position property.
	 */
	public void undefinePosition() {
		position.undefine();
	}

	/**
	 * Get the position history property.
	 * 
	 * @return The position history property.
	 */
	public IntArrayProperty getPositionHistoryProperty() {
		return positionHistory;
	}

	/**
	 * Get the position history.
	 * 
	 * @return The position history.
	 */
	public int[] getPositionHistory() {
		return positionHistory.getValue();
	}

	/**
	 * Set the position history.
	 * 
	 * @param history
	 *            The new position history.
	 */
	public void setPositionHistory(int[] history) {
		this.positionHistory.setValue(history);
	}

	/**
	 * Find out if the position history property has been defined.
	 * 
	 * @return True if the position history property has been defined, false
	 *         otherwise.
	 */
	public boolean isPositionHistoryDefined() {
		return positionHistory.isDefined();
	}

	/**
	 * Undefine the position history property.
	 */
	public void undefinePositionHistory() {
		positionHistory.undefine();
	}

	/**
	 * Get the direction property.
	 * 
	 * @return The direction property.
	 */
	public IntProperty getDirectionProperty() {
		return direction;
	}

	/**
	 * Get the direction.
	 * 
	 * @return The direction.
	 */
	public int getDirection() {
		return direction.getValue();
	}

	/**
	 * Set the direction.
	 * 
	 * @param direction
	 *            The new direction.
	 */
	public void setDirection(int direction) {
		this.direction.setValue(direction);
	}

	/**
	 * Find out if the direction property has been defined.
	 * 
	 * @return True if the direction property has been defined, false otherwise.
	 */
	public boolean isDirectionDefined() {
		return direction.isDefined();
	}

	/**
	 * Undefine the direction property.
	 */
	public void undefineDirection() {
		direction.undefine();
	}

	/**
	 * Get the stamina property.
	 * 
	 * @return The stamina property.
	 */
	public IntProperty getStaminaProperty() {
		return stamina;
	}

	/**
	 * Get the stamina of this human.
	 * 
	 * @return The stamina.
	 */
	public int getStamina() {
		return stamina.getValue();
	}

	/**
	 * Set the stamina of this human.
	 * 
	 * @param stamina
	 *            The new stamina.
	 */
	public void setStamina(int stamina) {
		this.stamina.setValue(stamina);
	}

	/**
	 * Find out if the stamina property has been defined.
	 * 
	 * @return True if the stamina property has been defined, false otherwise.
	 */
	public boolean isStaminaDefined() {
		return stamina.isDefined();
	}

	/**
	 * Undefine the stamina property.
	 */
	public void undefineStamina() {
		stamina.undefine();
	}

	/**
	 * Get the hp property.
	 * 
	 * @return The hp property.
	 */
	public IntProperty getHPProperty() {
		return hp;
	}

	/**
	 * Get the hp of this human.
	 * 
	 * @return The hp of this human.
	 */
	public int getHP() {
		return hp.getValue();
	}

	/**
	 * Set the hp of this human.
	 * 
	 * @param newHP
	 *            The new hp.
	 */
	public void setHP(int newHP) {
		this.hp.setValue(newHP);
	}

	/**
	 * Find out if the hp property has been defined.
	 * 
	 * @return True if the hp property has been defined, false otherwise.
	 */
	public boolean isHPDefined() {
		return hp.isDefined();
	}

	/**
	 * Undefine the hp property.
	 */
	public void undefineHP() {
		hp.undefine();
	}

	/**
	 * Get the damage property.
	 * 
	 * @return The damage property.
	 */
	public IntProperty getDamageProperty() {
		return damage;
	}

	/**
	 * Get the damage of this human.
	 * 
	 * @return The damage of this human.
	 */
	public int getDamage() {
		return damage.getValue();
	}

	/**
	 * Set the damage of this human.
	 * 
	 * @param damage
	 *            The new damage.
	 */
	public void setDamage(int damage) {
		this.damage.setValue(damage);
	}

	/**
	 * Find out if the damage property has been defined.
	 * 
	 * @return True if the damage property has been defined, false otherwise.
	 */
	public boolean isDamageDefined() {
		return damage.isDefined();
	}

	/**
	 * Undefine the damage property.
	 */
	public void undefineDamage() {
		damage.undefine();
	}

	/**
	 * Get the buriedness property.
	 * 
	 * @return The buriedness property.
	 */
	public IntProperty getBuriednessProperty() {
		return buriedness;
	}

	/**
	 * Get the buriedness of this human.
	 * 
	 * @return The buriedness of this human.
	 */
	public int getBuriedness() {
		return buriedness.getValue();
	}

	/**
	 * Set the buriedness of this human.
	 * 
	 * @param buriedness
	 *            The new buriedness.
	 */
	public void setBuriedness(int buriedness) {
		this.buriedness.setValue(buriedness);
	}

	/**
	 * Find out if the buriedness property has been defined.
	 * 
	 * @return True if the buriedness property has been defined, false
	 *         otherwise.
	 */
	public boolean isBuriednessDefined() {
		return buriedness.isDefined();
	}

	/**
	 * Undefine the buriedness property.
	 */
	public void undefineBuriedness() {
		buriedness.undefine();
	}

	/**
	 * Get the X property.
	 * 
	 * @return The X property.
	 */
	public IntProperty getXProperty() {
		return x;
	}

	/**
	 * Get the X coordinate of this human.
	 * 
	 * @return The x coordinate of this human.
	 */
	public int getX() {
		return x.getValue();
	}

	/**
	 * Set the X coordinate of this human.
	 * 
	 * @param x
	 *            The new x coordinate.
	 */
	public void setX(int x) {
		this.x.setValue(x);
	}

	/**
	 * Find out if the x property has been defined.
	 * 
	 * @return True if the x property has been defined, false otherwise.
	 */
	public boolean isXDefined() {
		return x.isDefined();
	}

	/**
	 * Undefine the X property.
	 */
	public void undefineX() {
		x.undefine();
	}

	/**
	 * Get the y property.
	 * 
	 * @return The y property.
	 */
	public IntProperty getYProperty() {
		return y;
	}

	/**
	 * Get the y coordinate of this human.
	 * 
	 * @return The y coordinate of this human.
	 */
	public int getY() {
		return y.getValue();
	}

	/**
	 * Set the y coordinate of this human.
	 * 
	 * @param y
	 *            The new y coordinate.
	 */
	public void setY(int y) {
		this.y.setValue(y);
	}

	/**
	 * Find out if the y property has been defined.
	 * 
	 * @return True if the y property has been defined, false otherwise.
	 */
	public boolean isYDefined() {
		return y.isDefined();
	}

	/**
	 * Undefine the y property.
	 */
	public void undefineY() {
		y.undefine();
	}

	/**
	 * Get the travel distance property.
	 * 
	 * @return The travel distance property.
	 */
	public IntProperty getTravelDistanceProperty() {
		return travelDistance;
	}

	/**
	 * Get the travel distance.
	 * 
	 * @return The travel distance.
	 */
	public int getTravelDistance() {
		return travelDistance.getValue();
	}

	/**
	 * Set the travel distance.
	 * 
	 * @param d
	 *            The new travel distance.
	 */
	public void setTravelDistance(int d) {
		this.travelDistance.setValue(d);
	}

	/**
	 * Find out if the travel distance property has been defined.
	 * 
	 * @return True if the travel distance property has been defined, false
	 *         otherwise.
	 */
	public boolean isTravelDistanceDefined() {
		return travelDistance.isDefined();
	}

	/**
	 * Undefine the travel distance property.
	 */
	public void undefineTravelDistance() {
		travelDistance.undefine();
	}

	/**
	 * Get the entity represented by the position property. The result will be
	 * null if the position property has not been set or if the entity reference
	 * is invalid.
	 * 
	 * @param model
	 *            The WorldModel to look up entity references.
	 * @return The entity represented by the position property.
	 */
	public StandardEntity getPosition(WorldModel<? extends StandardEntity> model) {
		if (!position.isDefined()) {
			return null;
		}
		return model.getEntity(position.getValue());
	}

	public Entity getPosition(worldModel.WorldModel worldModel) {
		if (!position.isDefined()) {
			return null;
		}
		return worldModel.getEntity(position.getValue());
	}

	public void updateDistance(worldModel.WorldModel worldModel) {
		//
		HashMap<Area, Float> copy = new HashMap<Area, Float>();
		worldModel.clearAreas();
		distance = new HashMap<Area, Integer>();
		worldModel.Area area = ((rescuecore2.standard.entities.Area) getPosition(worldModel)).area;
		distance.put(area, 0);
		ArrayList<Entrance> layer = new ArrayList<Entrance>();
		ArrayList<Entrance> enterances = area.entrances;
		layer.addAll(enterances);
		for (Entrance e : layer)
			e.mark = true;
		while (layer.size() > 0) {
			ArrayList<Entrance> newLayer = new ArrayList<Entrance>();
			for (Entrance entrance : layer) {
				if (entrance.isItConnectedToNeighbour
						&& !entrance.neighbour.mark) {
					entrance.neighbour.mark = true;
					entrance.neighbour.lastEntrance = entrance;
					for (Entrance internal : entrance.neighbour.internalEntrances)
						if (!internal.mark) {
							newLayer.add(internal);
							internal.mark = true;
							internal.lastEntrance = entrance.neighbour;
						}
					if (entrance.neighbour.area.checkForAmbForBFS
							|| (worldModel.refuges.size() == 0 && entrance.neighbour.area instanceof Road)) {
						float dist = 0;
						Entrance copyOfEnterance = entrance.neighbour;
						while (copyOfEnterance.lastEntrance != null) {
							dist += copyOfEnterance.center
									.distance((copyOfEnterance.lastEntrance.center));
							copyOfEnterance = copyOfEnterance.lastEntrance;
						}
						copy.put(entrance.neighbour.area, dist);
						distance.put(entrance.neighbour.area,
								(int) dist / 30000);
					}
				}
			}
			layer = newLayer;
		}
		ArrayList<Integer> road = new ArrayList<Integer>();
		for (Entry<Area, Integer> dist : distance.entrySet()) {
			if (dist.getKey() instanceof Road)
				road.add(dist.getValue());
		}
		int bank = Integer.MAX_VALUE;
		for (int r : road) {
			if (r < bank)
				bank = r;
		}
		HashMap<Area, Integer> copyOfDistance = (HashMap<Area, Integer>) distance
				.clone();
		for (Entry<Area, Integer> dist : copyOfDistance.entrySet()) {
			if (dist.getKey() instanceof Road && dist.getValue() != bank
					&& !dist.getKey().checkForAmbForBFS)
				distance.remove(dist.getKey());
		}
	}

	/**
	 * Set the position of this human.
	 * 
	 * @param newPosition
	 *            The new position.
	 * @param newX
	 *            The x coordinate of this agent.
	 * @param newY
	 *            The y coordinate if this agent.
	 */
	public void setPosition(EntityID newPosition, int newX, int newY) {
		this.position.setValue(newPosition);
		this.x.setValue(newX);
		this.y.setValue(newY);
	}

	public Point2D getPositionPoint() {
		if (!isXDefined() || !isYDefined())
			return null;
		return new Point2D(getX(), getY());
	}

}
