package worldModel;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;

import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;

public class Area implements Serializable, Comparable<Area> {
	private static final long serialVersionUID = 6721173662089821339L;
	public boolean visited = false;
	public int surf = 0;
	public int counterForFindingHighways = 0;
	public int bfsDist = -1;
	private int weight = -10;
	public boolean isGoodForFireSearch = false;
	public transient rescuecore2.standard.entities.Area serverArea = null;
	public int id = -1;
	public Geometry circumstance = null;
	public transient Geometry completeShape = null;
	public transient Geometry shape = null;
	public ArrayList<Entrance> entrances = new ArrayList<Entrance>();
	public transient ArrayList<Blockade> blockades = new ArrayList<Blockade>();
	// public HashSet<Coordinate> matchedCoords = new HashSet<Coordinate>();
	public transient ArrayList<Area> nearAreas50000 = new ArrayList<Area>();
	public transient ArrayList<Area> nearAreas = new ArrayList<Area>();
	public transient ArrayList<Building> nearBuildings = new ArrayList<Building>();
	public ArrayList<Integer> nearBuildingsInts = new ArrayList<>();
	public transient ArrayList<Road> nearRoads = new ArrayList<Road>();
	public ArrayList<Integer> nearRoadsInts = new ArrayList<>();
	public transient ArrayList<Area> nearAreasIncludeSelf = new ArrayList<Area>();
	public transient ArrayList<Building> nearBuildingsIncludeSelf = new ArrayList<Building>();
	public transient ArrayList<Road> nearRoadsIncludeSelf = new ArrayList<Road>();
	public int zoneNumberForPolice = -1;
	public transient boolean[] isEmpty = null;
	//TODO CVmarke uses??
	public transient HashSet<Entrance> CVmarked = new HashSet<Entrance>();
	public transient ArrayList<Area> fireMarked = new ArrayList<Area>();
	public transient Geometry areaWithOutBlock = null;
	public boolean isReachable = false;
	public boolean checkForAmbForBFS = false;
	public int distanceFromSelf = -1;
	public int distFromeSelfWOB = -1;
	public int zoneNum = -1;
	public boolean checkedForTail = false;
	// last time visited
	public int LTV = -1;
	// heard last time visit
	public int HLTV = -1;
	// last time seen
	public int LTS = -1;
	// heard last time seen
	private int HLTS = -1;
	// heard lts for fire
	public int FHLTS = -1;
	public int num_zon = -1;
	public boolean isNearFire = false;
	public boolean check = false;
	public boolean buildingReachabilityCalculated = false;
	public boolean buildingReachability = false;
	public int fgNum = -1;
	public boolean checkSituation4Fire = false;
	public int heardlastTimeSeen = -1;
	public boolean canBeVisited = true;
	public transient ArrayList<FireBrigade> fireBrigades = new ArrayList<FireBrigade>();

	public void setTransients(WorldModel worldModel, GeometryFactory geom) {
		for (Entrance e : this.entrances) {
			e.setTransients(geom, this);
		}
		CVmarked = new HashSet<>();
		fireMarked = new ArrayList<>();
		fireBrigades = new ArrayList<>();
		blockades = new ArrayList<>();
		nearAreas = new ArrayList<>();
		nearRoads = new ArrayList<>();
		nearBuildings = new ArrayList<>();
		nearAreasIncludeSelf = new ArrayList<>();
		nearRoadsIncludeSelf = new ArrayList<>();
		nearBuildingsIncludeSelf = new ArrayList<>();
		nearAreas50000 = new ArrayList<>();
		for (int i : nearRoadsInts)
			nearRoads.add(worldModel.roads.get(i));
		for (int i : nearBuildingsInts)
			nearBuildings.add(worldModel.buildings.get(i));
		Collections.sort(nearRoads, worldModel.areaComparator);
		Collections.sort(nearBuildings, worldModel.areaComparator);
		nearAreas.addAll(nearRoads);
		nearAreas.addAll(nearBuildings);
		if (this instanceof Building)
			nearBuildingsIncludeSelf.add((Building) this);
		else if (this instanceof Road)
			nearRoadsIncludeSelf.add((Road) this);
		nearBuildingsIncludeSelf.addAll(nearBuildings);
		nearRoadsIncludeSelf.addAll(nearRoads);
		nearAreasIncludeSelf.addAll(nearRoadsIncludeSelf);
		nearAreasIncludeSelf.addAll(nearBuildingsIncludeSelf);
		for (StandardEntity e : worldModel.getModel().getObjectsInRange(
				getID(), worldModel.maxDistanceExtinguish))
			if (e instanceof rescuecore2.standard.entities.Area
					&& e.getID().getValue() != getID().getValue())
				nearAreas50000
						.add(((rescuecore2.standard.entities.Area) e).area);
		if (this instanceof Building)
			((Building) this).setTransients(worldModel);

		completeShape = GeometryTools.createPolygonFromApexes(
				serverArea.getApexList()).buffer(5); //new :-s
		shape = circumstance;
		areaWithOutBlock = shape;
		isEmpty = new boolean[worldModel.simAgent.cvRecognizer];
	}

	public void setHLTS(int time, int agentID,
			HashMap<Building, Integer> lastSeenBy, boolean setLastSeenBy) {
		HLTS = time;
		if (this instanceof worldModel.Building && setLastSeenBy)
			lastSeenBy.put((worldModel.Building) this, agentID);

	}

	public void setHLTS(int time, int agentID,
			HashMap<Building, Integer> lastSeenBy) {
		if (time > HLTS)
			setHLTS(time, agentID, lastSeenBy, true);
	}

	public int getHLTS() {
		return HLTS;
	}

	public int getX() {
		return serverArea.getX();
	}

	public int getY() {
		return serverArea.getY();
	}

	public int getLTS() {
		return (Math.max(LTS, HLTS));
	}

	public int getLTV() {
		return (Math.max(LTV, HLTV));
	}

	public int subWeight = 0;

	public int getWeight() {
		if (weight == -10)
			// TODO LTV ya LTS ?
			return getLTV();
		return weight;
	}

	public void setWeight(int weight) {
		this.weight = weight;
	}

	public Area(rescuecore2.standard.entities.Area area) {
		serverArea = area;
		serverArea.area = this;
		id = serverArea.getID().getValue();
		circumstance = GeometryTools
				.createPolygonFromApexes(area.getApexList());
		completeShape = circumstance.buffer(5);
		areaWithOutBlock = circumstance;
	}

	public void setAreaAsTarget(boolean isTarget) {
		for (Entrance enterance : entrances)
			enterance.isTarget = isTarget;
	}

	public EntityID getID() {
		return serverArea.getID();
	}

	public ArrayList<EntityID> getNeighbours() {
		return (ArrayList<EntityID>) serverArea.getNeighbours();
	}

	public String toString() {
		return "Area: " + serverArea.getID().getValue();
	}

	public void updateBlockades(StandardWorldModel model, int AGENT_RAD) {
		blockades = new ArrayList<Blockade>();
		if (serverArea.isBlockadesDefined())
			for (EntityID blockadeID : serverArea.getBlockades()) {
				rescuecore2.standard.entities.Blockade serverBlockade = (rescuecore2.standard.entities.Blockade) model
						.getEntity(blockadeID);
				if (serverBlockade == null)
					continue;
				Blockade blockade = new Blockade(serverBlockade, AGENT_RAD);
				blockades.add(blockade);
			}
	}

	public Edge getEdgeTo(EntityID neighbour) {
		for (Edge next : serverArea.getEdges()) {
			if (neighbour.equals(next.getNeighbour())) {
				return next;
			}
		}
		return null;
	}

	public boolean isBlockadesDefined() {
		return serverArea.isBlockadesDefined();
	}

	public ArrayList<Blockade> getBlockades() {
		return blockades;
	}

	public HashSet<Blockade> getBadBlockades(Coordinate firstPoint,
			Coordinate secondPoint, int rad, GeometryFactory geom) {
		HashSet<Blockade> badBlockades = new HashSet<Blockade>();
		LineSegment l = new LineSegment(firstPoint, secondPoint);
		Geometry gLine = l.toGeometry(geom).buffer(rad);
		for (Blockade blockade : blockades) {
			if (blockade.mainShape.intersects(gLine)
					|| blockade.mainShape.contains(gLine))
				badBlockades.add(blockade);
		}

		return badBlockades;
	}

	public void clearMarks() {
		for (Entrance enterance : entrances) {
			enterance.mark = false;
			enterance.lastEntrance = null;
			enterance.area.checkSituation4Fire = false;
		}
	}

	public int compareTo(Area o) {
		if (id > o.id)
			return 1;
		if (id < o.id)
			return -1;
		return 0;
	}

	public double distance(Area a) {
		return Math.hypot(getX() - a.getX(), getY() - a.getY());
	}

	public double distance(Coordinate c) {
		return Math.hypot(getX() - c.x, getY() - c.y);
	}
}
