package worldModel;

import java.io.Serializable;
import java.util.ArrayList;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;

public class Entrance implements Serializable, Comparable<Entrance> {
	private static final long serialVersionUID = 4164450602418567158L;

	public transient Geometry shapeInArea = null;
	public transient Area area = null;
	public Entrance neighbour = null;
	public LineSegment lineSegment = null;
	public transient Geometry lineSegmentGeom = null;
	private transient ArrayList<Coordinate> coords = new ArrayList<Coordinate>();
	private transient ArrayList<Geometry> points = new ArrayList<Geometry>();
	public transient ArrayList<Entrance> internalEntrances = new ArrayList<Entrance>();
	public boolean isItConnectedToNeighbour = true;
	public boolean updated = false;
	public boolean mark = false;

	public transient Coordinate center;
	public int id = -1;

	// BFS params
	public boolean marked = false;
	public boolean isReachable = false;
	public int distanceFromSelf = -1;
	public boolean isTarget = false;
	public transient Entrance lastEntrance = null;
	public double distForDijk = 1000000000;
	public transient Entrance prev = null;
	public transient Entrance preEntrance = null;

	public transient GeometryFactory geom = null;

	public void setTransients(GeometryFactory geom, Area area) {
		this.geom = geom;
		this.area = area;
		coords = new ArrayList<>();
		resetShapeInArea();
		lineSegmentGeom = lineSegment.toGeometry(geom);
		center = lineSegment.midPoint();
		for (int i = 0; i < 2; i++)
			coords.add(lineSegment.getCoordinate(i));
		updatePointsByCoords();
	}

	public Entrance(Area area, LineSegment lineSegment, GeometryFactory geom,
			int id) {
		this.id = id;
		this.area = area;
		this.lineSegment = lineSegment;
		this.geom = geom;
		lineSegmentGeom = lineSegment.toGeometry(geom);
		center = lineSegment.midPoint();
		for (int i = 0; i < 2; i++)
			coords.add(lineSegment.getCoordinate(i));
		updatePointsByCoords();
		resetShapeInArea();
	}

	public void resetShapeInArea() {
		shapeInArea = area.circumstance;
	}

	public void updatePointsByCoords() {
		points = new ArrayList<Geometry>();
		for (Coordinate c : coords) {
			points.add(geom.createPoint(c));
		}
	}

	public void updatePoints(ArrayList<Blockade> blockades) {
		coords = new ArrayList<Coordinate>();
		Geometry newShape = lineSegmentGeom;
		for (Blockade blockade : blockades) {
			newShape = newShape.difference(blockade.shape);
		}
		for (int i = 0; i < newShape.getNumGeometries(); i++) {
			Geometry s = newShape.getGeometryN(i);
			for (Coordinate c : s.getCoordinates())
				coords.add(c);
		}
		updatePointsByCoords();
	}

	public void updateConnectivity(ArrayList<Blockade> blockades,
			boolean isNeighbourInChangeSet) {
		if (isNeighbourInChangeSet || isItConnectedToNeighbour) {
			// System.out.println("here: " + this);
			isItConnectedToNeighbour = false;
			if (blockades.size() == 0)
				isItConnectedToNeighbour = true;
			else if (points.size() > 1 && neighbour.points.size() > 1) {
				Geometry shape = lineSegmentGeom.buffer(10, 1);
				for (Blockade blockade : blockades)
					shape = shape.difference(blockade.circumstance);

				for (int i = 0; i < shape.getNumGeometries(); i++) {
					Geometry s = shape.getGeometryN(i);

					boolean isPointInShape = false;
					for (Geometry p : points)
						if (s.contains(p)) {
							isPointInShape = true;
							break;
						}
					if (!isPointInShape)
						continue;
					isPointInShape = false;
					for (Geometry p : neighbour.points)
						if (s.contains(p)) {
							isPointInShape = true;
							break;
						}
					if (isPointInShape) {
						isItConnectedToNeighbour = true;
						break;
					}
				}
			}

			neighbour.isItConnectedToNeighbour = isItConnectedToNeighbour;
		}
		// System.out.println("at end: " + isItConnectedToNeighbour);
		// System.out.println("neighbour: " +
		// neighbour.isItConnectedToNeighbour);
	}

	public ArrayList<Geometry> getPoints() {
		return points;
	}

	public int compareTo(Entrance other) {
		if (distForDijk > other.distForDijk)
			return 1;
		if (distForDijk < other.distForDijk)
			return -1;
		if (area.getID().getValue() > other.area.getID().getValue())
			return 1;
		if (area.getID().getValue() < other.area.getID().getValue())
			return -1;
		if (id > other.id)
			return 1;
		if (id < other.id)
			return -1;
		return 0;
	}

	public String toString() {
		// return area.id + "-" + neighbour.area.id;
		return area.getID().getValue() + "-" + neighbour.area.id + " "
				+ GeometryTools.printLine(lineSegment);
	}
}
