package worldModel;

import geometry.Mathematic;

import java.util.ArrayList;

import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.Human;

import com.infomatiq.jsi.Rectangle;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.triangulate.Segment;

public class GeometryTools {
	public static Polygon createPolygonFromApexes(int[] apexes) {
		ArrayList<Coordinate> coords = new ArrayList<Coordinate>();
		for (int i = 0; i < apexes.length; i += 2)
			coords.add(new Coordinate(apexes[i], apexes[i + 1]));
		coords.add(new Coordinate(apexes[0], apexes[1]));
		return createPolygon(coords);
	}

	public static float getDeltaAngle(float angle1, float angle2) {
		angle1 = absoluteAngle(angle1);
		angle2 = absoluteAngle(angle2);
		return normalizeAngle(angle1 - angle2);
	}

	// public static Coordinate minus(Coordinate point1, Coordinate point2) {
	// double a = 0, b = 0;
	// System.out.println("p1x : " + point1.x + "p1y : " + point1.y);
	// System.out.println("p2x : " + point2.x + "p2y : " + point2.y);
	// a = (double) point1.x - (double) point2.x;
	// b = (double) point1.y - (double) point2.y;
	// System.out.println("a: " + a);
	// System.out.println("b: " + b);
	// Coordinate h = new Coordinate(a, b);
	// return h;
	// }

	public static float getTheta(Segment a) {
		return (float) a.getLineSegment().angle();
	}

	public static float absoluteAngle(float theta) {
		while (theta > 2 * Math.PI || theta < 0) {
			if (theta > Math.PI)
				theta -= 2 * Math.PI;
			else
				theta += 2 * Math.PI;
		}

		return theta;
	}

	public static boolean isIntersectBetweenLines(Segment l1, Segment l2) {
		return isIntersectBetweenLines(l1.getStart(), l1.getEnd(),
				l2.getStart(), l2.getEnd());
	}

	public static int getDirection(Coordinate p1, Coordinate p2, Coordinate p3) {
		// int v1 = (p2.getY() - p1.getY()) * (p3.getX() - p2.getX());
		// int v2 = (p3.getY() - p2.getY()) * (p2.getX() - p1.getX());
		int p1X = (int) p1.x;
		int p1Y = (int) p1.y;
		int p2X = (int) p2.x;
		int p2Y = (int) p2.y;
		int p3X = (int) p3.x;
		int p3Y = (int) p3.y;

		int minX = Math.min(p1X, Math.min(p2X, p3X));
		int minY = Math.min(p1Y, Math.min(p2Y, p3Y));
		p1X -= minX;
		p2X -= minX;
		p3X -= minX;
		p1Y -= minY;
		p2Y -= minY;
		p3Y -= minY;

		long v1 = ((long) p2Y - (long) p1Y) * ((long) p3X - (long) p2X);
		long v2 = ((long) p3Y - (long) p2Y) * ((long) p2X - (long) p1X);
		if (v1 > v2)
			return 1;
		if (v1 < v2)
			return -1;
		return 0;
	}

	public static Geometry coordToGeometry(Coordinate c, GeometryFactory geom) {
		return geom.createPoint(c);
	}

	public static boolean isIntersectBetweenLines(Coordinate p1, Coordinate q1,
			Coordinate p2, Coordinate q2) {
		int p1q1p2 = getDirection(p1, q1, p2);
		int p1q1q2 = getDirection(p1, q1, q2);
		int p2q2p1 = getDirection(p2, q2, p1);
		int p2q2q1 = getDirection(p2, q2, q1);
		if (p1q1p2 * p1q1q2 != 1 && p2q2p1 * p2q2q1 != 1)
			return true;

		if ((p1q1p2 == 0 && p1q1q2 == 0 && p2q2p1 == 0 && p2q2q1 == 0)
				&& (Mathematic.isInBetween((float) p1.x, (float) p2.x,
						(float) q1.x)
						|| Mathematic.isInBetween((float) p1.x, (float) q2.x,
								(float) q1.x) || Mathematic.isInBetween(
						(float) p2.x, (float) p1.x, (float) q2.x))
				&& (Mathematic.isInBetween((float) p1.y, (float) p2.y,
						(float) q1.y)
						|| Mathematic.isInBetween((float) p1.y, (float) q2.y,
								(float) q1.y) || Mathematic.isInBetween(
						(float) p2.y, (float) p1.y, (float) q2.y)))
			return true;

		return false;
	}

	public static float getDegree(Coordinate pos1, Coordinate pos2,
			Coordinate pos3) {
		Coordinate s = new Coordinate();
		Coordinate d = new Coordinate();
		s.x = pos2.x;
		double m = pos3.x;
		s.x = s.x - m;
		s.y = pos2.y - pos3.y;
		d.x = pos1.x - pos2.x;
		d.y = pos1.y - pos2.y;
		Segment v1 = new Segment(new Coordinate(0, 0), s), v2 = new Segment(
				new Coordinate(0, 0), d);
		return normalizeAngle(getTheta(v1) - getTheta(v2));
	}

	public static boolean isClockWise(Coordinate p1, Coordinate p2,
			Coordinate p3) {
		// if (p1 == null || p2 == null || p3 == null)
		// System.out.println("(tmp)p1: " + p1 + " (BP)p2: " + p2
		// + " (MP)p3: " + (p3));
		return getDegree(p1, p2, p3) > 0;
	}

	public static float normalizeAngle(float theta) {
		while (theta > Math.PI || theta < -Math.PI) {
			if (theta > Math.PI)
				theta -= 2 * Math.PI;
			else
				theta += 2 * Math.PI;
		}

		return theta;
	}

	public static Polygon createPolygon(ArrayList<Coordinate> coordinates) {
		LinearRing shell = new GeometryFactory().createLinearRing(coordinates
				.toArray(new Coordinate[0]));
		return new GeometryFactory().createPolygon(shell, null);
	}

	public static LineSegment getClearLineSegment(Human agent, int targetX,
			int targetY, int clearLength, int clearRad, GeometryFactory geom) {
		Vector2D v = new Vector2D(targetX - agent.getX(), targetY
				- agent.getY());
		v = v.normalised().scale(clearLength);
		return new LineSegment(
				new Coordinate(agent.getX(), agent.getY()),
				new Coordinate(agent.getX() + v.getX(), agent.getY() + v.getY()));
	}

	public static Geometry getClearArea(Human agent, int targetX, int targetY,
			int clearLength, int clearRad, GeometryFactory geom) {
		return getClearLineSegment(agent, targetX, targetY, clearLength,
				clearRad, geom).toGeometry(geom).buffer(clearRad);
	}

	public static Rectangle makeRectangle(Coordinate[] coordinates) {
		int x1 = Integer.MAX_VALUE;
		int x2 = -Integer.MAX_VALUE;
		int y1 = Integer.MAX_VALUE;
		int y2 = -Integer.MAX_VALUE;
		for (Coordinate c : coordinates) {
			if (c.x < x1)
				x1 = (int) c.x;
			if (c.x > x2)
				x2 = (int) c.x;
			if (c.y < y1)
				y1 = (int) c.y;
			if (c.y > y2)
				y2 = (int) c.y;
		}
		// System.out.println(x1 + "  " + y1 + "  " + x2 + "  " + y2);
		return new Rectangle(x1, y1, x2, y2);
	}

	public static Rectangle makeRectangle(Coordinate[] coordinates, Boolean b) {
		int x1 = Integer.MAX_VALUE;
		int x2 = -Integer.MAX_VALUE;
		int y1 = Integer.MAX_VALUE;
		int y2 = -Integer.MAX_VALUE;
		for (Coordinate c : coordinates) {
			if (c.x < x1)
				x1 = (int) c.x;
			if (c.x > x2)
				x2 = (int) c.x;
			if (c.y < y1)
				y1 = (int) c.y;
			if (c.y > y2)
				y2 = (int) c.y;
		}
		// if (b)
		// System.out.println(x1 + "  " + y1 + "  " + x2 + "  " + y2);
		return new Rectangle(x1, y1, x2, y2);
	}

	public static Rectangle makeRectangle(Geometry shape) {
		return makeRectangle(shape.getCoordinates());
	}

	public static Rectangle makeRectangle(LineSegment lineSegment) {
		Coordinate[] coordinates = { lineSegment.getCoordinate(0),
				lineSegment.getCoordinate(1) };
		return makeRectangle(coordinates);
	}

	public static String printShape(Geometry shape) {
		String res = "";
		// for (int i = 0; i < shape.getCoordinates().length - 1; i++) {
		// res += "line(" + (int) shape.getCoordinates()[i].x + ", "
		// + (int) shape.getCoordinates()[i].y + ")("
		// + (int) shape.getCoordinates()[i + 1].x + ", "
		// + (int) shape.getCoordinates()[i + 1].y + ") 4; ";
		// }
		return res;
	}

	public static String printShape(Geometry shape, int color) {
		String res = "";
		// for (int i = 0; i < shape.getCoordinates().length - 1; i++) {
		// res += "line(" + (int) shape.getCoordinates()[i].x + ", "
		// + (int) shape.getCoordinates()[i].y + ")("
		// + (int) shape.getCoordinates()[i + 1].x + ", "
		// + (int) shape.getCoordinates()[i + 1].y + ") " + color
		// + "; ";
		// }
		return res;
	}

	public static String printShapeByCircle(Geometry shape) {
		String res = "";
		// for (int i = 0; i < shape.getCoordinates().length; i++) {
		// res += "circle(" + (int) shape.getCoordinates()[i].x + ", "
		// + (int) shape.getCoordinates()[i].y + ", 500) 4; ";
		// }
		return res;
	}

	public static String printLine(LineSegment ls) {
		return "line(" + (int) ls.getCoordinate(0).x + ", "
				+ (int) ls.getCoordinate(0).y + ")("
				+ (int) ls.getCoordinate(1).x + ", "
				+ (int) ls.getCoordinate(1).y + ") 4; ";
	}

	public static String printPoint(Coordinate coordinate) {
		return "circle(" + (int) coordinate.x + ", " + (int) coordinate.y
				+ ", 500) 4; ";
	}

	public static String printLine(double x1, double y1, double x2, double y2) {
		return "line(" + (int) x1 + ", " + (int) y1 + ")(" + (int) x2 + ", "
				+ (int) y2 + ") 0;";
	}

	public static String printCircle(double x, double y, int rad) {
		return "circle(" + (int) x + ", " + (int) y + ", " + rad + ") 4;";
	}

	public static String printLine(Coordinate start, Coordinate end) {
		return "line(" + (int) start.x + ", " + (int) start.y + ")("
				+ (int) end.x + ", " + (int) end.y + ") 4; ";
	}
}
